Package: dygraphs
Type: Package
Title: Interface to 'Dygraphs' Interactive Time Series Charting Library
Version: 1.1.1.6
Authors@R: c(
    person("Dan", "Vanderkam", role = c("aut", "cph"),
       comment = "dygraphs library in htmlwidgets/lib, http://dygraphs.com/"),
    person("Petr", "Shevtsov", role = c("cre", "cph"),
       email = "petr.shevtsov@gmail.com"),
    person("JJ", "Allaire", role = c("aut")),
    person(family = "RStudio", role = "cph"),
    person("Jonathan", "Owen", role = c("aut", "cph")),
    person("Daniel", "Gromer", role = c("aut", "cph")),
    person("Benoit", "Thieurmel", role = c("aut", "cph")),
    person("Kent", "Laukhuf", role = "ctb"),
    person(family = "jQuery Foundation", role = "cph",
           comment = "jQuery library"),
    person(family = "jQuery contributors", role = c("ctb", "cph"),
           comment = "jQuery library; authors listed in inst/htmlwidgets/lib/jquery/AUTHORS.txt")
    )
URL: https://github.com/rstudio/dygraphs
BugReports: https://github.com/rstudio/dygraphs/issues
Description: An R interface to the 'dygraphs' JavaScript charting library
    (a copy of which is included in the package). Provides rich facilities
    for charting time-series data in R, including highly configurable
    series- and axis-display and interactive features like zoom/pan and
    series/point highlighting.
License: MIT + file LICENSE
Depends: R (>= 3.0)
Imports: magrittr, htmlwidgets (>= 0.6), htmltools (>= 0.3.5), zoo (>=
        1.7-10), xts (>= 0.9-7)
Suggests: testthat
Enhances: rmarkdown (>= 0.3.3), shiny (>= 0.10.2.1)
RoxygenNote: 6.0.1
NeedsCompilation: no
Packaged: 2018-07-11 21:29:28 UTC; petr
Author: Dan Vanderkam [aut, cph] (dygraphs library in htmlwidgets/lib,
    http://dygraphs.com/),
  Petr Shevtsov [cre, cph],
  JJ Allaire [aut],
  RStudio [cph],
  Jonathan Owen [aut, cph],
  Daniel Gromer [aut, cph],
  Benoit Thieurmel [aut, cph],
  Kent Laukhuf [ctb],
  jQuery Foundation [cph] (jQuery library),
  jQuery contributors [ctb, cph] (jQuery library; authors listed in
    inst/htmlwidgets/lib/jquery/AUTHORS.txt)
Maintainer: Petr Shevtsov <petr.shevtsov@gmail.com>
Repository: CRAN
Date/Publication: 2018-07-11 22:00:03 UTC
Built: R 4.6.0; ; 2026-04-26 16:02:04 UTC; unix

Package: dygraphs
Type: Package
Title: Interface to 'Dygraphs' Interactive Time Series Charting Library
Version: 1.1.1.6
Authors@R: c(
    person("Dan", "Vanderkam", role = c("aut", "cph"),
       comment = "dygraphs library in htmlwidgets/lib, http://dygraphs.com/"),
    person("Petr", "Shevtsov", role = c("cre", "cph"),
       email = "petr.shevtsov@gmail.com"),
    person("JJ", "Allaire", role = c("aut")),
    person(family = "RStudio", role = "cph"),
    person("Jonathan", "Owen", role = c("aut", "cph")),
    person("Daniel", "Gromer", role = c("aut", "cph")),
    person("Benoit", "Thieurmel", role = c("aut", "cph")),
    person("Kent", "Laukhuf", role = "ctb"),
    person(family = "jQuery Foundation", role = "cph",
           comment = "jQuery library"),
    person(family = "jQuery contributors", role = c("ctb", "cph"),
           comment = "jQuery library; authors listed in inst/htmlwidgets/lib/jquery/AUTHORS.txt")
    )
URL: https://github.com/rstudio/dygraphs
BugReports: https://github.com/rstudio/dygraphs/issues
Description: An R interface to the 'dygraphs' JavaScript charting library
    (a copy of which is included in the package). Provides rich facilities
    for charting time-series data in R, including highly configurable
    series- and axis-display and interactive features like zoom/pan and
    series/point highlighting.
License: MIT + file LICENSE
Depends: R (>= 3.0)
Imports: magrittr, htmlwidgets (>= 0.6), htmltools (>= 0.3.5), zoo (>=
        1.7-10), xts (>= 0.9-7)
Suggests: testthat
Enhances: rmarkdown (>= 0.3.3), shiny (>= 0.10.2.1)
RoxygenNote: 6.0.1
NeedsCompilation: no
Packaged: 2018-07-11 21:29:28 UTC; petr
Author: Dan Vanderkam [aut, cph] (dygraphs library in htmlwidgets/lib,
    http://dygraphs.com/),
  Petr Shevtsov [cre, cph],
  JJ Allaire [aut],
  RStudio [cph],
  Jonathan Owen [aut, cph],
  Daniel Gromer [aut, cph],
  Benoit Thieurmel [aut, cph],
  Kent Laukhuf [ctb],
  jQuery Foundation [cph] (jQuery library),
  jQuery contributors [ctb, cph] (jQuery library; authors listed in
    inst/htmlwidgets/lib/jquery/AUTHORS.txt)
Maintainer: Petr Shevtsov <petr.shevtsov@gmail.com>
Repository: CRAN
Date/Publication: 2018-07-11 22:00:03 UTC

# Generated by roxygen2: do not edit by hand

export("%>%")
export(as.yearmon)
export(as.yearqtr)
export(dyAnnotation)
export(dyAxis)
export(dyBarChart)
export(dyBarSeries)
export(dyCSS)
export(dyCallbacks)
export(dyCandlestick)
export(dyCandlestickGroup)
export(dyCrosshair)
export(dyDataHandler)
export(dyDependency)
export(dyErrorFill)
export(dyEvent)
export(dyFilledLine)
export(dyGroup)
export(dyHighlight)
export(dyLegend)
export(dyLimit)
export(dyMultiColumn)
export(dyMultiColumnGroup)
export(dyOptions)
export(dyPlotter)
export(dyPlugin)
export(dyRangeSelector)
export(dyRebase)
export(dyRibbon)
export(dyRoller)
export(dySeries)
export(dySeriesData)
export(dyShading)
export(dyShadow)
export(dyStackedBarChart)
export(dyStackedBarGroup)
export(dyStackedLineGroup)
export(dyStackedRibbonGroup)
export(dyStemSeries)
export(dyUnzoom)
export(dygraph)
export(dygraphOutput)
export(renderDygraph)
importFrom(grDevices,rgb)
importFrom(htmltools,htmlDependency)
importFrom(htmlwidgets,JS)
importFrom(magrittr,"%>%")
importFrom(stats,end)
importFrom(stats,start)
importFrom(zoo,as.yearmon)
importFrom(zoo,as.yearqtr)

# dygraphs JavaScript charting library

The dygraphs JavaScript library produces interactive, zoomable charts of time series:

![sample graph](Screenshot.png)

Learn more about it at [dygraphs.com](http://www.dygraphs.com).

Get help with dygraphs on [Stack Overflow][] (preferred) and [Google Groups][].

## Features
* Plots time series without using an external server or Flash
* Supports [error bands][] around data series
* Interactive [pan and zoom][]
* Displays values [on mouseover][]
* Adjustable [averaging period][]
* Extensive set of [options][] for customization.
* Compatible with the [Google Visualization API][gviz]

## Minimal Example
```html
<html>
<head>
<script type="text/javascript" src="dygraph.js"></script>
<link rel="stylesheet" type="text/css" href="dygraph.css" />
</head>
<body>
<div id="graphdiv"></div>
<script type="text/javascript"><!--//--><![CDATA[//><!--
  Dygraph.onDOMready(function onDOMready() {  // or jQuery $() etc.
    g = new Dygraph(
        document.getElementById("graphdiv"),  // containing div
        "Date,Temperature\n" +                // the data series
        "2008-05-07,75\n" +
        "2008-05-08,70\n" +
        "2008-05-09,80\n",
        { }                                   // the options
      );
  });
//--><!]]></script>
</body>
</html>
```

Learn more by reading [the tutorial][] and seeing demonstrations of what
dygraphs can do in the [gallery][]. You can get `dygraph.js` and `dygraph.css`
from [UNPKG][], [cdnjs][] or [from NPM][npm] (see below).

## Usage with a module loader

Get dygraphs from NPM:

    npm install dygraphs

**Do not install from the git repository!** Always use a tarball install,
which contains the prebuilt files; `npm` fails to build the source code
on install from github. (The tarball from the GitHub Registry is fine.)

You'll find pre-built JS & CSS files in `node_modules/dygraphs/dist/`. If you're
using a module bundler like browserify or webpack, you can import dygraphs:

```js
import Dygraph from 'dygraphs';
// or: const Dygraph = require('dygraphs');

const g = new Dygraph('graphdiv', data, { /* options */ });
```

Check out the [dygraphs-es6 repo][] for a fully-worked example.

## Development

To get going, install the following Debian packages…

 - `jq`
 - `mksh`
 - `pax`
 - `python3`

… clone the repo and run:

    npm install
    npm run build-jsonly

Then open `tests/demo.html` in your browser.

Read more about the dygraphs development process in the [developer guide](/DEVELOP.md).

## License(s)
dygraphs is available under the MIT license, included in [LICENSE.txt](./LICENSE.txt).

[UNPKG]: https://unpkg.com/dygraphs/
[cdnjs]: https://cdnjs.com/libraries/dygraph
[the tutorial]: http://www.dygraphs.com/tutorial.html
[gallery]: http://www.dygraphs.com/gallery
[error bands]: http://dygraphs.com/tests/legend-values.html
[pan and zoom]: http://dygraphs.com/tests/link-interaction.html
[on mouseover]: http://dygraphs.com/tests/legend-values.html
[averaging period]: http://dygraphs.com/tests/temperature-sf-ny.html
[options]: http://www.dygraphs.com/options.html
[Stack Overflow]: https://stackoverflow.com/questions/tagged/dygraphs?sort=votes&pageSize=50
[Google Groups]: http://groups.google.com/group/dygraphs-users
[gviz]: http://dygraphs.com/data.html#datatable
[npm]: https://www.npmjs.com/package/dygraphs
[dygraphs-es6 repo]: https://github.com/danvk/dygraphs-es6

Online sources of some of these documents locally mirrored
so these can be used without needing to visit external services:

• datahandler-proposal.pdf
  https://docs.google.com/document/d/1IWmX4oDbQbVtUoRNzSRG3yMpoBQ7LseVCQhGnuOZz_A
• hairlines-annotations.pdf
  https://docs.google.com/document/d/1OHNE8BNNmMtFlRQ969DACIYIJ9VVJ7w3dSPRJDEeIew
• legend-formatter.md
  https://github.com/danvk/dygraphs/pull/683
• smooth-plotter.md + smooth-plotter-p*
  https://github.com/danvk/dygraphs/pull/469

Tips on adding entries to the gallery

* You can use the id in conjunction with specialized CSS.

* Most files don't have a lot of HTML, but some have tons. See
  independent-series for tips on adding a lot of HTML.

* You can also define a function called cleanup, which is the
  opposite of setup. It's perfect for stopping timers, for instance.

* Callbacks in HTML widgets. You might assume you can use
  <button onclick="func">
  and then define func in your run script:

  run: function() { function func() { .. }; }

  but because func is defined in function scope the button can't call it.
  The solution is to put func in the global scope, which is the same as window
  scope:

  run: function() { function window.func() { .. }; }

* You'll find it easier if you convert all the double-quotes in your HTML to
  single quotes.

TODOs:
  * Remove the awful tables in interaction.html (that I wrote)
  * Move all style to css
  * Add margin-left: 30px to #workarea, and then make #workarea h2 { margin-left: -30px}
  * Remove runtime warnings.

library(dygraphs)

shinyUI(fluidPage(
  
  titlePanel("Predicted Deaths from Lung Disease (UK)"),
  
  sidebarLayout(
    sidebarPanel(
      numericInput("months", label = "Months to Predict", 
                   value = 72, min = 12, max = 144, step = 12),
      selectInput("interval", label = "Prediction Interval",
                  choices = c("0.80", "0.90", "0.95", "0.99"),
                  selected = "0.95"),
      checkboxInput("showgrid", label = "Show Grid", value = TRUE),
      hr(),
      div(strong("From: "), textOutput("from", inline = TRUE)),
      div(strong("To: "), textOutput("to", inline = TRUE)),
      div(strong("Date clicked: "), textOutput("clicked", inline = TRUE)),
      div(strong("Nearest point clicked: "), textOutput("point", inline = TRUE)),
      br(),
      helpText("Click and drag to zoom in (double click to zoom back out).")
    ),
    mainPanel(
      dygraphOutput("dygraph")
    )
  )
))

library(dygraphs)

hw <- HoltWinters(ldeaths)
p <- predict(hw, n.ahead = 36, prediction.interval = TRUE, level = 0.95)
ts <- cbind(ldeaths, p)

dygraph(ts, "Deaths from Lung Disease (UK)") %>%
  dySeries("ldeaths", label = "Deaths") %>%
  dySeries(c("p.lwr", "p.fit", "p.upr"), label = "Predicted") %>%
  dyRangeSelector()

library(dygraphs)
library(datasets)

shinyServer(function(input, output) {
  
  predicted <- reactive({
    hw <- HoltWinters(ldeaths)
    predict(hw, n.ahead = input$months, 
            prediction.interval = TRUE,
            level = as.numeric(input$interval))
  })
  
  output$dygraph <- renderDygraph({
    dygraph(predicted(), main = "Predicted Deaths/Month") %>%
      dySeries(c("lwr", "fit", "upr"), label = "Deaths") %>%
      dyOptions(drawGrid = input$showgrid)
  })
  
  output$from <- renderText({
    strftime(req(input$dygraph_date_window[[1]]), "%d %b %Y")      
  })
  
  output$to <- renderText({
    strftime(req(input$dygraph_date_window[[2]]), "%d %b %Y")
  })
  
  output$clicked <- renderText({
    strftime(req(input$dygraph_click$x), "%d %b %Y")
  })

  output$point <- renderText({
    paste0('X = ', strftime(req(input$dygraph_click$x_closest_point), "%d %b %Y"), 
         '; Y = ', req(input$dygraph_click$y_closest_point))
  })
})

#' dygraph exported operators and S3 methods
#' 
#' The following functions are imported and then re-exported
#' from the dygraphs package to avoid listing the magrittr and
#' zoo packages as Depends of dygraphs.
#' 
#' @name dygraphs-exports
NULL

#' @importFrom magrittr %>%
#' @name %>%
#' @export
#' @rdname dygraphs-exports
NULL

#' @importFrom zoo as.yearmon
#' @name as.yearmon
#' @export as.yearmon
#' @rdname dygraphs-exports
NULL

#' @importFrom zoo as.yearqtr
#' @name as.yearqtr
#' @export as.yearqtr
#' @rdname dygraphs-exports 
NULL

#' Annotation for dygraph chart
#' 
#' Define a text annotation for a data-point on a dygraph chart.
#' 
#' @param dygraph Dygraph to add an annotation to
#' @param x Either numeric or date value indicating where to place the
#'   annotation. For date value, this should be of class \code{POSIXct} or
#'   convertible to \code{POSIXct}.
#' @param text Text to overlay on the chart at the location of x
#' @param tooltip Additional tooltip text to display on mouse hover
#' @param width Width (in pixels) of the annotation flag.
#' @param height Height (in pixels) of the annotation flag.
#' @param cssClass CSS class to use for styling the annotation.
#' @param tickHeight Height of the tick mark (in pixels) connecting the point to
#'   its flag or icon.
#' @param attachAtBottom If true, attach annotations to the x-axis, rather than 
#'   to actual points.
#' @param clickHandler JavaScript function to call when an annotation is 
#'   clicked.
#' @param mouseOverHandler JavaScript function to call when the mouse hovers 
#'   over an annotation.
#' @param mouseOutHandler JavaScript function to call when the mouse exits an 
#'   annotation.
#' @param dblClickHandler JavaScript function to call when an annotation is 
#'   double clicked.
#' @param series Series to attach the annotation to. By default, the last series
#'   defined using \code{\link{dySeries}}.
#'   
#' @note Annotations are bound to specific series in the input data. If you have
#'   only one series or if you define annotations immediately after a call to 
#'   \code{\link{dySeries}} then you need not specify the series explicitly. 
#'   Otherwise, you should use the \code{series} parameter to indicate which
#'   series the annotation should be bound to.
#'   
#'   Annotation event handlers can also specified globally (see 
#'   \code{\link{dyCallbacks}}).
#'   
#' @note See the
#'   \href{https://rstudio.github.io/dygraphs/gallery-annotations.html}{online
#'   documentation} for additional details and examples.
#'   
#' @return Dygraph with specified annotation
#' 
#' @examples 
#' library(dygraphs)
#'   
#' dygraph(presidents, main = "Presidential Approval") %>%
#'   dyAxis("y", valueRange = c(0, 100)) %>%
#'   dyAnnotation("1950-7-1", text = "A", tooltip = "Korea") %>%
#'   dyAnnotation("1965-1-1", text = "B", tooltip = "Vietnam")   
#'   
#' @export
dyAnnotation <- function(dygraph,
                         x,
                         text, 
                         tooltip = NULL,
                         width = NULL,
                         height = NULL,
                         cssClass = NULL,
                         tickHeight = NULL,
                         attachAtBottom = FALSE,
                         clickHandler = NULL,
                         mouseOverHandler = NULL,
                         mouseOutHandler = NULL,
                         dblClickHandler = NULL,
                         series = NULL) {
  
  # convert x to date format then to a suitable time value if it is date format
  if (dygraph$x$format == "date")
    x <- asISO8601Time(x)
  
  # validate series if specified
  if (!is.null(series) && ! series %in% dygraph$x$attrs$labels) {
    stop("The specified series was not found. Valid series names are: ", 
         paste(dygraph$x$attrs$labels[-1], collapse = ", "))
  }
  
  # default the series if necessary
  if (is.null(series))
    series <- dygraph$x$attrs$labels[length(dygraph$x$attrs$labels)]
  
  # create annotation
  annotation <- list()
  annotation$x <- x
  annotation$shortText <- text
  annotation$text <- tooltip
  annotation$width <- width
  annotation$height <- height
  annotation$cssClass <- cssClass
  annotation$tickHeight <- tickHeight
  annotation$attachAtBottom <- attachAtBottom
  annotation$clickHandler <- JS(clickHandler)
  annotation$mouseOverHandler <- JS(mouseOverHandler)
  annotation$mouseOutHandler <- JS(mouseOutHandler)
  annotation$dblClickHandler <- JS(dblClickHandler)
  annotation$series <- series
  
  # add it
  dygraph$x$annotations[[length(dygraph$x$annotations) + 1]] <- annotation

  # return the dygraph
  dygraph
}

#' @rdname Plotters
#' @export
dyCandlestick <- function(dygraph, compress = FALSE) {
  path <- system.file("plotters/candlestick.js", package = "dygraphs")
  path <- normalizePath(path)
  dygraph <- dyPlotter(dygraph = dygraph,
                       name = "CandlestickPlotter",
                       path = path,
                       version = "1.0")

  if (compress) {
    path <- system.file("plugins/compress.js", package = "dygraphs")
    path <- normalizePath(path)
    dygraph <- dyDataHandler(dygraph = dygraph,
                             name = "CompressHandler",
                             path = path,
                             version = "1.0")
    dygraph <- dyAxis(dygraph, "x", valueFormatter = "
    function(millis) {
      var monthNames = ['Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun',
                        'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec'];
      var scale = this.dataHandler_.compressTitle;
      var date = new Date(millis);
      if (scale == 'yearly')
        return date.getFullYear();
      else if (scale == 'quarterly')
        return moment(millis).fquarter(1);
      else if (scale == 'monthly')
        return monthNames[date.getMonth()] + ', ' + date.getFullYear();
      else if (scale == 'daily' || scale == 'weekly')
        return monthNames[date.getMonth()] + ', ' +
          date.getDate() + ', ' +
          date.getFullYear();
        else
          return date.toLocaleString();
    }")
  }

  # return dygraph
  dygraph
}

#' CSS for dygraph labels and legend
#' 
#' Apply custom CSS to the text drawn within a dygraph. See the 
#' \href{http://dygraphs.com/css.html}{CSS documentation} on the dygraphs 
#' website for additional details on which styles are available.
#' 
#' @param dygraph Dygraph to add CSS styles to
#' @param css Path to css file to be used for styling textual elements of the 
#'   graph.
#'   
#' @return dygraph with additional CSS styles
#'   
#' @note See the 
#'   \href{https://rstudio.github.io/dygraphs/gallery-css-styling.html}{CSS 
#'   Styling} article on the dygraphs for R website for additional details.
#'   
#'   Note that CSS styles are global so will affect all dygraphs on a given web
#'   page. This also implies that for a page with multiple plots you only need
#'   to specify styles for the first one (alternatively you can just add them
#'   directly to the page by other means).
#'   
#' @export
dyCSS <- function(dygraph, css) {
  
  # read css
  dygraph$x$css <- paste(readLines(css, warn = FALSE), collapse = "\n")
  
  # return modified dygraph
  dygraph
}

#' dygraph options
#' 
#' Add options to a dygraph plot.
#' 
#' @param dygraph Dygraph to add options to
#' @param stackedGraph If set, stack series on top of one another rather than 
#'   drawing them independently. The first series specified in the input data 
#'   will wind up on top of the chart and the last will be on bottom.
#' @param fillGraph Should the area underneath the graph be filled? This option 
#'   is not compatible with error bars. This option can also be set on a 
#'   per-series basis.
#' @param fillAlpha Transparency for filled regions of the plot. A value of 0.0 
#'   means that the fill will not be drawn, whereas a value of 1.0 means that 
#'   the fill will be as dark as the line of the series itself.
#' @param stepPlot When set, display the graph as a step plot instead of a line 
#'   plot. This option can also be set on a per-series basis.
#' @param stemPlot When set, display the graph as a stem plot instead of a line
#'   plot. This option can also be set on a per-series basis.
#' @param drawPoints Draw a small dot at each point, in addition to a line going
#'   through the point. This makes the individual data points easier to see, but
#'   can increase visual clutter in the chart. This option can also be set on a 
#'   per-series basis.
#' @param pointSize The size of the dot to draw on each point in pixels. A dot 
#'   is always drawn when a point is "isolated", i.e. there is a missing point 
#'   on either side of it. This also controls the size of those dots. This 
#'   option can also be set on a per-series basis.
#' @param pointShape The shape of the dot to draw. Can be one of the following:
#'   "dot" (default), "triangle", "square", "diamond", "pentagon", "hexagon",
#'   "circle", "star", "plus" or "ex". This option can also be set on a per-series
#'   basis.
#' @param drawGapEdgePoints Draw points at the edges of gaps in the data. This 
#'   improves visibility of small data segments or other data irregularities.
#' @param connectSeparatedPoints Usually, when dygraphs encounters a missing 
#'   value in a data series, it interprets this as a gap and draws it as such. 
#'   If, instead, the missing values represents an x-value for which only a 
#'   different series has data, then you'll want to connect the dots by setting 
#'   this to true.
#' @param strokeWidth The width of the lines connecting data points. This can be
#'   used to increase the contrast or some graphs. This option can also be set 
#'   on a per-series basis.
#' @param strokePattern A custom pattern array where the even index is a draw 
#'   and odd is a space in pixels. If null then it draws a solid line. The array
#'   should have a even length as any odd length array could be expressed as a 
#'   smaller even length array. This is used to create dashed lines. This option
#'   can also be set on a per-series basis.
#' @param strokeBorderWidth Draw a border around graph lines to make crossing 
#'   lines more easily distinguishable. Useful for graphs with many lines. This 
#'   option can also be set on a per-series basis.
#' @param strokeBorderColor Color for the line border used if 
#'   \code{strokeBorderWidth} is set. This option can also be set on a 
#'   per-series basis.
#' @param plotter A function (or array of functions) which plot each data series
#'   on the chart. May also be set on a per-series basis. See the 
#'   \href{http://dygraphs.com/tests/plotters.html}{dygraphs documentation} for 
#'   additional details on plotting functions.
#' @param colors Character vector of colors for the data series. These can be of
#'   the form "#AABBCC" or "rgb(255,100,200)" or "yellow", etc. If not 
#'   specified, equally-spaced points around a color wheel are used. This option
#'   can also be set on a per-series basis. Note that in both global and 
#'   per-series specification of custom colors you must provide a color for all 
#'   series being displayed. Note also that global and per-series color 
#'   specification cannot be mixed.
#' @param colorValue If custom colors are not specified, value of the data 
#'   series colors, as in hue/saturation/value (0.0-1.0, default 0.5).
#' @param colorSaturation If custom colors are not specified, saturation of the 
#'   automatically-generated data series colors (0.0-1.0, default 0.5).
#' @param drawXAxis Whether to draw the x-axis. Setting this to false also 
#'   prevents x-axis ticks from being drawn and reclaims the space for the chart
#'   grid/lines.
#' @param drawYAxis Whether to draw the y-axis. Setting this to false also 
#'   prevents y-axis ticks from being drawn and reclaims the space for the chart
#'   grid/lines.
#' @param includeZero Usually, dygraphs will use the range of the data plus some
#'   padding to set the range of the y-axis. If this option is set, the y-axis 
#'   will always include zero, typically as the lowest value. This can be used 
#'   to avoid exaggerating the variance in the data.
#' @param drawAxesAtZero When set, draw the X axis at the Y=0 position and the Y
#'   axis at the X=0 position if those positions are inside the graph's visible 
#'   area. Otherwise, draw the axes at the bottom or left graph edge as usual.
#' @param logscale When set the graph shows the y-axis in log scale. Any values 
#'   less than or equal to zero are not displayed.
#' @param axisTickSize The spacing between axis labels and tick marks.
#' @param axisLineColor Color of the x- and y-axis lines. Accepts any value 
#'   which the HTML canvas strokeStyle attribute understands, e.g. 'black' or 
#'   'rgb(0, 100, 255)'. This can also be set on a per-axis basis.
#' @param axisLineWidth Thickness (in pixels) of the x- and y-axis lines. This 
#'   can also be set on a per-axis basis.
#' @param axisLabelColor Color for x- and y-axis labels. This is a CSS color 
#'   string. This may also be set on a per-axis basis.
#' @param axisLabelFontSize Size of the font (in pixels) to use in the axis 
#'   labels, both x- and y-axis. This may also be set on a per-axis basis.
#' @param axisLabelWidth Width (in pixels) of the containing divs for x- and 
#'   y-axis labels.
#' @param drawGrid Whether to display grid lines in the chart. This may be set 
#'   on a per-axis basis to define the visibility of each axis' grid separately.
#'   Defaults to \code{TRUE} for x and y, and \code{FALSE} for y2.
#' @param gridLineColor The color of the grid lines. This option can also be set
#'   on a per-series basis.
#' @param gridLineWidth Thickness (in pixels) of the grid lines drawn under the 
#'   chart. This option can also be set on a per-series basis.
#' @param titleHeight Height of the chart title, in pixels. This also controls 
#'   the default font size of the title. If you style the title on your own, 
#'   this controls how much space is set aside above the chart for the title's 
#'   div.
#' @param rightGap Number of pixels to leave blank at the right edge of the 
#'   Dygraph. This makes it easier to highlight the right-most data point.
#' @param digitsAfterDecimal Unless it's run in scientific mode (see the 
#'   \code{sigFigs} option), dygraphs displays numbers with 
#'   \code{digitsAfterDecimal} digits after the decimal point. Trailing zeros 
#'   are not displayed, so with a value of 2 you'll get '0', '0.1', '0.12', 
#'   '123.45' but not '123.456' (it will be rounded to '123.46'). Numbers with 
#'   absolute value less than 0.1^digitsAfterDecimal (i.e. those which would 
#'   show up as '0.00') will be displayed in scientific notation.
#' @param labelsKMB Show K/M/B for thousands/millions/billions on y-axis.
#' @param labelsKMG2 Show k/M/G for kilo/Mega/Giga on y-axis. This is different 
#'   than \code{labelsKMB} in that it uses base 2, not 10.
#' @param labelsUTC Show date/time labels according to UTC (instead of local 
#'   time). Note that this option cannot is incompatible with
#'   \code{useDataTimezone} (you must use one or the other).
#' @param maxNumberWidth When displaying numbers in normal (not scientific) 
#'   mode, large numbers will be displayed with many trailing zeros (e.g. 
#'   100000000 instead of 1e9). This can lead to unwieldy y-axis labels. If 
#'   there are more than maxNumberWidth digits to the left of the decimal in a 
#'   number, dygraphs will switch to scientific notation, even when not 
#'   operating in scientific mode. If you'd like to see all those digits, set 
#'   this to something large, like 20 or 30.
#' @param sigFigs By default, dygraphs displays numbers with a fixed number of 
#'   digits after the decimal point. If you'd prefer to have a fixed number of 
#'   significant figures, set this option to that number of significant figures.
#'   A value of 2, for instance, would cause 1 to be display as 1.0 and 1234 to 
#'   be displayed as 1.23e+3.
#' @param panEdgeFraction A value representing the farthest a graph may be 
#'   panned, in percent of the display. For example, a value of 0.1 means that 
#'   the graph can only be panned 10% pased the edges of the displayed values. 
#'   null means no bounds.
#' @param animatedZooms Set this option to animate the transition between zoom 
#'   windows. Applies to programmatic and interactive zooms. Note that if you 
#'   also set a drawCallback, it will be called several times on each zoom. If 
#'   you set a zoomCallback, it will only be called after the animation is 
#'   complete.
#' @param mobileDisableYTouch Set this option to automatically disable touch
#'   events on the Y axis for mobile devices (since this interferes with 
#'   swiping/scrolling on mobile devices).
#' @param timingName Set this option to log timing information. The value of the
#'   option will be logged along with the timing, so that you can distinguish 
#'   multiple dygraphs on the same page.
#' @param useDataTimezone Whether to use the time zone of the underlying xts 
#'   object for display. Defaults to \code{FALSE} which uses the time zone of 
#'   the client workstation. Note that this option is incompatible with 
#'   \code{labelsUTC} (you must use one or other other).
#' @param retainDateWindow Whether to retain the user's current date window 
#'   (zoom level) when updating an existing dygraph with new data and/or 
#'   options.
#' @param disableZoom Set this option to disable click and drag zooming.
#'   
#' @return dygraph with additional options
#'   
#' @note See the \href{https://rstudio.github.io/dygraphs/}{online documentation}
#'   for additional details and examples.
#'   
#' @export
dyOptions <- function(dygraph,
                      stackedGraph = FALSE,
                      fillGraph = FALSE,
                      fillAlpha = 0.15,
                      stepPlot = FALSE,
                      stemPlot = FALSE,
                      drawPoints = FALSE,
                      pointSize = 1.0,
                      pointShape = c("dot",
                                     "triangle",
                                     "square",
                                     "diamond",
                                     "pentagon",
                                     "hexagon",
                                     "circle",
                                     "star",
                                     "plus",
                                     "ex"),
                      drawGapEdgePoints = FALSE,
                      connectSeparatedPoints = FALSE,
                      strokeWidth = 1.0,
                      strokePattern = NULL,
                      strokeBorderWidth = NULL,
                      strokeBorderColor = "white",
                      plotter = NULL,
                      colors = NULL,
                      colorValue = 0.5,
                      colorSaturation = 1.0,
                      drawXAxis = TRUE,
                      drawYAxis = TRUE,
                      includeZero = FALSE,
                      drawAxesAtZero = FALSE,
                      logscale = FALSE,
                      axisTickSize = 3.0,
                      axisLineColor = "black",
                      axisLineWidth = 0.3,
                      axisLabelColor = "black",
                      axisLabelFontSize = 14,
                      axisLabelWidth = 60,
                      drawGrid = TRUE,
                      gridLineColor = NULL,
                      gridLineWidth = 0.3,
                      titleHeight = NULL,
                      rightGap = 5,
                      digitsAfterDecimal = 2,
                      labelsKMB = FALSE,
                      labelsKMG2 = FALSE,
                      labelsUTC = FALSE,
                      maxNumberWidth = 6,
                      sigFigs = NULL,
                      panEdgeFraction = NULL,
                      animatedZooms = FALSE,
                      mobileDisableYTouch = TRUE,
                      timingName = NULL,
                      useDataTimezone = FALSE,
                      retainDateWindow = FALSE,
                      disableZoom = FALSE) {
  
  # validate that labelsUTC and useDataTimezone aren't specified together
  if (!missing(labelsUTC) && !missing(useDataTimezone))
    stop("You cannot specify both labelsUTC and useDateTimezone (choose one)")
  
  # Resolve stemPlot into a custom plotter if necessary
  plotter <- resolveStemPlot(stemPlot, plotter)
  
  options <- list()
  options$stackedGraph <- stackedGraph
  options$fillGraph <- fillGraph
  options$fillAlpha = fillAlpha
  options$stepPlot <- stepPlot
  options$drawPoints <- drawPoints
  options$pointSize <- pointSize
  options$drawGapEdgePoints <- drawGapEdgePoints
  options$connectSeparatedPoints <- connectSeparatedPoints
  options$strokeWidth <- strokeWidth
  options$strokePattern <- strokePattern
  options$strokeBorderWidth <- strokeBorderWidth
  options$strokeBorderColor <- strokeBorderColor
  options$plotter <- JS(plotter)
  if (!is.null(colors))
    options$colors <- as.list(colors)
  options$colorValue <- colorValue
  options$colorSaturation <- colorSaturation
  options$axes$x$drawAxis <- drawXAxis
  options$axes$y$drawAxis <- drawYAxis
  options$includeZero <- includeZero
  options$drawAxesAtZero <- drawAxesAtZero
  options$logscale <- logscale
  options$axisTickSize <- axisTickSize
  options$axisLineColor <- axisLineColor
  options$axisLineWidth <- axisLineWidth
  options$axisLabelColor <- axisLabelColor
  options$axisLabelFontSize <- axisLabelFontSize
  options$axisLabelWidth <- axisLabelWidth
  options$drawGrid <- drawGrid
  options$gridLineColor <- gridLineColor
  options$gridLineWidth <- gridLineWidth
  options$titleHeight <- titleHeight
  options$rightGap <- rightGap
  options$digitsAfterDecimal <- digitsAfterDecimal
  options$labelsKMB <- labelsKMB
  options$labelsKMG2 <- labelsKMG2
  options$labelsUTC <- labelsUTC
  options$maxNumberWidth <- maxNumberWidth
  options$sigFigs <- sigFigs
  options$panEdgeFraction <- panEdgeFraction
  options$animatedZooms <- animatedZooms
  options$mobileDisableYTouch <- mobileDisableYTouch
  options$timingName <- timingName
  if (!missing(retainDateWindow))
    options$retainDateWindow <- retainDateWindow
  options$disableZoom <- disableZoom
  
  # merge options into attrs
  dygraph$x$attrs <- mergeLists(dygraph$x$attrs, options)
   
  # use data timezone
  data.timezone <- attr(attr(dygraph$x, "time"),"tzone")
  if (is.null(data.timezone)) {
    data.timezone <- ""
  }
  dygraph$x$fixedtz <- useDataTimezone
  if (useDataTimezone && (data.timezone == "")) {
    warning("Can't use data time zone: no 'tzone' attribute in data")
    dygraph$x$fixedtz <- FALSE
  }
  dygraph$x$tzone <- data.timezone
  
  # set point shape
  pointShape <- match.arg(pointShape)
  if (pointShape != "dot") {
    dygraph$x$pointShape <- pointShape
  }

  # return modified dygraph
  dygraph
}

#' Employ a dygraph plotter on a series, a group of series, or the whole dygraph
#' 
#' Plotters provide variuos ways to customize how your data appears
#' on the dygraph.  Series-based plotters allow users to mix-and-match different plotters on a 
#' per-series or (with dyGroup) a per-group basis.  See \code{\link{dyPlotter}} for additional detail.
#'
#'
#' @section Available plotters:
#'
#' 
#' Currently the dygraphs package provides the following plotters:
#' 
#' \describe{
#'    \item{dyBarChart()}{Draws a bar plot rather than a line plot. If the provided dygraph
#' features more than one series, dyBarChart will call dyMultiColumn instead.}
#' 
#'    \item{dyStackedBarChart()}{Draws a bar chart stacking all the underlying series.}
#'
#'    \item{dyMultiColumn()}{Draws multiple column bar chart.} 
# 
#'    \item{dyBarSeries()}{Draws a single set of bars for just the provided series.}
#' 
#'    \item{dyStemSeries()}{Draws a single set of stems for just the provided series.}
#' 
#'    \item{dyShadow()}{An extraction of the _fillplotter from dygraph-combined-dev.js, 
#'    drawing the filled area without the line.}
#' 
#'    \item{dyFilledLIne()}{An extraction of the _fillplotter and _lineplotter combo 
#'    from dygraph-combined-dev.js. dyFilledLine allows users to fill only a single series.}
#' 
#'    \item{dyMultiColumnGroup()}{The multicolumn plotter, but on a subset of the series, leaving 
#'    the others for other plotters.}
#' 
#'    \item{dyCandlestick()}{Draw a candlestick chart.}
#' 
#'    \item{dyCandleStickGroup()}{Employed on the provided series, 
#'    but still plotting the others.}
#'    
#'    \item{dyStackerBarGroup()}{Return the data group as stacked bars}
#' 
#'    \item{dyStackerRibbonGroup()}{Return the data group as stacked ribbons}
#' 
#' }
#' 
#' 
#' @param dygraph Dygraph to add plotter to
#' @param name name - or chrarcter vector of names - of (the) series within the data set
#' @param compress (For dyCandlestick) If true, compress data yearly, quarterly, monthly, weekly or daily
#' according to overall amount of bars and/or current zoom level.
#' @param ... additional options to pass to dySeries
#'
#'
#' @return A dygraph with the specified plotter(s) employed.
#' 
#' @name Plotters
#' @examples 
#' ## The following two examples will results in the same dygraph:
#' 
#' dygraph(mdeaths) %>% 
#'   dyBarChart()
#'  
#' lungDeaths <- cbind(mdeaths, fdeaths)
#' dygraph(lungDeaths) %>%
#'   dyMultiColumn()
#'   
# lungDeaths <- cbind(mdeaths, fdeaths)
# dygraph(lungDeaths) %>%
#   dyStackedBarChart()
#'
#' ## Per-series plotters:
#' 
#' lungDeaths <- cbind(mdeaths, fdeaths)
#' dygraph(lungDeaths) %>%
#'   dyBarSeries('fdeaths')
#' 
#' lungDeaths <- cbind(mdeaths, fdeaths)
#' dygraph(lungDeaths) %>%
#'   dyStemSeries('fdeaths')
#'   
#' lungDeaths <- cbind(mdeaths, fdeaths)
#' dygraph(lungDeaths) %>%
#'   dyShadow('fdeaths')
#' 
#' lungDeaths <- cbind(mdeaths, fdeaths)
#' dygraph(lungDeaths) %>%
#'   dyFilledLine('fdeaths')
#'
#' ## A bunch of different plotters together:
#' 
#' lungDeaths <- cbind(fdeaths, mdeaths, ldeaths, foo = fdeaths/2, bar = fdeaths/3)
#' dygraph(lungDeaths) %>%
#'   dyRangeSelector() %>%
#'   dyBarSeries('bar') %>% 
#'   dyStemSeries('mdeaths') %>% 
#'   dyShadow('foo') %>% 
#'   dyFilledLine('fdeaths')
#'
#' ## Group-based plotters:
#'   
# lungDeaths <- cbind(mdeaths, fdeaths, ldeaths)
# dygraph(lungDeaths) %>%
#   dyMultiColumnGroup(c('fdeaths', 'mdeaths'))
#'
#' ## Candlestick plotters:
#'   
#' library(xts)
#' data(sample_matrix)
#' library(dygraphs)
#' dygraph(sample_matrix) %>%
#'   dyCandlestick()
#'   
#' sample<-data.frame(sample_matrix)
#' sample_2<-sample*2
#' names(sample_2)<-c('O', 'H', 'L', 'C')
#' sample<-cbind(sample, sample_2)
#' dygraph(sample) %>% 
#'   dyOptions(stackedGraph = TRUE) %>% 
#'   dyCandlestickGroup(c('Open', 'High', 'Low', 'Close')) %>% 
#'   dyCandlestickGroup(c('O', 'H', 'L', 'C'))
#'
#' ## Stacked Bar and Ribbon Graphs:  
#' 
#' dygraph(lungDeaths) %>% 
#'   dySeries('mdeaths', axis = 'y2') %>%
#'   dyAxis('y', valueRange = c(-100, 1000)) %>% 
#'   dyStackedBarGroup(c('ldeaths', 'fdeaths'))
#'   
#' lungDeaths <- cbind(ldeaths, fdeaths, mdeaths, 
#'                     additive = rep.int(200, length(ldeaths)),
#'                     line = rep.int(3000, length(ldeaths)))
#' dygraph(lungDeaths) %>% 
#'   dySeries('line', strokePattern = 'dashed') %>% 
#'   dySeries('ldeaths', stepPlot = TRUE) %>% 
#'   dyStackedBarGroup(c('additive', 'mdeaths')) %>% 
#'   dyStackedRibbonGroup(c('fdeaths', 'ldeaths'))
NULL

#' @rdname Plotters
#' @export
dyBarChart <- function(dygraph) {
  if (length(dygraph$x$data)>2) {
    dygraph <- dyPlotter(dygraph = dygraph,
              name = "MultiColumn",
              path = system.file("plotters/multicolumn.js", package = "dygraphs"))
    
  } else {
    dygraph <- dyPlotter(dygraph = dygraph,
              name = "BarChart",
              path = system.file("plotters/barchart.js", package = "dygraphs"))
  }
  
  return (dygraph)
}


#' @rdname Plotters
#' @export
dyStackedBarChart <- function(dygraph) {
  dygraph <- dyPlotter(dygraph = dygraph,
              name = "StackedBarChart",
              path = system.file("plotters/stackedbarchart.js", package = "dygraphs"))
  
  names<-dygraph$x$attrs$labels[-1]
  
  dygraph <- computeYaxisRange(dygraph, names)
  dygraph 
  
}


#' @rdname Plotters
#' @export
dyMultiColumn <- function(dygraph) {
  dyPlotter(dygraph = dygraph,
            name = "MultiColumn",
            path = system.file("plotters/multicolumn.js", package = "dygraphs"))
}


#' @rdname Plotters
#' @export
dyBarSeries <- function(dygraph, name, ...) {
  file <- system.file("plotters/barseries.js", package = "dygraphs")
  plotter_ <- paste0(readLines(file, skipNul = T), collapse = "\n")
  
  dots <- list(...)
  do.call('dySeries', c(list(dygraph = dygraph, name = name, plotter = plotter_), dots))
  
}


#' @rdname Plotters
#' @export
dyStemSeries <- function(dygraph, name, ...) {
  file <- system.file("plotters/stemplot.js", package = "dygraphs")
  plotter_ <- paste0(readLines(file, skipNul = T), collapse = "\n")
  
  dots <- list(...)
  do.call('dySeries', c(list(dygraph = dygraph, name = name, plotter = plotter_), dots))
}


#' @rdname Plotters
#' @export
dyShadow <- function(dygraph, name, ...) {
  file <- system.file("plotters/fillplotter.js", package = "dygraphs")
  plotter_ <- paste0(readLines(file, skipNul = T), collapse = "\n")
  
  dots <- list(...)
  do.call('dySeries', c(list(dygraph = dygraph, name = name, plotter = plotter_), dots))
}


#' @rdname Plotters
#' @export
dyFilledLine <- function(dygraph, name, ...) {
  file <- system.file("plotters/filledline.js", package = "dygraphs")
  plotter_ <- paste0(readLines(file, skipNul = T), collapse = "\n")
  
  dots <- list(...)
  do.call('dySeries', c(list(dygraph = dygraph, name = name, plotter = plotter_), dots))
}


#' @rdname Plotters
#' @export
dyErrorFill <- function(dygraph, name, ...) {
  file <- system.file("plotters/errorplotter.js", package = "dygraphs")
  plotter_ <- paste0(readLines(file, skipNul = T), collapse = "\n")
  
  dots <- list(...)
  do.call('dySeries', c(list(dygraph = dygraph, name = name, plotter = plotter_), dots))
}


#' @rdname Plotters
#' @export
dyMultiColumnGroup <- function(dygraph, name, ...) {
  file <- system.file("plotters/multicolumngroup.js", package = "dygraphs")
  plotter_ <- paste0(readLines(file, skipNul = T), collapse = "\n")
  
  dots <- list(...)
  do.call('dyGroup', c(list(dygraph = dygraph, name = name, plotter = plotter_), dots))
}


#' @rdname Plotters
#' @export
dyCandlestickGroup <- function(dygraph, name, ...) {
  file <- system.file("plotters/candlestickgroup.js", package = "dygraphs")
  plotter_ <- paste0(readLines(file, skipNul = T), collapse = "\n")
  
  dots <- list(...)
  do.call('dyGroup', c(list(dygraph = dygraph, name = name, plotter = plotter_), dots))
}


#' @rdname Plotters
#' @export
dyStackedBarGroup <- function(dygraph, name, ...) {
  dots <- list(...)
  
  if(length(name) < 2) {
    dygraph <- do.call('dyBarSeries', c(list(dygraph = dygraph, name = unlist(name)), dots))
    return(dygraph)
  }
  
  file <- system.file("plotters/stackedbargroup.js", package = "dygraphs")
  plotter_ <- paste0(readLines(file, skipNul = T), collapse = "\n")
  
  dygraph <- do.call('dyGroup', c(list(dygraph = dygraph, name = name, plotter = plotter_), dots))

  dygraph <- computeYaxisRange(dygraph, name)
  dygraph 
}

#' @rdname Plotters
#' @export
dyStackedLineGroup <- function(dygraph, name, ...) {
  dots <- list(...)
  
  if(length(name) < 2) {
    # just rely on the base package
    return(dygraph)
  }
  
  file <- system.file("plotters/stackedlinegroup.js", package = "dygraphs")
  plotter_ <- paste0(readLines(file, skipNul = T), collapse = "\n")
  
  dygraph <- do.call('dyGroup', c(list(dygraph = dygraph, name = name, plotter = plotter_), dots))

  dygraph <- computeYaxisRange(dygraph, name)
  dygraph 
}
 
 
#' @rdname Plotters
#' @export
dyStackedRibbonGroup <- function(dygraph, name, ...) {
  dots <- list(...)
  
  if(length(name) < 2) {
    dygraph <- do.call('dyFilledLine', c(list(dygraph = dygraph, name = name), dots))
    return(dygraph)
  }
  
  file <- system.file("plotters/stackedribbongroup.js", package = "dygraphs")
  plotter_ <- paste0(readLines(file, skipNul = T), collapse = "\n")
  
  dygraph <- do.call('dyGroup', c(list(dygraph = dygraph, name = name, plotter = plotter_), dots))

  dygraph <- computeYaxisRange(dygraph, name)
  dygraph 
}

computeYaxisRange <- function(dygraph, name){ 
  # most of what happens from here on out is a simplified version of the 
  # stackPoints and computeYaxis functions in the underlying dygraph package.
  # Since we can't modify the Yaxis range from within the specialized plotter,
  # we need to calculate an appropriate valueRange for the group's axis here, then
  # reconcile that against user-provided ranges and then pass into the widget... this
  # way we ensure that the newly stacked data (but initially unstacked to the widget) 
  # doesn't get cutoff by an axis range - computed by the widget - that won't consider
  # the points stacked
  attrs <- dygraph$x$attrs

  data <- attr(dygraph$x, "data")
  cols <- which(names(data) %in% name)
  
  #get all the series, minus the x axis, that are not part of the group
  data_ <- data[-c(1, cols)]
  name_ <- names(data)[which(!names(data) %in% name)][-1]
  
  #grab the axis for the group, we'll calculate the new range only on this axis  
  series <- attrs$series[[name[1]]]
  axisNm <- series$axis
  if (!is.null(axisNm)) {
    axis <- attrs$axes[[axisNm]]
    valueRange <- axis$valueRange
  } else {
    axisNm<-"y"
    axis <- attrs$axes[[axisNm]] <- NULL
    valueRange <- NULL
  }
  
  if(is.null(valueRange)) {
    valueRange <- c(0, 0)
  }
 
  # get the group data fields
  data <- data[cols]
 
  for(i in 1:length(data)) {
      # get the data points
      points <- data[i][[1]]
      # fill NAs... we're not saving this data back into the graph, so this is OK
      is.na(points) <- 0
      
      # add to cumulativeList
      if(!exists('cumulativeYval')) cumulativeYval <- points
      else cumulativeYval <- cumulativeYval + points
      # calculate extremes
      extremes <- range(cumulativeYval)
  }
  
  for(i in 1:length(data_)) {
    if (length(data_)==0) break
    
    # ranges are calcuated separately, so skip those from other axes
    series_ <- attrs$series[[name_[i]]]
    if (!is.null(series_$axis) && series_$axis != axisNm) next
    
    points <- data_[i][[1]]
  
    # fill NAs
    is.na(points) <- 0
    
    # getExtremes
    extremes_ <- range(points)
    extremes[1] <- min(extremes[1], extremes_[1])  
    extremes[2] <- max(extremes[2], extremes_[2])  
  }
  
  valueRange[1] <- min(extremes[1], valueRange[1])
  valueRange[2] <- max(extremes[2], valueRange[2])
    
  # add a little padding since we're hard-setting the range
  valueRange[2] <- valueRange[2] + 0.05 * abs(valueRange[2] - valueRange[1])
  
  axis$options$valueRange <- valueRange
  attrs$axes[[axisNm]] <- axis$options  
  
  # return modified dygraph
  dygraph$x$attrs <- mergeLists(dygraph$x$attrs, attrs)
  return (dygraph)
}

#' Add external assets as a dygraph dependency
#'
#' @param dygraph Dygraph to add dependency to
#' @param dependency An HTML dependency
#'
#' @return A dygraph with the specified dependency added.
#'
#' @export
dyDependency <- function(dygraph, dependency) {
  if (is.null(dygraph$dependencies)) {
    dygraph$dependencies <- list()
  }
  dygraph$dependencies[[length(dygraph$dependencies) + 1]] <- dependency

  # return dygraph
  dygraph
}

#' Include a dygraph plugin
#'
#' @param dygraph Dygraph to add plugin to
#' @param name Name of plugin
#' @param path Path to plugin JavaScript file
#' @param options Named list of options to pass to plugin constructor
#' @param version Plugin version (e.g. version of package which provides the plugin)
#'
#' @importFrom htmltools htmlDependency
#' @return A dygraph with the specified plugin enabled.
#'
#' @details
#' You can use dygraphs plugins to customize the appearance of dygraphs as well
#' as add new interactive behaviors. For additional information on creating
#' dygraphs plugins see the
#' \href{https://github.com/danvk/dygraphs/tree/master/src/plugins}{dygraphs plugins}
#' documentation.
#
#' Once you've created a dygraphs plugins you can use the dyPlugin function to
#' create an R wrapper for it. See
#' \href{https://rstudio.github.io/dygraphs/gallery-plugins.html}{https://rstudio.github.io/dygraphs/gallery-plugins.html}
#' for details on how to do this.
#'
#' @export
dyPlugin <- function(dygraph, name, path, options = list(), version = "1.0") {
  path <- normalizePath(path)
  pluginDependency <- htmlDependency(paste0("Dygraph.Plugins.", name),
                                     version,
                                     src = dirname(path),
                                     script = basename(path),
                                     all_files = FALSE)
  dygraph <- dyDependency(dygraph, pluginDependency)

  # add the plugin and it's options (will be evaluated by renderValue)
  if (is.null(dygraph$x$plugins)) {
    dygraph$x$plugins <- list()
  }
  if (length(options) == 0) {
    options <- JS("{}")
  }
  # merge options if plugin was already added
  if (name %in% names(dygraph$x$plugins)) {
    current.options <- dygraph$x$plugins[[name]]
    i <- match(names(current.options), names(options))
    i <- is.na(i)
    if (any(i)) {
      options[names(current.options)[which(i)]] = current.options[which(i)]
    }
  }
  dygraph$x$plugins[[name]] <- options

  # return dygraph
  dygraph
}

#' Include a dygraph plotter
#'
#' @param dygraph Dygraph to add plotter to
#' @param name Name of plotter
#' @param path Path to plotter JavaScript file
#' @param version Plotter version (e.g. version of package which provides the plotter)
#'
#' @importFrom htmltools htmlDependency
#' @return A dygraph with the specified plotter enabled.
#'
#' @export
dyPlotter <- function(dygraph, name, path, version = "1.0") {
  path <- normalizePath(path)
  plotterDependency <- htmlDependency(paste0("Dygraph.Plotters.", name),
                                      version,
                                      src = dirname(path),
                                      script = basename(path),
                                      all_files = FALSE)
  dygraph <- dyDependency(dygraph, plotterDependency)

  dygraph$x$plotter <- name

  # return dygraph
  dygraph
}

#' Include a dygraph data handler
#'
#' @param dygraph Dygraph to add data handler to
#' @param name Name of data handler
#' @param path Path to data handler JavaScript file
#' @param version Data handler version (e.g. version of package which provides the
#' data handler)
#'
#' @importFrom htmltools htmlDependency
#' @return A dygraph with the specified data handler enabled.
#'
#' @export
dyDataHandler <- function(dygraph, name, path, version = "1.0") {
  path <- normalizePath(path)
  dataHandlerDependency <- htmlDependency(paste0("Dygraph.DataHandlers.", name),
                                          version,
                                          src = dirname(path),
                                          script = basename(path),
                                          all_files = FALSE)
  dygraph <- dyDependency(dygraph, dataHandlerDependency)

  dygraph$x$dataHandler <- name

  # return dygraph
  dygraph
}

#' dygraph interactive range selection and zooming
#' 
#' Add a range selector to the bottom of the chart that allows users to pan and 
#' zoom to various date ranges.
#' 
#' @inheritParams dyOptions
#' 
#' @param dygraph Dygraph to add range selector to
#' @param dateWindow Initially zoom in on a section of the graph. Is a two 
#'   element vector [earliest, latest], where earliest/latest objects 
#'   convertible via \code{as.POSIXct}.
#' @param height Height, in pixels, of the range selector widget.
#' @param fillColor The range selector mini plot fill color. This can be of the 
#'   form "#AABBCC" or "rgb(255,100,200)" or "yellow". You can also specify ""
#'   to turn off fill.
#' @param strokeColor The range selector mini plot stroke color. This can be of 
#'   the form "#AABBCC" or "rgb(255,100,200)" or "yellow". You can also specify 
#'   "" to turn off stroke.
#' @param keepMouseZoom Keep mouse zoom when adding a range selector
#' 
#' @return A dygraph that displays a range selector
#'   
#' @examples
#' library(dygraphs)
#' 
#' dygraph(nhtemp, main = "New Haven Temperatures") %>% 
#'   dyRangeSelector()   
#'  
#' dygraph(nhtemp, main = "New Haven Temperatures") %>% 
#'   dyRangeSelector(dateWindow = c("1920-01-01", "1960-01-01"))   
#'     
#' dygraph(nhtemp, main = "New Haven Temperatures") %>% 
#'   dyRangeSelector(height = 20, strokeColor = "")     
#'
#' @note See the
#' \href{https://rstudio.github.io/dygraphs/gallery-range-selector.html}{online
#' documentation} for additional details and examples.
#' 
#' Shiny applications can respond to changes in the \code{dateWindow}
#' via a special date window shiny input value. For example, if the 
#' output id of a dygraph is `series` then the current date window
#' can be read from \code{input$series_date_window} as an array of 
#' two date values (from and to).
#' 
#' @export
dyRangeSelector <- function(dygraph,
                            dateWindow = NULL, 
                            height = 40,
                            fillColor = " #A7B1C4",
                            strokeColor = "#808FAB", 
                            keepMouseZoom = TRUE,
                            retainDateWindow = FALSE) {
  
  selector <- list()
  selector$showRangeSelector = TRUE
  if (!is.null(dateWindow)) {
    if (length(dateWindow) != 2)
      stop("dateWindow must be vector of length 2 that is convertible to POSIXct")
    if (dygraph$x$format == "date")
      selector$dateWindow <- sapply(USE.NAMES = FALSE, dateWindow, asISO8601Time)
    else
      selector$dateWindow <- dateWindow
  }
  selector$rangeSelectorHeight <- height
  selector$rangeSelectorPlotFillColor <- fillColor
  selector$rangeSelectorPlotStrokeColor <- strokeColor
  if (!missing(retainDateWindow))
    selector$retainDateWindow <- retainDateWindow
  
  if(keepMouseZoom)
    selector$interactionModel= JS("Dygraph.Interaction.defaultModel")
  
  # merge selector
  dygraph$x$attrs <- mergeLists(dygraph$x$attrs, selector)
  
  # return modified dygraph
  dygraph
}

context("dyRangeSelector")

test_that("range selector creation", {
  d <- dygraph(nhtemp, main = "New Haven Temperatures") %>% 
         dyRangeSelector()   
  expect_identical(d$x$attrs$showRangeSelector, TRUE)
})

#' dygraph series group
#' 
#' Add a data series group to a dygraph plot. Note that options will use the default 
#' global setting (as determined by \code{\link{dyOptions}}) when not specified 
#' explicitly. Importantly, any dySeries options passed can be passed as a vector of values 
#' and will be replicated across all series named as part of the group. If arguments differ in
#' length than the number of series named, then the argument vector will be 
#' cycled across the named series.
#' 
#' NOTE: dyGroup will turn off \code{stackedGraph}, as the option will calculated cumulatives using
#' all series in the underlying dygraph, not just a subset.
#' 
#' The dyGroup function can also replicated similar arguments across multiple series, or 
#' be used to apply a grouped custom plotter - i.e., multi-column plotter - to a subset of the
#' dygraph data series.
#' 
#' @inheritParams dySeries
#' @inheritParams dyOptions
#'   
#' @param dygraph Dygraph to add a series definition to
#' @param name character vector of the series within data set. If no name is specified then 
#'   series are bound to implicitly based on their order within the underlying 
#'   time series object. This parameter can also be a character vector of length
#'   3 that specifies a set of input column names to use as the lower, value,
#'   and upper for a series with a shaded bar drawn around it.
#' @param label Labels to display for series (uses name if no label defined)
#' @param color Color for series. These can be of the form "#AABBCC" or 
#'   "rgb(255,100,200)" or "yellow", etc. If not specified then the global 
#'   colors option (typically based on equally-spaced points around a color wheel). 
#' @param axis Y-axis to associate the series with ("y" or "y2")
#' @param stepPlot When set, display the graph as a step plot instead of a line 
#'   plot.
#' @param stemPlot When set, display the graph as a stem plot instead of a line
#'   plot.
#' @param fillGraph Should the area underneath the graph be filled? This option 
#'   is not compatible with error bars.
#' @param drawPoints Draw a small dot at each point, in addition to a line going
#'   through the point. This makes the individual data points easier to see, but
#'   can increase visual clutter in the chart.
#' @param pointSize The size of the dot to draw on each point in pixels. A dot 
#'   is always drawn when a point is "isolated", i.e. there is a missing point 
#'   on either side of it. This also controls the size of those dots.
#' @param pointShape The shape of the dot to draw. Can be one of the following:
#'   "dot" (default), "triangle", "square", "diamond", "pentagon", "hexagon",
#'   "circle", "star", "plus" or "ex".
#' @param strokeWidth The width of the lines connecting data points. This can be
#'   used to increase the contrast or some graphs.
#' @param strokePattern A predefined stroke pattern type ("dotted", "dashed", or
#'   "dotdash") or a custom pattern array where the even index is a draw and odd
#'   is a space in pixels. If \code{NULL} then it draws a solid line. The array 
#'   should have an even length as any odd length array could be expressed as 
#'   a smaller even length array.
#' @param strokeBorderWidth Draw a border around graph lines to make crossing 
#'   lines more easily distinguishable. Useful for graphs with many lines.
#' @param strokeBorderColor Color for the line border used if 
#'   \code{strokeBorderWidth} is set.
#' 
#' @param plotter A function which plots the series group. See the 
#'   \href{http://dygraphs.com/tests/plotters.html}{dygraphs documentation} for 
#'   additional details on plotting functions.
#'   
#' @return Dygraph with additional series
#'   
#' @examples
#' \dontrun{
#' library(dygraphs)
#' 
#' lungDeaths <- cbind(ldeaths, mdeaths, fdeaths)
#' 
#' dygraph(lungDeaths, main = "Deaths from Lung Disease (UK)") %>%
#'   dySeries("fdeaths", stepPlot = TRUE, color = "red") %>% 
#'   dyGroup(c("mdeaths", "ldeaths"), drawPoints = TRUE, color = c("blue", "green"))
#' }   
#'
#' @note See the 
#'   \href{https://rstudio.github.io/dygraphs/gallery-series-options.html}{online
#'   documentation} for additional details and examples.
#'   
#' @export
dyGroup <- function(dygraph,
                     name = NULL, 
                     label = NULL,
                     color = NULL,
                     axis = "y", 
                     stepPlot = NULL,
                     stemPlot = NULL,
                     fillGraph = NULL,
                     drawPoints = NULL,
                     pointSize = NULL,
                     pointShape = NULL,
                     strokeWidth = NULL,
                     strokePattern = NULL,
                     strokeBorderWidth = NULL,
                     strokeBorderColor = NULL,
                     plotter = NULL) {
  
  
  # get a reference to the underlying data and labels
  data <- attr(dygraph$x, "data")
  labels <- names(data)

  if (length(plotter) > 1) message('dyGroup: pass only a single plotter option')
  
  
  # auto-bind name if necessary
  autobind <- attr(dygraph$x, "autoSeries")
  if (length(name) == 1) {
    dygraph<-dySeries(dygraph = dygraph, 
                      name = name, 
                      label = label, 
                      color = color, 
                      plotter = plotter)
    return(dygraph)
  }
  
  # when setting up the first color, we start handling colors here 
  if (!is.null(color) && is.null(dygraph$x$attrs$colors)) {
      colors <- dygraphColors(dygraph, length(labels) - 1)
      dygraph$x$attrs$colors <- colors
  }

  # prepare the colors list for processing 
  if (!is.null(dygraph$x$attrs$colors)) {
     colors <- dygraph$x$attrs$colors
     names(colors) <- dygraph$x$attrs$labels[-1]
  }
    
  # Plotter-mod!  Added the plotter != NULL test to keep base capability while
  # expanding to include group plotters
  # Get the cols where this series is located and verify that they are
  # available within the underlying dataset
  cols <- which(labels %in% name)
  if (length(cols) != length(name)) {
    stop("One or more of the specified series were not found. ",
         "Valid series names are: ", paste(labels[-1], collapse = ", "))
  }
  
  # Data series named here are "consumed" from the automatically generated
  # list of series (they'll be added back in below)
  cols <- which(dygraph$x$attrs$labels %in% name)
  dygraph$x$data <- dygraph$x$data[-c(cols)]
  dygraph$x$attrs$labels <- dygraph$x$attrs$labels[-c(cols)]

   
  # MUST turn off native stacking option, as underlying dygraph 
  # will include custom-plotted points in the stacked calculation
  if (length(dygraph$x$attrs$stackedGraph) > 0) {
    if (dygraph$x$attrs$stackedGraph) 
						warning("dyGroup is incompatible with stackedGraph... stackedGraph now FALSE")

    dygraph$x$attrs$stackedGraph <- FALSE;
  }
   
  # Resolve stemPlot into a custom plotter if necessary
  plotter <- resolveStemPlot(stemPlot, plotter)
  
  if (!is.null(pointShape))
    dygraph$x$pointShape <- list()
   
  l <- length(name)
  
  # copy attrs for modification
  attrs <- dygraph$x$attrs
  
  # repeat (most of) the steps from dySeries, just in a loop 
  for (i in 1:l) {
    
    # create series object
    series <- list()
    series$name <- name[i]
    
    # take the passed options and extend to the length of the name vector; it's
    # up to the User to make sure the vectors are of the desired length
    suppressWarnings({
      # for the axis, however, we enforce the same axis across all series named
      # in the group.  We can't stop the user from changing the axis of one or more
      # series later, but at least we can control for some mistakes here
      series$options$axis <- rep(match.arg(axis, c("y", "y2")), 
                      			 			length.out = l)[1]
      if (!is.null(stepPlot)) 
				series$options$stepPlot <- rep(stepPlot, length.out = l)[i]

      if (!is.null(fillGraph)) 
				series$options$fillGraph <- rep(fillGraph, length.out = l)[i]
      
			if (!is.null(drawPoints)) 
				series$options$drawPoints <- rep(drawPoints, length.out = l)[i]

      if (!is.null(pointSize)) 
				series$options$pointSize <- rep(pointSize, length.out = l)[i]
      
			if (!is.null(strokeWidth)) 
				series$options$strokeWidth <- rep(strokeWidth, length.out = l)[i]
      
			if (!is.null(strokePattern)) 
				series$options$strokePattern <- rep(resolveStrokePattern(strokePattern), length.out = l)[i]
      
			if (!is.null(strokeBorderWidth)) 
				series$options$strokeBorderWidth <- rep(strokeBorderWidth, length.out = l)[i]
      
			if (!is.null(strokeBorderColor)) 
				series$options$strokeBorderColor <- rep(strokeBorderColor, length.out = l)[i]
     
      # one can use this to pass a group plotter or any combination of individual series plotters 
      series$options$plotter <- JS(plotter)
    })
    
    # KEY!  Adding a group designator to aid in group plotters
    # By concatenating the names provided in the name age, it becomes
    # a unique identifier than won't be duplicated unless the entire group of names
    # passed gets re-passed together a second time, which would obviously override
    # the first set of options
    series$options$group <- paste0(name, collapse = "")
  
    seriesData <- data[[series$name]]
    
    # grab the colors for the series being processed
  if (!is.null(dygraph$x$attrs$colors)) {
      currColors <- colors[names(colors) %in% name[i]]
      
			if (!is.null(color)) 
				currColors[[series$name]] <- color[i]
      
      colors <- colors[!names(colors) %in% name[i]]
      colors[[series$name]] <- currColors[[series$name]]

    # compensating for the bug whereas a single series dygraph with specified color
    # shows up as black since the DateTime series tries to take the first color
    	if (length(colors) == 1) colors <- c(colors, colors)
      
      attrs$colors <- colors
      names(attrs$colors) <- NULL
    }
    
    # default the label if we need to
    if (is.null(series$label))
      series$label <- series$name  
    
    # add label
    attrs$labels <- c(attrs$labels, series$label)
    
    # get whatever options might have previously existed for the series, then merge
    base <- attrs$series[[series$label]]
	 	series$options <- mergeLists(base, series$options)
    
    # set options
    attrs$series[[series$label]] <- series$options
    
    # set attrs
    dygraph$x$attrs <- attrs
    
    # set point shape
    if (!is.null(pointShape[i])) {
      shapes <- c("dot", "triangle", "square", "diamond", "pentagon",
                  "hexagon", "circle", "star", "plus", "ex")

      if (!is.element(pointShape[i], shapes)) {
        stop("Invalid value for pointShape parameter. ",
             "Should be one of the following: ",
             "'dot', 'triangle', 'square', 'diamond', 'pentagon', ",
             "'hexagon', 'circle', 'star', 'plus' or 'ex'")
      }
  
      if (pointShape[i] != "dot")
        dygraph$x$pointShape[[series$label]] <- rep(pointShape, length.out = l)
      
    }
    
    # add data
    dygraph$x$data[[length(dygraph$x$data) + 1]] <- seriesData
  }  
  
  # return modified dygraph
  dygraph
}

#' The dyUnzoom plugin adds an "Unzoom" button to the graph when it's displaying
#' in a zoomed state (this is a bit more discoverable than the default double-
#' click gesture for unzooming).
#'
#' @param dygraph Dygraph to add plugin to
#'
#' @return Dygraph with Unzoom plugin enabled
#'
#' @examples
#' library(dygraphs)
#' dygraph(ldeaths) %>%
#'   dyRangeSelector() %>%
#'   dyUnzoom()
#'
#' @export
dyUnzoom <-function(dygraph) {
  dyPlugin(
    dygraph = dygraph,
    name = "Unzoom",
    path = system.file("plugins/unzoom.js", package = "dygraphs")
  )
}

#' The dyCrosshair plugin draws a crosshair line over the point closest to the
#' mouse when the user hovers over the graph. It has a "direction" option which
#' is provided in the R wrapper function and then forwarded to the plugin using
#' the "options" argument to dyPlugin.
#'
#' @param dygraph Dygraph to add plugin to
#' @param direction Crosshair direction. Valid options are: "both", "horizontal",
#' "vertical"
#'
#' @return Dygraph with Crosshair plugin enabled
#'
#' @examples
#' library(dygraphs)
#' dygraph(ldeaths) %>%
#'   dyRangeSelector() %>%
#'   dyCrosshair(direction = "vertical")
#'
#' @export
dyCrosshair <- function(dygraph, direction = c("both", "horizontal", "vertical")) {
  dyPlugin(
    dygraph = dygraph,
    name = "Crosshair",
    path = system.file("plugins/crosshair.js", package = "dygraphs"),
    options = list(direction = match.arg(direction))
  )
}

#' dyRibbon plugin adds a horizontal band of colors that runs through the chart. It
#' can be useful to visualize categorical variables
#' (http://en.wikipedia.org/wiki/Categorical_variable) that change over time (along
#' the x-axis).
#'
#' @param dygraph Dygraph to add plugin to
#' @param data Vector of numeric values in the range from 0 to 1
#' @param palette Vector with colors palette
#' @param parser JavaScrip function (function (data, dygraph)) returning the array of
#' numeric values. Parser is used if no data was provided
#' @param top Vertical position of the top edge of ribbon relative to chart height.
#' @param bottom Vertical position of the bottom edge of ribbon relative to chart height.
#'
#' @examples
#' \dontrun{
#' library(quantmod)
#'
#' getSymbols("SPY", from = "2016-12-01", auto.assign=TRUE)
#'
#' difference <- SPY[, "SPY.Open"] - SPY[, "SPY.Close"]
#' decreasing <- which(difference < 0)
#' increasing <- which(difference > 0)
#'
#' dyData <- SPY[, "SPY.Close"]
#'
#' ribbonData <- rep(0, nrow(dyData))
#' ribbonData[decreasing] <- 0.5
#' ribbonData[increasing] <- 1
#'
#' dygraph(dyData) %>%
#'     dyRibbon(data = ribbonData, top = 0.1, bottom = 0.02)
#' }
#' 
#' @export
dyRibbon <- function(dygraph, data = NULL, palette = NULL, parser = NULL, top = 1, bottom = 0) {
  dyPlugin(
    dygraph = dygraph,
    name = "Ribbon",
    path = system.file("plugins/ribbon.js", package = "dygraphs"),
    options = list(data = data,
                   parser = JS(parser),
                   options = list(palette = palette,
                                  top = top,
                                  bottom = bottom))
  )
}

mergeLists <- function (base_list, overlay_list, recursive = TRUE) {
  if (length(base_list) == 0)
    overlay_list
  else if (length(overlay_list) == 0)
    base_list
  else {
    merged_list <- base_list
    for (name in names(overlay_list)) {
      base <- base_list[[name]]
      overlay <- overlay_list[[name]]
      if (is.list(base) && is.list(overlay) && recursive)
        merged_list[[name]] <- mergeLists(base, overlay)
      else {
        merged_list[[name]] <- NULL
        merged_list <- append(merged_list,
                              overlay_list[which(names(overlay_list) %in% name)])
      }
    }
    merged_list
  }
}

asISO8601Time <- function(x) {
  if (!inherits(x, "POSIXct"))
    x <- as.POSIXct(x, tz = "GMT")
  format(x, format = "%04Y-%m-%dT%H:%M:%OS3Z", tz = 'GMT')
}

resolveStrokePattern <- function(strokePattern) {
  if (is.character(strokePattern)) {
    if (strokePattern == "dotted") 
      strokePattern <- c(2, 2)
    else if (strokePattern == "dashed")
      strokePattern <- c(7, 3)
    else if (strokePattern == "dotdash")
      strokePattern <- c(7, 2, 2, 2)
    else if (strokePattern == "solid")
      strokePattern <- c(1, 0)
    else
      stop("Invalid stroke pattern: valid values are dotted, ",
           "dashed, and dotdash")
  }
  strokePattern
}

#' @importFrom stats end start
defaultPeriodicity <- function (data) {
  periodicity <- structure(
    list(difftime = structure(0, units = "secs", class = "difftime"),
         frequency = 0,
         start = start(data),
         end = end(data),
         units = "secs",
         scale = "seconds",
         label = "second"),
    class = "periodicity")
}

#' dygraph axis
#' 
#' Define options for an axis on a dygraph plot. Note that options will use the 
#' default global setting (as determined by \code{\link{dyOptions}}) when not 
#' specified explicitly.
#' 
#' @inheritParams dyOptions
#'   
#' @param dygraph Dygraph to add an axis definition to
#' @param name Axis name ('x', 'y', or 'y2')
#' @param label Label to display for axis (defaults to none).
#' @param valueRange Explicitly set the vertical range of the graph to 
#'   \code{c(low, high)}. This may be set on a per-axis basis to define each 
#'   y-axis separately. If either limit is unspecified, it will be calculated 
#'   automatically (e.g. \code{c(NULL, 30)} to automatically calculate just the 
#'   lower bound).
#' @param logscale When set for the y-axis or x-axis, the graph shows that axis
#'   in log scale. Any values less than or equal to zero are not displayed.
#'   Showing log scale with ranges that go below zero will result in an
#'   unviewable graph. Not compatible with showZero. connectSeparatedPoints is
#'   ignored. This is ignored for date-based x-axes.
#' @param ticker This lets you specify an arbitrary JavaScript function to 
#'   generate tick marks on an axis. The tick marks are an array of (value, 
#'   label) pairs. The built-in functions go to great lengths to choose good 
#'   tick marks so, if you set this option, you'll most likely want to call one 
#'   of them and modify the result. See dygraph-tickers.js and the 
#'   \href{http://dygraphs.com/options.html}{dygraphs documentation} for 
#'   additional details).
#' @param rangePad Add the specified amount of extra space (in pixels) around 
#'   the value range to ensure points at the edges remain visible.
#' @param labelWidth Width of the div which contains the y-axis label. Since the
#'   y-axis label appears rotated 90 degrees, this actually affects the height 
#'   of its div.
#' @param labelHeight Height of the x-axis label, in pixels. This also controls 
#'   the default font size of the x-axis label. If you style the label on your 
#'   own, this controls how much space is set aside below the chart for the 
#'   x-axis label's div.
#' @param axisHeight Height, in pixels, of the x-axis. If not set explicitly, 
#'   this is computed based on \code{axisLabelFontSize} and \code{axisTickSize}.
#' @param axisLineColor Color of the x- and y-axis lines. Accepts any value 
#'   which the HTML canvas strokeStyle attribute understands, e.g. 'black' or 
#'   'rgb(0, 100, 255)'.
#' @param axisLineWidth Thickness (in pixels) of the x- and y-axis lines.
#' @param axisLabelColor Color for x- and y-axis labels. This is a CSS color 
#'   string. This may also be set globally using \code{dyOptions}.
#' @param axisLabelFontSize Size of the font (in pixels) to use in the axis 
#'   labels, both x- and y-axis. This may also be set globally using 
#'   \code{dyOptions}.
#' @param axisLabelWidth Width, in pixels, of the axis labels
#' @param axisLabelFormatter JavaScript function to call to format the tick 
#'   values that appear along an axis (see the 
#'   \href{http://dygraphs.com/options.html}{dygraphs documentation} for 
#'   additional details).
#' @param valueFormatter JavaScript function to call to provide a custom display
#'   format for the values displayed on mouseover (see the 
#'   \href{http://dygraphs.com/options.html}{dygraphs documentation} for 
#'   additional details).
#' @param pixelsPerLabel Number of pixels to require between each x- and 
#'   y-label. Larger values will yield a sparser axis with fewer ticks. Defaults
#'   to 50 (x-axis) or 30 (y-axes).
#' @param drawGrid Whether to display grid lines in the chart.
#' @param gridLineColor The color of the grid lines.
#' @param gridLineWidth Thickness (in pixels) of the grid lines drawn under the 
#'   chart.
#' @param independentTicks Only valid for y and y2, has no effect on x: This 
#'   option defines whether the y axes should align their ticks or if they 
#'   should be independent. Possible combinations: 1.) y=true, y2=false 
#'   (default): y is the primary axis and the y2 ticks are aligned to the the 
#'   ones of y. (only 1 grid) 2.) y=false, y2=true: y2 is the primary axis and 
#'   the y ticks are aligned to the the ones of y2. (only 1 grid) 3.) y=true, 
#'   y2=true: Both axis are independent and have their own ticks. (2 grids) 4.) 
#'   y=false, y2=false: Invalid configuration causes an error.
#'   
#' @return Axis options
#'   
#' @note See the 
#'   \href{https://rstudio.github.io/dygraphs/gallery-axis-options.html}{online 
#'   documentation} for additional details and examples.
#'   
#' @examples
#' library(dygraphs)
#' 
#' dygraph(nhtemp, main = "New Haven Temperatures") %>%
#'   dyAxis("y", label = "Temp (F)", valueRange = c(40, 60)) %>%
#'   dyOptions(axisLineWidth = 1.5, fillGraph = TRUE, drawGrid = FALSE)
#'     
#' @export
dyAxis <- function(dygraph,
                   name, 
                   label = NULL, 
                   valueRange = NULL,
                   logscale = NULL,
                   ticker = NULL,
                   rangePad = NULL,
                   labelWidth = NULL,
                   labelHeight = NULL,
                   axisHeight = NULL,
                   axisLineColor = NULL,
                   axisLineWidth = NULL,
                   pixelsPerLabel = NULL,
                   axisLabelColor = NULL,
                   axisLabelFontSize = NULL,
                   axisLabelWidth = NULL,
                   axisLabelFormatter = NULL,
                   valueFormatter = NULL,
                   drawGrid = NULL,
                   gridLineColor = NULL,
                   gridLineWidth = NULL,
                   independentTicks = NULL) {
  
  if (!name %in% c("x", "y", "y2"))
    stop("Axis name must be 'x', 'y', or 'y2'")
  
  # copy attrs for modification
  attrs <- dygraph$x$attrs
  
  axis <- list()
  axis$name <- name
  axis$label <- label
  axis$options <- list()
  axis$options$valueRange <- valueRange
  
  if (!is.null(logscale)) {
    if (identical(name, "x"))
      axis$options$logscale <- logscale
    else
      attrs$logscale <- logscale
  }
  
   
  axis$options$ticker <- JS(ticker)
  if (!is.null(axisHeight)) {
    if (name == "x")
      attrs$xAxisHeight <- axisHeight
    else
      stop("axisHeight is only applicable to the x axis")
  }
  
  if (!is.null(rangePad))
    attrs[[sprintf("%sRangePad", axis$name)]] <- rangePad
  
  if (!is.null(labelWidth))
    attrs[[sprintf("%sLabelWidth", axis$name)]] <- labelWidth
  if (!is.null(labelHeight))
    attrs[[sprintf("%sLabelHeight", axis$name)]] <- labelHeight
  
  axis$options$axisLineColor <- axisLineColor
  axis$options$axisLineWidth <- axisLineWidth
  if (!is.null(pixelsPerLabel))
    axis$options$pixelsPerLabel <- pixelsPerLabel 
  axis$options$axisLabelColor <- axisLabelColor
  axis$options$axisLabelFontSize <- axisLabelFontSize
  
  if (!is.null(axisLabelWidth))
    axis$options$axisLabelWidth <- axisLabelWidth
  
  axis$options$axisLabelFormatter <- JS(axisLabelFormatter)
  axis$options$valueFormatter <- JS(valueFormatter)
  axis$options$drawGrid <- drawGrid
  axis$options$gridLineColor <- gridLineColor
  axis$options$gridLineWidth <- gridLineWidth
  axis$options$independentTicks <- independentTicks
 
  # label and options
  if (!is.null(axis$label))
    attrs[[sprintf("%slabel", axis$name)]] <- axis$label
  attrs$axes[[axis$name]] <- axis$options  
  
  # return modified dygraph
  dygraph$x$attrs <- mergeLists(dygraph$x$attrs, attrs)
  dygraph
}

context("dyAxis")

test_that("axis creation", {
  d <- dygraph(ldeaths) %>% dyAxis("x", label = "x-axis", drawGrid = FALSE)
  expect_identical(d$x$attrs$xlabel, "x-axis")
  expect_identical(d$x$attrs$axes$x$drawGrid, FALSE)
})

#' dygraph region shading
#' 
#' Specify that a region of a dygraph be drawn with a background shading
#' 
#' @param dygraph Dygraph to add shading to
#' @param from Date/time or numeric to shade from (for date/time this must be a
#'   \code{as.POSIXct} object or another object convertible via
#'   \code{as.POSIXct}).
#' @param to Date/time or numeric to shade to (for date/time this must be a
#'   \code{as.POSIXct} object or another object convertible via
#'   \code{as.POSIXct}).
#' @param color Color of shading. This can be of the form "#AABBCC" or 
#'   "rgb(255,100,200)" or "yellow". Defaults to a very light gray.
#' @param axis Axis to apply shading.  Choices are "x" or "y".
#'   
#' @return A dygraph with the specified shading
#'  
#' @examples 
#' library(dygraphs)
#' 
#' dygraph(nhtemp, main = "New Haven Temperatures") %>% 
#'   dyShading(from = "1920-1-1", to = "1930-1-1") %>%
#'   dyShading(from = "1940-1-1", to = "1950-1-1")
#'   
#' dygraph(nhtemp, main = "New Haven Temperatures") %>% 
#'   dyShading(from = "48", to = "52", axis = "y") %>%
#'   dyShading(from = "50", to = "50.1", axis = "y", color = "black")
#'     
#' @note See the
#'   \href{https://rstudio.github.io/dygraphs/gallery-annotations.html}{online
#'   documentation} for additional details and examples.
#'   
#' @export
dyShading <- function(dygraph, from, to, color = "#EFEFEF", axis = "x") {
  
  # create shading
  shading <- list()
  shading$from <- ifelse(axis == "x" && dygraph$x$format == "date",
                         asISO8601Time(from), from)
  shading$to <- ifelse(axis == "x" && dygraph$x$format == "date",
                       asISO8601Time(to), to)
  shading$color <- color
  shading$axis <- axis
 
  # add it to list of shadings
  dygraph$x$shadings[[length(dygraph$x$shadings) + 1]] <- shading
  
  # return modified dygraph
  dygraph
}

context("dyShading")

test_that("shading creation", {
  d <- dygraph(nhtemp, main = "New Haven Temperatures") %>% 
         dyShading(from = "1920-1-1", to = "1930-1-1") %>%
         dyShading(from = "1940-1-1", to = "1950-1-1")   
  expect_equal(length(d$x$shadings), 2)
  expect_equal(length(d$x$shadings[[1]]), 4)
  expect_equal(length(d$x$shadings[[2]]), 4)
})

#' Callbacks for dygraph events
#' 
#' Set JavaScript callbacks for various dygraph events. See the 
#' \href{http://dygraphs.com/options.html}{dygraph options} reference for 
#' additional details on the signature of each callback.
#' 
#' @param dygraph Dygraph to add callbacks to
#' @param clickCallback A function to call when the canvas is clicked.
#' @param drawCallback When set, this callback gets called every time the 
#'   dygraph is drawn. This includes the initial draw, after zooming and 
#'   repeatedly while panning.
#' @param highlightCallback When set, this callback gets called every time a new
#'   point is highlighted.
#' @param pointClickCallback A function to call when a data point is clicked. 
#'   and the point that was clicked.
#' @param underlayCallback When set, this callback gets called before the chart 
#'   is drawn.
#' @param unhighlightCallback When set, this callback gets called every time the
#'   user stops highlighting any point by mousing out of the graph.
#' @param zoomCallback A function to call when the zoom window is changed 
#'   (either by zooming in or out).
#' @param drawHighlightPointCallback Draw a custom item when a point is 
#'   highlighted. Default is a small dot matching the series color. This method 
#'   should constrain drawing to within pointSize pixels from (cx, cy)
#' @param drawPointCallback Draw a custom item when drawPoints is enabled. 
#'   Default is a small dot matching the series color. This method should 
#'   constrain drawing to within pointSize pixels from (cx, cy).
#' @param annotationClickHandler JavaScript function to call when an annotation
#'   is clicked. This can also be specified on a per-annotation basis.
#' @param annotationMouseOverHandler JavaScript function to call when the mouse
#'   hovers over an annotation. This can also be specified on a per-annotation
#'   basis.
#' @param annotationMouseOutHandler JavaScript function to call when the mouse
#'   exits an annotation. This can also be specified on a per-annotation basis.
#' @param annotationDblClickHandler JavaScript function to call when an
#'   annotation is double clicked. This can also be specified on a
#'   per-annotation basis.
#'   
#' @return Dygraph with callbacks
#'   
#' @export
dyCallbacks <- function(dygraph,
                        clickCallback = NULL,
                        drawCallback = NULL,
                        highlightCallback = NULL,
                        pointClickCallback = NULL,
                        underlayCallback = NULL,
                        unhighlightCallback = NULL,
                        zoomCallback = NULL,
                        drawHighlightPointCallback = NULL,
                        drawPointCallback = NULL,
                        annotationClickHandler = NULL,
                        annotationMouseOverHandler = NULL,
                        annotationMouseOutHandler = NULL,
                        annotationDblClickHandler = NULL) {
  
  callbacks <- list()
  callbacks$clickCallback <- JS(clickCallback)
  callbacks$drawCallback <- JS(drawCallback)
  callbacks$highlightCallback <- JS(highlightCallback)
  callbacks$pointClickCallback <- JS(pointClickCallback)
  callbacks$underlayCallback <- JS(underlayCallback)
  callbacks$unhighlightCallback <- JS(unhighlightCallback)
  callbacks$zoomCallback <- JS(zoomCallback)
  callbacks$drawHighlightPointCallback <- JS(drawHighlightPointCallback)
  callbacks$drawPointCallback <- JS(drawPointCallback)
  callbacks$annotationClickHandler <- JS(annotationClickHandler)
  callbacks$annotationMouseOverHandler <- JS(annotationMouseOverHandler)
  callbacks$annotationMouseOutHandler <- JS(annotationMouseOutHandler)
  callbacks$annotationDblClickHandler <- JS(annotationDblClickHandler)
  
  dygraph$x$attrs <- mergeLists(dygraph$x$attrs, callbacks)
  
  dygraph
}

#' dygraph legend
#' 
#' Configure options for the dygraph series legend.
#' 
#' @param dygraph Dygraph to configure legend options for.
#' @param show When to display the legend. Specify "always" to always show the 
#'   legend. Specify "onmouseover" to only display it when a user mouses over 
#'   the chart. Specify "follow" to have the legend show as overlay to the chart
#'   which follows the mouse. The default behavior is "auto", which results in
#'   "always" when more than one series is plotted and "onmouseover" when only a
#'   single series is plotted.
#' @param width Width (in pixels) of the div which shows the legend.
#' @param showZeroValues Show zero value labels in the legend.
#' @param labelsDiv Show data labels in an external div, rather than on the 
#'   graph. This value should be a div element id.
#' @param labelsSeparateLines Put a <br/> between lines in the label string. 
#'   Often used in conjunction with \code{labelsDiv}.
#' @param hideOnMouseOut Whether to hide the legend when the mouse leaves the 
#'   chart area. This option applies when \code{show} is set to "onmouseover". 
#'   Note that this also affects the hiding of the \code{\link{dyHighlight}} on 
#'   mouse out.
#'   
#' @return A dygraph with customized legend options
#'  
#' @examples
#' library(dygraphs)
#' 
#' dygraph(nhtemp, main = "New Haven Temperatures") %>% 
#'   dySeries("V1", label = "Temperature (F)") %>%
#'   dyLegend(show = "always", hideOnMouseOut = FALSE)   
#'     
#' @note See the 
#'   \href{https://rstudio.github.io/dygraphs/gallery-plot-labels.html}{online
#'   documentation} for additional details and examples.
#'   
#' @export
dyLegend <- function(dygraph,
                     show = c("auto", "always", "onmouseover", "follow", "never"),
                     width = 250,
                     showZeroValues = TRUE,
                     labelsDiv = NULL,
                     labelsSeparateLines = FALSE,
                     hideOnMouseOut = TRUE) {
  
  legend <- list()
  legend$legend <- match.arg(show)
  if (legend$legend == "never") {
    legend$legend <- NULL
    legend$showLabelsOnHighlight <- FALSE
  }
  legend$labelsDivWidth <- width
  legend$labelsShowZeroValues <- showZeroValues
  legend$labelsDiv <- labelsDiv
  legend$labelsSeparateLines <- labelsSeparateLines
  legend$hideOverlayOnMouseOut <- hideOnMouseOut
   
  # merge legend
  dygraph$x$attrs <- mergeLists(dygraph$x$attrs, legend)
  
  # return modified dygraph
  dygraph
}

context("dyLegend")

test_that("legend creation", {
  d <- dygraph(ldeaths) %>% dyLegend(show = "always", hideOnMouseOut = TRUE)
  expect_identical(d$x$attrs$legend, "always")
  expect_identical(d$x$attrs$hideOverlayOnMouseOut, TRUE)
})

#' dygraph interactive plot for time series data
#' 
#' R interface to interactive time series plotting using the 
#' \href{http://dygraphs.com}{dygraphs} JavaScript library.
#' 
#' @inheritParams htmlwidgets::createWidget
#' 
#' @param data Either time series data or numeric data. For time series, this 
#'   must be an \link[xts]{xts} object or an object which is convertible to 
#'   \code{xts}. For numeric data, this must be a named list or data frame,
#'   where the first element/column provides x-axis values and all subsequent
#'   elements/columns provide one or more series of y-values.
#' @param periodicity Periodicity of time series data (automatically detected
#'   via \link[xts:periodicity]{xts::periodicity} if not specified).
#' @param main Main plot title (optional)
#' @param xlab X axis label
#' @param ylab Y axis label
#' @param group Group to associate this plot with. The x-axis zoom level of 
#'   plots within a group is automatically synchronized.
#' @param width Width in pixels (optional, defaults to automatic sizing)
#' @param height Height in pixels (optional, defaults to automatic sizing)
#'   
#' @return Interactive dygraph plot
#'   
#' @note
#' See the \href{https://rstudio.github.io/dygraphs/}{online documentation} for
#' additional details and examples.
#' 
#' @examples 
#' library(dygraphs)
#' lungDeaths <- cbind(mdeaths, fdeaths)
#' dygraph(lungDeaths)
#' 
#' indoConc <- Indometh[Indometh$Subject == 1, c("time", "conc")]
#' dygraph(indoConc)
#' 
#' @export
dygraph <- function(data, main = NULL, xlab = NULL, ylab = NULL,
                    periodicity = NULL, group = NULL, 
                    elementId = NULL, width = NULL, height = NULL) {
  
  # Test whether x-axis are dates or numeric
  if (xts::xtsible(data)) {
    
    if (!xts::is.xts(data))
      data <- xts::as.xts(data)
    format <- "date"
    
  } else if (is.list(data) && is.numeric(data[[1]])) {
    
    if (is.null(names(data)))
      stop("For numeric values, 'data' must be a named list or data frame")
    format <- "numeric"
    
  } else {
    stop("Unsupported type passed to argument 'data'.")
  }
  
  if (format == "date") {
    
    # auto-detect periodicity if not otherwise specified
    if (is.null(periodicity)) {
      if (nrow(data) < 2) {
        periodicity <- defaultPeriodicity(data)
      } else {
        periodicity <- xts::periodicity(data)
      }
    }
    
    # extract time
    time <- time(data)
    
    # get data as a named list
    data <- zoo::coredata(data)
    data <- unclass(as.data.frame(data))
    
    # merge time back into list and convert to JS friendly string
    timeColumn <- list()
    timeColumn[[periodicity$label]] <- asISO8601Time(time)
    data <- append(timeColumn, data)
  } else {
    # Convert data to list if it was data frame
    data <- as.list(data)
  }
  
  # create native dygraph attrs object
  attrs <- list()
  attrs$title <- main
  attrs$xlabel <- xlab
  attrs$ylabel <- ylab
  attrs$labels <- names(data)
  attrs$legend <- "auto"
  attrs$retainDateWindow <- FALSE
  attrs$axes$x <- list() 
  attrs$axes$x$pixelsPerLabel <- 60
   
  # create x (dygraph attrs + some side data)
  x <- list()
  x$attrs <- attrs
  x$scale <- if (format == "date") periodicity$scale else NULL
  x$group <- group
  x$annotations <- list()
  x$shadings <- list()
  x$events <- list()
  x$format <- format # Add format for further processing here
  
  # Add attributes required for defining custom series. When a dySeries call
  # is made it places series definition in "manual mode". In this case we
  # need to save the original data.
  attr(x, "time") <- if (format == "date") time else NULL
  attr(x, "data") <- data
  attr(x, "autoSeries") <- 2
  
  # add data (strip names first so we marshall as a 2d array)
  names(data) <- NULL
  x$data <- data
    
  # create widget
  htmlwidgets::createWidget(
    name = "dygraphs",
    x = x,
    width = width,
    height = height,
    htmlwidgets::sizingPolicy(viewer.padding = 10, browser.fill = TRUE),
    elementId = elementId
  )
}

#' Shiny bindings for dygraph
#' 
#' Output and render functions for using dygraph within Shiny 
#' applications and interactive Rmd documents.
#' 
#' @param outputId output variable to read from
#' @param width,height Must be a valid CSS unit (like \code{"100\%"},
#'   \code{"400px"}, \code{"auto"}) or a number, which will be coerced to a
#'   string and have \code{"px"} appended.
#' @param expr An expression that generates a dygraph
#' @param env The environment in which to evaluate \code{expr}.
#' @param quoted Is \code{expr} a quoted expression (with \code{quote()})? This 
#'   is useful if you want to save an expression in a variable.
#'   
#' @name dygraph-shiny
#'
#' @export
dygraphOutput <- function(outputId, width = "100%", height = "400px") {
  htmlwidgets::shinyWidgetOutput(outputId, "dygraphs", width, height)
}

#' @rdname dygraph-shiny
#' @export
renderDygraph <- function(expr, env = parent.frame(), quoted = FALSE) {
  if (!quoted) { expr <- substitute(expr) } # force quoted
  htmlwidgets::shinyRenderWidget(expr, dygraphOutput, env, quoted = TRUE)
}

context("dyEvent")

test_that("event creation", {
  d <- dygraph(presidents, main = "Presidential Approval") %>%
         dyEvent(x = "1950-6-30", "Korea", labelLoc = "bottom") %>%
         dyEvent(x = "1965-2-09", "Vietnam", labelLoc = "bottom")
  dd <- dygraph(presidents, main = "Presidential Approval") %>%
         dyEvent(x = c("1950-6-30", "1965-2-09"), c("Korea", "Vietnam"), labelLoc = "bottom")
  expect_equal(length(d$x$events), 2)
  expect_identical(d, dd)
  expect_identical(d$x$events[[1]]$label, "Korea")
  expect_identical(d$x$events[[2]]$label, "Vietnam")
})

context("dySeries")

test_that("series creation", {
  d <- dygraph(ldeaths) %>% dySeries(label = "series", fillGraph = FALSE)
  expect_identical(d$x$attrs$labels[[2]], "series")
  expect_identical(d$x$attrs$series[[1]]$fillGraph, FALSE)
})

context("dyRoller")

test_that("roller creation", {
  d <- dygraph(discoveries, main = "Important Discoveries") %>% 
         dyRoller(rollPeriod = 5)
  expect_identical(d$x$attrs$showRoller, TRUE)
  expect_identical(d$x$attrs$rollPeriod, 5)
})

context("dygraph")

test_that("dygraph creation", {
  expect_is(dygraph(ldeaths), class = "dygraphs")
  expect_identical(dygraph(ldeaths, xlab = "x")$x$attrs$xlabel, "x")
  expect_identical(dygraph(ldeaths, ylab = "y")$x$attrs$ylabel, "y")
})

context("dyHighlight")

test_that("highlight creation", {
  lungDeaths <- cbind(ldeaths, mdeaths, fdeaths)
  d <- dygraph(lungDeaths, main = "Deaths from Lung Disease (UK)") %>%
         dyHighlight(highlightCircleSize = 5, 
                     highlightSeriesBackgroundAlpha = 0.2,
                     hideOnMouseOut = FALSE)   
  expect_identical(d$x$attrs$highlightCircleSize, 5)
  expect_identical(d$x$attrs$highlightSeriesBackgroundAlpha, 0.2)
})

#' dygraph data series
#' 
#' Add a data series to a dygraph plot. Note that options will use the default 
#' global setting (as determined by \code{\link{dyOptions}}) when not specified 
#' explicitly. When no \code{dySeries} is specified for a plot then all series 
#' within the underlying data are plotted.
#' 
#' @inheritParams dyOptions
#'   
#' @param dygraph Dygraph to add a series definition to
#' @param name Name of series within data set. If no name is specified then 
#'   series are bound to implicitly based on their order within the underlying 
#'   time series object. This parameter can also be a character vector of length
#'   3 that specifies a set of input column names to use as the lower, value,
#'   and upper for a series with a shaded bar drawn around it.
#' @param label Label to display for series (uses name if no label defined)
#' @param color Color for series. These can be of the form "#AABBCC" or 
#'   "rgb(255,100,200)" or "yellow", etc. If not specified then the global 
#'   colors option (typically based on equally-spaced points around a color wheel). 
#' @param axis Y-axis to associate the series with ("y" or "y2")
#' @param stepPlot When set, display the graph as a step plot instead of a line 
#'   plot.
#' @param stemPlot When set, display the graph as a stem plot instead of a line
#'   plot.
#' @param fillGraph Should the area underneath the graph be filled? This option 
#'   is not compatible with error bars.
#' @param drawPoints Draw a small dot at each point, in addition to a line going
#'   through the point. This makes the individual data points easier to see, but
#'   can increase visual clutter in the chart.
#' @param pointSize The size of the dot to draw on each point in pixels. A dot 
#'   is always drawn when a point is "isolated", i.e. there is a missing point 
#'   on either side of it. This also controls the size of those dots.
#' @param pointShape The shape of the dot to draw. Can be one of the following:
#'   "dot" (default), "triangle", "square", "diamond", "pentagon", "hexagon",
#'   "circle", "star", "plus" or "ex".
#' @param strokeWidth The width of the lines connecting data points. This can be
#'   used to increase the contrast or some graphs.
#' @param strokePattern A predefined stroke pattern type ("dotted", "dashed", or
#'   "dotdash") or a custom pattern array where the even index is a draw and odd
#'   is a space in pixels. If \code{NULL} then it draws a solid line. The array 
#'   should have an even length as any odd length array could be expressed as 
#'   a smaller even length array.
#' @param strokeBorderWidth Draw a border around graph lines to make crossing 
#'   lines more easily distinguishable. Useful for graphs with many lines.
#' @param strokeBorderColor Color for the line border used if 
#'   \code{strokeBorderWidth} is set.
#' @param plotter A function which plots the data series. May also be set on on 
#'   a global basis using \code{dyOptions}. See the 
#'   \href{http://dygraphs.com/tests/plotters.html}{dygraphs documentation} for 
#'   additional details on plotting functions.
#'   
#' @return Dygraph with additional series
#'   
#' @examples
#' library(dygraphs)
#' 
#' lungDeaths <- cbind(ldeaths, mdeaths, fdeaths)
#' 
#' dygraph(lungDeaths, main = "Deaths from Lung Disease (UK)") %>%
#'   dySeries("mdeaths", drawPoints = TRUE, color = "blue") %>%
#'   dySeries("fdeaths", stepPlot = TRUE, color = "red")   
#'   
#' @note See the 
#'   \href{https://rstudio.github.io/dygraphs/gallery-series-options.html}{online
#'   documentation} for additional details and examples.
#'   
#' @export
dySeries <- function(dygraph,
                     name = NULL, 
                     label = NULL,
                     color = NULL,
                     axis = "y", 
                     stepPlot = NULL,
                     stemPlot = NULL,
                     fillGraph = NULL,
                     drawPoints = NULL,
                     pointSize = NULL,
                     pointShape = NULL,
                     strokeWidth = NULL,
                     strokePattern = NULL,
                     strokeBorderWidth = NULL,
                     strokeBorderColor = NULL,
                     plotter = NULL) {
  
  # get a reference to the underlying data and labels
  data <- attr(dygraph$x, "data")
  labels <- names(data)
  
  # when setting up the first color, we start handling colors here 
  if (!is.null(color) && is.null(dygraph$x$attrs$colors)) {
      colors <- dygraphColors(dygraph, length(labels) - 1)
      dygraph$x$attrs$colors <- colors
  }

  # prepare the colors list for processing 
  if (!is.null(dygraph$x$attrs$colors)) {
     colors <- dygraph$x$attrs$colors
     names(colors) <- dygraph$x$attrs$labels[-1]
  }
   
  # auto-bind name if necessary
  autobind <- attr(dygraph$x, "autoSeries")
  if (is.null(name))
    name <- labels[[autobind]]
  attr(dygraph$x, "autoSeries") <- autobind + length(name)
  
  # ensure that name is of length 1 or 3
  if (length(name) != 1 && length(name)  != 3) {
    stop("The name parameter must either be a character vector ",
         "of length one or three")
  }
  
  # Get the cols where this series is located and verify that they are
  # available within the underlying dataset
  cols <- which(labels %in% name)
  if (length(cols) != length(name)) {
    stop("One or more of the specified series were not found. ",
         "Valid series names are: ", paste(labels[-1], collapse = ", "))
  }
  
  
  # Data series named here are "consumed" from the automatically generated
  # list of series (they'll be added back in below)
  cols <- which(dygraph$x$attrs$labels %in% name)
  dygraph$x$data <- dygraph$x$data[-c(cols)]
  dygraph$x$attrs$labels <- dygraph$x$attrs$labels[-c(cols)]
   
  # Resolve stemPlot into a custom plotter if necessary
  plotter <- resolveStemPlot(stemPlot, plotter)
  
  # create series object
  series <- list()
  series$name <- name
  series$label <- label
  series$options <- list()
  series$options$axis <- match.arg(axis, c("y", "y2"))
  series$options$stepPlot <- stepPlot
  series$options$fillGraph <- fillGraph
  series$options$drawPoints <- drawPoints
  series$options$pointSize <- pointSize
  series$options$strokeWidth <- strokeWidth
  series$options$strokePattern <- resolveStrokePattern(strokePattern)
  series$options$strokeBorderWidth <- strokeBorderWidth
  series$options$strokeBorderColor <- strokeBorderColor
  series$options$plotter <- JS(plotter)

 
  # copy attrs for modification
  attrs <- dygraph$x$attrs
  
  # resolve multi-series
  if (length(series$name) == 3) {
    
    # find column indexes within the data
    cols <- integer(3)
    for (i in 1:3) {
      col <- which(labels == series$name[[i]])
      if (length(col) != 1)
        stop("Series name '", series$name[[i]], "' not found in input data")

      cols[[i]] <- col
    }
    
    # mark attrs as containing custom bars
    attrs$customBars <- TRUE
    
    # fixup name
    series$name <- series$name[[2]]
    
    # compute series data
    seriesData <- toMultiSeries(data[[cols[[1]]]],
                                data[[cols[[2]]]],
                                data[[cols[[3]]]])
    
  } else {
    # select series data
    seriesData <- data[[series$name]]
  }
  
  # grab the colors for the series being processed
  if (!is.null(dygraph$x$attrs$colors)) {
    currColors <- colors[names(colors) %in% name]

    if (!is.null(color)) 
			currColors[[series$name]] <- color
    
    colors <- colors[!names(colors) %in% name]
    colors[[series$name]] <- currColors[[series$name]]
    
    # compensating for the bug whereas a single series dygraph with specified color
    # shows up as black since the DateTime series tries to take the first color
    if (length(colors) == 1) colors <- c(colors, colors)
    
    attrs$colors <- colors
    names(attrs$colors) <- NULL
  }
  
  
  # default the label if we need to
  if (is.null(series$label))
    series$label <- series$name  
   
  # add label
  attrs$labels <- c(attrs$labels, series$label)
   
  # set options
  attrs$series[[series$label]] <- series$options
 
  # set attrs
  dygraph$x$attrs <- attrs

  # set point shape
  if (!is.null(pointShape)) {
    shapes <- c("dot", "triangle", "square", "diamond", "pentagon",
                "hexagon", "circle", "star", "plus", "ex")

    if (!is.element(pointShape, shapes)) {
      stop("Invalid value for pointShape parameter. ",
           "Should be one of the following: ",
           "'dot', 'triangle', 'square', 'diamond', 'pentagon', ",
           "'hexagon', 'circle', 'star', 'plus' or 'ex'")
    }

    if (pointShape != "dot") {
      dygraph$x$pointShape <- list()
      dygraph$x$pointShape[[series$label]] <- pointShape
    }
  }
  
  # add data
  dygraph$x$data[[length(dygraph$x$data) + 1]] <- seriesData
  
  # if we are using custom bars then fixed non custom-bar data
  # series to still display
  if (isTRUE(attrs$customBars)) {
    for (i in 2:length(dygraph$x$data)) {
      if (length(dygraph$x$data[[i]][[1]]) == 1) {
        values <- dygraph$x$data[[i]]
        dygraph$x$data[[i]] <- toMultiSeries(values, values, values)
      }
    }
  }
  
  # return modified dygraph
  dygraph
}


#' Add series data to dygraph
#' 
#' Add an additional column of series data to a dygraph. This
#' is typically used in the construction of custom series types
#' (e.g. log scaled, smoothed, etc.)
#' 
#' @param dygraph Dygraph to add data to
#' @param name Name of series
#' @param values Data values
#' 
#' @return Dygraph with additional series data 
#' 
#' @export 
dySeriesData <- function(dygraph, name, values) {
  
  # add values
  data <- attr(dygraph$x, "data")
  data[[name]] <- values
  attr(dygraph$x, "data") <- data
  
  # return modified dygraph
  dygraph
}


# return a list of three element arrays 
toMultiSeries <- function(lower, value, upper) {  
  series <- vector(mode = "list", length = length(value))
  for (i in 1:length(series))
    series[[i]] <- c(lower[[i]], value[[i]], upper[[i]])

  series
}

# provide a custom plotter if stemPlot has been specified
resolveStemPlot <- function(stemPlot, plotter) {
  
  # check for stemPlot argument
  if (isTRUE(stemPlot)) {
    
    # verify that a custom plotter hasn't been specified
    if (!is.null(plotter)) {
      stop("stemPlot provides it's own plotter so is incompatible with ",
           "specifying a custom plotter", call. = FALSE)
    }
    
    # provide custom plotter JS
    "function stemPlotter(e) { 
       var ctx = e.drawingContext; 
       var points = e.points; 
       var y_bottom = e.dygraph.toDomYCoord(0);
       ctx.fillStyle = e.color; 
       for (var i = 0; i < points.length; i++) { 
          var p = points[i]; 
          var center_x = p.canvasx;
          var center_y = p.canvasy; 
          ctx.beginPath(); 
          ctx.moveTo(center_x, y_bottom); 
          ctx.lineTo(center_x, center_y); 
          ctx.stroke();
          ctx.beginPath(); 
          ctx.arc(center_x, center_y, 3, 0, 2*Math.PI); 
          ctx.stroke();
       }
    }"
  } else {
    # specified plotter
    plotter
  }
}

dygraphColors <- function(dygraph, num) {

  # These are used for when no custom colors are specified.
  sat <- dygraph$x$attrs$colorSaturation %||% 1.0
  val <- dygraph$x$attrs$colorValue %||% 0.5
  half <- ceiling(num / 2)
  
  colors<-c()
  
  for (i in 0:(num-1)) {
    # alternate colors for high contrast.
    idx <- ifelse(i %% 2, (half + (i + 1) / 2), ceiling((i + 1) / 2))
    hue <- (1.0 * idx / (1 + num))
    color <- hsvToRGB(hue, sat, val)
    colors <- c(colors, color)

  }

  return(colors)
}

#' @importFrom grDevices rgb
hsvToRGB <- function (hue, saturation, value) {
  if (saturation == 0) {
    red = value
    green = value
    blue = value

  } else {
    i <- floor(hue * 6)
    f <- (hue * 6) - i
    p <- value * (1 - saturation)
    q <- value * (1 - (saturation * f))
    t <- value * (1 - (saturation * (1 - f)))

    # converting the switch from the JS library to a vector selection    
    red <- c(value, q, p, p, t, value, value)
    green <- c(t, value, value, q, p, p, t)
    blue <- c(p, p, t, value, value, q, p)
    
    r <- red[i + 1]
    g <- green[i + 1]
    b <- blue[i + 1]
  }

  red <- floor(255 * r + 0.5)
  green <- floor(255 * g + 0.5)
  blue <- floor(255 * b + 0.5)
  return (rgb(red, green, blue, maxColorValue = 255))
}

`%||%` <- function(a, b){
  if(!is.null(a)) return(a)
  return(b)
}

#' Rebase data handler for straw broom charts with Dygraph
#'
#' Draw a straw broom chart.
#'
#' @param dygraph Dygraph to draw chart on
#' @param value Value to rebase to
#' @param percent If true, ingnore value argument and rebase to percentage difference
#'
#' @return Dygraph with specified straw broom chart
#'
#' @examples
#' \dontrun{
#' library(quantmod)
#' tickers <- c("AAPL", "MSFT")
#' getSymbols(tickers)
#' closePrices <- do.call(merge, lapply(tickers, function(x) Cl(get(x))))
#' dateWindow <- c("2008-01-01", "2009-01-01")
#' dygraph(closePrices, main = "Value", group = "stock") %>%
#'   dyRebase(value = 100) %>%
#'   dyRangeSelector(dateWindow = dateWindow)
#' dygraph(closePrices, main = "Percent", group = "stock") %>%
#'   dyRebase(percent = TRUE) %>%
#'   dyRangeSelector(dateWindow = dateWindow)
#' dygraph(closePrices, main = "None", group = "stock") %>%
#'   dyRangeSelector(dateWindow = dateWindow)
#' }
#'
#' @export
dyRebase <- function(dygraph, value = 100, percent = FALSE) {
  if (percent) {
    base <- "percent"
  } else if (is.numeric(value)) {
    base <- value
  } else {
    stop("value must be of numeric type", call. = FALSE)
  }

  dyPlugin(dygraph = dygraph,
           name = "Rebase",
           path = system.file("plugins/rebase.js", package = "dygraphs"),
           options = base)
}

#' dygraph rolling average period text box
#' 
#' Add a rolling average period text box to the bottom left of the plot. Y 
#' values are averaged over the specified number of time scale units.
#' 
#' @param dygraph Dygraph to add roller to
#' @param showRoller Whether to show the roller
#' @param rollPeriod Number of time scale units (e.g. days, months, years) to 
#'   average values over.
#'   
#' @return A dygraph that displays a range selector
#'   
#' @examples 
#' library(dygraphs)
#' 
#' dygraph(discoveries, main = "Important Discoveries") %>% 
#'   dyRoller(rollPeriod = 5)
#' 
#' @note See the
#' \href{https://rstudio.github.io/dygraphs/gallery-roll-periods.html}{online
#' documentation} for additional details and examples.
#' 
#' @export
dyRoller <- function(dygraph,
                     showRoller = TRUE,
                     rollPeriod = 1) {
  
  roller <- list()
  roller$showRoller <- showRoller
  roller$rollPeriod <- rollPeriod
  
  # merge roller
  dygraph$x$attrs <- mergeLists(dygraph$x$attrs, roller)
  
  # return modified dygraph
  dygraph
}

#' dygraph series mouse-over highlighting
#' 
#' Configure options for data series mouse-over highlighting. Note that 
#' highlighting is always enabled for dygraphs so this function is used to 
#' customize rather than enable highlighting.
#' 
#' @param dygraph Dygraph to configure highlighting behavior for.
#' @param highlightCircleSize The size in pixels of the dot drawn over 
#'   highlighted points.
#' @param highlightSeriesBackgroundAlpha Fade the background while highlighting 
#'   series. 1=fully visible background (disable fading), 0=hidden background 
#'   (show highlighted series only).
#' @param highlightSeriesOpts When set, the options from this list are applied 
#'   to the series closest to the mouse pointer for interactive highlighting. 
#'   Example: list(strokeWidth = 3). See the documentation on 
#'   \code{\link{dySeries}} for additional details on options that can be set.
#' @param hideOnMouseOut Whether to hide the highlighting effects when the mouse
#'   leaves the chart area. Note that this also affects the hiding of the 
#'   \code{\link{dyLegend}} on mouse out.
#'   
#' @return A dygraph with customized highlighting options
#'   
#' @examples 
#' library(dygraphs)
#' lungDeaths <- cbind(ldeaths, mdeaths, fdeaths)
#' dygraph(lungDeaths, main = "Deaths from Lung Disease (UK)") %>%
#'   dyHighlight(highlightCircleSize = 5, 
#'               highlightSeriesBackgroundAlpha = 0.2,
#'               hideOnMouseOut = FALSE)   
#'  
#' @note See the
#' \href{https://rstudio.github.io/dygraphs/gallery-series-highlighting.html}{online
#' documentation} for additional details and examples.
#' 
#' @export
dyHighlight <- function(dygraph,
                        highlightCircleSize = 3,
                        highlightSeriesBackgroundAlpha = 0.5,
                        highlightSeriesOpts = list(),
                        hideOnMouseOut = TRUE) {
  
  highlight <- list()
  highlight$highlightCircleSize <- highlightCircleSize
  highlight$highlightSeriesBackgroundAlpha <- highlightSeriesBackgroundAlpha
  highlight$highlightSeriesOpts <- highlightSeriesOpts
  highlight$hideOverlayOnMouseOut <- hideOnMouseOut
  
  # merge highlight
  dygraph$x$attrs <- mergeLists(dygraph$x$attrs, highlight)
  
  # return modified dygraph
  dygraph
}

#' dygraph event line
#' 
#' Add a vertical event line to a dygraph
#' 
#' @param dygraph Dygraph to add event line to
#' @param x Either numeric or date/time for the event, depending on the format
#'   of the x-axis of the dygraph. (For date/time must be a \code{POSIXct}
#'   object or another object convertible to \code{POSIXct} via
#'   \code{as.POSIXct})
#' @param label Label for event. Defaults to blank.
#' @param labelLoc Location for label (top or bottom)
#' @param color Color of event line. This can be of the form "#AABBCC" or 
#'   "rgb(255,100,200)" or "yellow". Defaults to black.
#' @param strokePattern A predefined stroke pattern type ("dotted", "dashed",
#'   "dotdash", or "solid") or a custom pattern array where the even index is 
#'   a draw and odd is a space in pixels. Defaults to dashed.
#' @param date (deprecated) See argument \code{x}.
#'   
#' @return A dygraph with the specified event line.
#'   
#' @note See the \href{https://rstudio.github.io/dygraphs/gallery-event-lines.html}{online 
#'   documentation} for additional details and examples.
#'   
#' @examples 
#' library(dygraphs)
#' 
#' dygraph(presidents, main = "Presidential Approval") %>%
#'   dyAxis("y", valueRange = c(0, 100)) %>%
#'   dyEvent("1950-6-30", "Korea", labelLoc = "bottom") %>%
#'   dyEvent("1965-2-09", "Vietnam", labelLoc = "bottom")

#' dygraph(presidents, main = "Presidential Approval") %>%
#'   dyAxis("y", valueRange = c(0, 100)) %>%
#'   dyEvent(c("1950-6-30", "1965-2-09"), c("Korea", "Vietnam"), labelLoc = "bottom") 
#'  
#' @export
dyEvent <- function(dygraph, 
                    x,
                    label = NULL, 
                    labelLoc = c("top", "bottom"),
                    color = "black", 
                    strokePattern = "dashed",
                    date) {
  
  # Check usage of deprecated argument 'date'
  if (!missing(date)) {
    x <- date
    warning("Argument 'date' is deprecated, please use argument 'x' instead")
  }
  
  # create events
  if (!is.null(label) && length(x) != length(label))
    stop("Length of 'x' and 'label' does not match")
  events <-
    lapply(
      seq_along(x), function(i)
      {
        if (is.null(label))
          eventLabel <- NULL
        else
          eventLabel <- label[i]
        list(
          pos = ifelse(dygraph$x$format == "date", asISO8601Time(x[i]), x[i])
         ,label = eventLabel
         ,labelLoc = match.arg(labelLoc, c("top", "bottom"))
         ,color = color
         ,strokePattern = resolveStrokePattern(strokePattern)
         ,axis = "x"
        )
      }
    )
 
  # add it to list of events
  dygraph$x$events <- c(dygraph$x$events, events)
  
  # return modified dygraph
  dygraph
}

#' dygraph limit line
#' 
#' Add a horizontal limit line to a dygraph
#' 
#' @param dygraph Dygraph to add limit line to
#' @param limit Numeric position of limit.
#' @param label Label for limit. Defaults to blank.
#' @param labelLoc Location for label (left or right).
#' @param color Color of limit line. This can be of the form "#AABBCC" or 
#'   "rgb(255,100,200)" or "yellow". Defaults to black.
#' @param strokePattern A predefined stroke pattern type ("dotted", "dashed",
#'   "dotdash", or "solid") or a custom pattern array where the even index is 
#'   a draw and odd is a space in pixels. Defaults to dashed.
#'   
#' @return A dygraph with the specified limit line.
#'   
#' @note See the \href{https://rstudio.github.io/dygraphs/gallery-event-lines.html}{online 
#'   documentation} for additional details and examples.
#'   
#' @examples 
#' library(dygraphs)
#' 
#' dygraph(presidents, main = "Presidential Approval") %>%
#'   dyAxis("y", valueRange = c(0, 100)) %>% 
#'   dyLimit(max(presidents, na.rm = TRUE), "Max",
#'           strokePattern = "solid", color = "blue")
#'  
#' @export
dyLimit <- function(dygraph, 
                    limit,
                    label = NULL, 
                    labelLoc = c("left", "right"),
                    color = "black", 
                    strokePattern = "dashed") {
  
  # create event
  event <- list()
  event$pos <- limit
  event$label <- label
  event$labelLoc <- match.arg(labelLoc)
  event$color <- color
  event$strokePattern <- resolveStrokePattern(strokePattern)
  event$axis <- "y"
  
  # add it to list of events
  dygraph$x$events[[length(dygraph$x$events) + 1]] <- event
  
  # return modified dygraph
  dygraph
}

#!/bin/mksh
# © 2019, 2023 mirabilos <m$(date +%Y)@mirbsd.de> Ⓕ CC0 or MirBSD or MIT

thisprojecttopdir=dygraphs-
thisprojectprefix=dygraphs_
thisprojectsuffix=.orig.tar.gz
set -A usecompressor -- gzip -n9
function thisprojectversionmapping {
	REPLY=${1#v}
}

export LC_ALL=C
unset LANGUAGE
set -e
set -o pipefail
case $KSH_VERSION {
(*MIRBSD\ KSH*) ;;
(*) echo E: do not call me with bash or something; exit 255 ;;
}

r=$(git rev-parse --show-toplevel)
[[ $r != /* ]] || r=$(realpath "$r")
if [[ $r != /* || "$(realpath .)" != "$r"?(/*) ]]; then
	print -ru2 -- "E: could not determine repository root"
	exit 1
fi
w=$PWD
cd "$r"

x=$(git status --porcelain) || x='<<ERROR>>'
if [[ -n $x ]]; then
	print -ru2 -- "N: git status info follows"
	print -r -- "$x" | sed 's/^/N:  /' >&2
	print -ru2 -- "E: source tree not clean"
	exit 1
fi

v=$(git describe --always --dirty --tags) || x='<<ERROR>>'
if [[ $v != [0-9A-Za-z]+([0-9A-Za-z.-]) ]]; then
	print -ru2 -- "E: cannot describe version"
	print -ru2 -- "N: $v"
	exit 1
fi
set -x
v=${|thisprojectversionmapping "$v";}
d=${thisprojecttopdir}$v
f=${thisprojectprefix}$v${thisprojectsuffix}
T=$(mktemp -d "$r/.tmp.XXXXXXXXXX")
trap 'set +e; cd "$r"; rm -rf "$T"' EXIT
mkdir -p "$T/$d"

x=$(git show --no-notes -s --pretty=tformat:%cd --date=format:%Y%m%d%H%M.%S)
[[ $x = 2[01][0-9][0-9][0-1][0-9][0-3][0-9][0-2][0-9][0-5][0-9].[0-6][0-9] ]]
git ls-tree -r --name-only -z HEAD | sort -z >"$T/.flst"

<"$T/.flst" xargs -0 touch -h -t "$x" --
<"$T/.flst" pax -rw -0 -l -pe "$T/$d/"
cd "$T"
jdupes -rL .

export d
<"$T/.flst" perl -0nle 'print "$ENV{d}/$_";' | \
    pax -w -0 -P -x ustar -o write_opt=nodir -b 512 -M dist | \
    "${usecompressor[@]}" >"$w/${f}~"
mv "$w/${f}~" "$w/$f"

#!/bin/mksh
case $KSH_VERSION {
(*MIRBSD\ KSH*) ;;
(*) echo E: do not call me with bash or something; exit 255 ;;
}

# initialisation
set -e
set -o pipefail
if command -v nodejs >/dev/null 2>&1; then
	node_js=nodejs
else
	node_js=node
fi
if [[ -e node_modules ]]; then
	# NPM setup
	babel_js=babel
else
	# Debian packaging
	babel_js=babeljs
fi
babelrc=$PWD/babel.config.json
set -x

# obtain dygraphs version…
v=$(sed -n '/^Dygraph.VERSION = "\(.*\)";$/s//\1/p' <src/dygraph.js)
if [[ -z $v ]]; then
	echo >&2 'E: could not determine version'
	exit 1
fi
# … as well as the version of the last release (for a snapshot)
if [[ $v = +([0-9]).+([0-9]).+([0-9])-* ]]; then
	relv=${v%%-*}
	IFS=.
	set -- $relv
	IFS=$' \t\n'
	relv=$1.$2.$(($3 - 1))
else
	relv=$v
fi

# build ES5- and browser-compatible code in a subdirectory
# and the code and test bundles as well as the minified CSS+JS

# copy/transform sources and licence info
rm -rf dist disttmp src-es5
mkdir disttmp
if [[ -e node_modules ]]; then
	pax -rw -l node_modules disttmp/
fi
pax -rw -l auto_tests src disttmp/

# licence headers for unminified and minified js, respectively
mksh scripts/txt2js.sh LICENSE.txt disttmp/LICENCE.js
header="/*! @license https://github.com/danvk/dygraphs/blob/v$relv/LICENSE.txt (MIT) */"

# prepare for building; avoid bad relative paths
cd disttmp
if [[ -e node_modules ]]; then
	PATH=$PWD/node_modules/.bin:$PATH
	export PATH
fi

# interpose browser-pack so we get the unminified prelude,
# as recently the minified file shipped will not work on
# PhantomJS or (other) older browsers
origbrowserpack=$(realpath "$($node_js -e 'process.stdout.write(require.resolve("browser-pack"));')")
origpreludefile=$(realpath "$origbrowserpack/../prelude.js")
mkdir .node_override
patchedbrowserpack=$PWD/.node_override/browser-pack.js
export origbrowserpack origpreludefile patchedbrowserpack
cat >"$patchedbrowserpack" <<\EOF
	const fs = require('fs');
	const orig = require(process.env.origbrowserpack);

	module.exports = function (opts) {
		if (!opts)
			opts = {};
		if (!opts.prelude && !opts.preludePath) {
			var pf = process.env.origpreludefile;
			var pc = fs.readFileSync(pf, 'utf8');
			pc = pc.trimEnd() + '/*}}}browser-pack*//*Dygraph(((*/';
			opts.preludePath = pf;
			opts.prelude = pc;
		}
		return orig(opts);
	}
EOF
export origbrowserify=$(realpath $(command -v browserify))
patchedbrowserify=$PWD/.node_override/browserify-patched.js
cat >"$patchedbrowserify" <<\EOF
	// monkey-patching require required 🙀
	const _Module = require('module');
	const _orig_require = _Module.prototype.require;
	_Module.prototype.require = function require(name) {
		if (typeof(name) === 'string' && name === 'browser-pack')
			return _orig_require(process.env.patchedbrowserpack);
		return _orig_require.apply(this, arguments);
	}
	require(process.env.origbrowserify);
EOF

# ES5-compatible source
$babel_js \
    --config-file "$babelrc" \
    --compact false \
    --source-maps inline \
    -d tests5 \
    auto_tests
$babel_js \
    --config-file "$babelrc" \
    --compact false \
    --source-maps inline \
    -d es5 \
    src
rm -rf auto_tests src

# get core-js from Debian, if needed, for XHR test
[[ -e node_modules ]] || ln -s /usr/share/nodejs/core-js .

# bundle dygraph.js{,.map} and tests.js with dev env
cp -r es5 src
mksh ../scripts/env-patcher.sh development src
$node_js "$patchedbrowserify" \
    -v \
    --debug \
    -p ../scripts/xfrmmodmap-dy.js \
    --full-paths \
    LICENCE.js \
    src/dygraph.js \
    >dygraph.tmp.js
$node_js "$patchedbrowserify" \
    -v \
    --debug \
    -p ../scripts/xfrmmodmap-t.js \
    --full-paths \
    tests5/tests/*.js \
    >tests.tmp.js
rm -rf src
python3 ../scripts/smap-out.py dygraph.tmp.js dygraph.js dygraph.js.map
python3 ../scripts/smap-out.py tests.tmp.js tests.tmp2.js tests.tmp.map
jq . <tests.tmp.map | perl -MCwd -pe \
    's!^ *"((?:\.\./)+)!Cwd::realpath($1) eq "/" ? "\"/" : $&!e;' \
    >tests.tmp2.map
python3 ../scripts/smap-in.py tests.tmp2.js tests.tmp2.map tests.js #--nonl

# bundle and minify dygraph.min.js{,.map} with prod env
cp -r es5 src
mksh ../scripts/env-patcher.sh production src
$node_js "$patchedbrowserify" \
    -v \
    --debug \
    -p ../scripts/xfrmmodmap-dy.js \
    --full-paths \
    src/dygraph.js \
    >dygraph.min.tmp.js
rm -rf src
python3 ../scripts/smap-out.py dygraph.min.tmp.js /dev/null dygraph.min.tmp.js.map

uglifyjs=$(uglifyjs --help 2>&1)
set -A compatopts -- --no-module --v8 --webkit
[[ $uglifyjs = *--no-module* ]] || unset compatopts[0]

uglifyjs \
    "${compatopts[@]}" \
    --compress \
    --mangle \
    --output-opts "preamble='$header'" \
    --source-map "content='dygraph.min.tmp.js.map',includeSources=true,url='dygraph.min.js.map'" \
    -o dygraph.min.js \
    dygraph.min.tmp.js

# copy out results
mkdir ../dist
mv dygraph.js dygraph.js.map dygraph.min.js dygraph.min.js.map tests.js ../dist/
mv es5 ../src-es5
cd ..
rm -rf disttmp

# minify CSS
cp css/dygraph.css dist/
cleancss css/dygraph.css -o dist/dygraph.min.css --source-map --source-map-inline-sources

# add (ES5-compatible) extras to dist
cd src-es5
pax -rw extras ../dist/
cd ..

#!/bin/mksh
set -exo pipefail

cd "$(dirname "$0")/.."

gitstatus=$(git status --porcelain)
if [[ -n $gitstatus ]]; then
	: repo unclean
	false
fi

curbranch=$(git branch --show-current)
[[ -n $curbranch ]]
curremote=$(git config --get "branch.$curbranch.remote")
[[ -n $curremote ]]
curremurl=$(git config --get "remote.$curremote.url")
[[ $curremurl = git@github.com:+([!\\\"])/+([!\\\"]).git ]]
shortrem=${curremurl%.git}
shortrem=${shortrem#git@github.com:}

curpkg=$(<package.json)
print -r -- "${curpkg%\}}" \
    ", \"repository\": { \"type\": \"git\", \"url\": \"https://github.com/$shortrem.git\" }" \
    ", \"name\": \"@$shortrem\"" \
    "}" | jq . >package.json

set +e
npm_config_registry=https://npm.pkg.github.com npm publish "$@"
rv=$?
git checkout package.json || rv=255
exit $rv

#!/bin/sh

sudoapt() {
	local _sudoapt_e=''

	if command -v eatmydata >/dev/null 2>&1; then
		_sudoapt_e=eatmydata
	fi
	sudo env DEBIAN_FRONTEND=noninteractive $_sudoapt_e apt-get "$@"
}
sudoagi='sudoapt --purge install --no-install-recommends -y'

set -ex

mksh -c true || {
	$sudoagi mksh
	exec mksh "$0" "$@"
	exit 255
}

export LC_ALL=C
unset LANGUAGE

set -o pipefail
rm -f .post.state
if test -h .post.state || test -e .post.state; then exit 255; fi
state=''

$sudoagi eatmydata
sudoapt clean
sudoapt update
#sudoapt --purge dist-upgrade -y
$sudoagi eatmydata git \
    ed jq node-jsdoc2 \
    libjs-bootstrap libjs-jquery libjs-jquery-ui \
    mksh pax python3

: drop any pre-installed PhantomJS as they cause breakage
bash -c 'set -o noglob; while true; do found=0; for x in $(which -a phantomjs); do test -e "$x" || continue; found=1; rm -f "$x"; done; test $found = 1 || break; done'

eatmydata env TMPDIR=/tmp npm install -g phantomjs@1.9.7-15
eatmydata env TMPDIR=/tmp npm install

(eatmydata npm run clean || :)
eatmydata npm run build
eatmydata npm run test || state="$state test"
eatmydata npm run test-min || state="$state test-min"
if [[ $GITHUB_REPOSITORY = danvk/dygraphs ]]; then
	eatmydata npm run coverage || state="$state coverage"
	eatmydata mksh scripts/post-coverage.sh || state="$state post-coverage"
	eatmydata mksh scripts/weigh-in.sh || state="$state weigh-in"
fi

rm -rf _site
mv site _site
cd _site
find . -type d -print0 | sort -z | xargs -0 mksh ../scripts/mkdiridx.sh
if [[ $GITHUB_REF = refs/heads/debian && $GITHUB_REPOSITORY = mirabilos/dygraphs ]]; then
	echo dygraph.github.mirsolutions.de >CNAME
	imprint_text='<a href="https://github.com/mirabilos/Impressum/tree/master/dygraphs#imprint-text" xml:lang="de-DE-1901"><b>Impressum</b> und Datenschutzerklärung</a>'
else
	imprint_text=
fi
if [[ -n $imprint_text ]]; then
	grep -FrlZ '@@@PLACE_IMPRINT_LINK_HERE_IF_NECESSARY@@@' . | \
	    xargs -0 perl -pi -e '
		s'\''<!--@@@IFIMPRINT:(.*?)@@@PLACE_IMPRINT_LINK_HERE_IF_NECESSARY@@@(.*?):FIIMPRINT@@@-->'\''$1.q{'"$imprint_text"'}.$2'\''eo
	    '
fi
cd ..
set -o noglob
echo $state >.post.state

#!/bin/mksh
# © 2022 mirabilos <m$(date +%Y)@mirbsd.de> Ⓕ MIT

set -eo pipefail
case $KSH_VERSION {
(*MIRBSD\ KSH*) ;;
(*) echo E: do not call me with bash or something; exit 255 ;;
}
mydir=$(realpath "$0/..")

infile=$1
outfile=$2
if [[ -z $infile || ! -s $infile || -z $outfile ]]; then
	print -ru2 "E: syntax error"
	exit 1
fi

if command -v nodejs >/dev/null 2>&1; then
	node_js=nodejs
else
	node_js=node
fi

rm -f "$outfile" "$outfile.tmp.js" "$outfile.tmp.js.map"
print -ru2 "I: converting $infile to $outfile"
$node_js "$mydir"/txt2js.js "$infile" "$outfile.tmp.js"
python3 "$mydir"/smap-in.py "$outfile.tmp.js" "$outfile.tmp.js.map" "$outfile" --nonl
rm -f "$outfile.tmp.js" "$outfile.tmp.js.map"

#!/bin/mksh
# Checks that no ".only" has made it into tests. This should never be committed,
# since it will disable the vast majority of tests.

if grep -R '\.only(' auto_tests/tests; then
  echo 'Remove .only from tests before committing.'
  exit 1
fi

exit 0

#!/bin/bash
# Continually regenerate these two files:
# - dist/dygraph.js
# - dist/tests.js
# As source/test files change.

# Clean background processes after quit
trap "trap - SIGTERM && kill -- -$$" SIGINT SIGTERM EXIT

mkdir -p dist  # in case it doesn't exist; watchify needs it

watchify \
  -v \
  -t babelify \
  -t [ envify --NODE_ENV development ] \
  --debug \
  --standalone Dygraph \
  -o dist/dygraph.js \
  src/dygraph.js &

watchify \
  -v \
  -t babelify \
  -t [ envify --NODE_ENV development ] \
  --debug \
  -o dist/tests.js \
  auto_tests/tests/*.js &

# Wait until background processes end
wait

#!/bin/mksh
set -ex
case $KSH_VERSION {
(*MIRBSD\ KSH*) ;;
(*) echo E: do not call me with bash or something; exit 255 ;;
}

# Build code, tests, browser bundles.
mksh scripts/build-js.sh

# Build documentation.
mksh scripts/build-docs.sh

# This is for on the webserver
rm -rf site _site
mkdir site
cd docroot
pax -rw . ../site/
set -A torm -- ../site/LICENSE.txt ../site/dist
[[ -n $IS_ACTUAL_DEBIAN_BUILD ]] || set -A torm+ -- ../site/.jslibs/*
rm "${torm[@]}"
[[ -n $IS_ACTUAL_DEBIAN_BUILD ]] || cp -L .jslibs/* ../site/.jslibs/
cd ..
pax -rw LICENSE.txt dist site/
rm -f site/dist/tests.js
find site -type d -print0 | xargs -0r chmod 0755 --
find site -type f -print0 | xargs -0r chmod 0644 --
set +ex
rv=0
x=$(find site \( ! -type d -a ! -type f -a ! -type l \) -ls) || {
	print -ru2 -- "E: could not check for bogus filetypes"
	rv=1
	x=
}
[[ -z $x ]] || {
	print -ru2 -- "E: bogus filetypes found"
	print -r -- "$x" | sed 's/^/N: /' >&2
	rv=1
}
if [[ $(find --help 2>&1) = *' -printf '* ]]; then
	x=$(find site -type l -printf '(%Y)%p\n') || {
		print -ru2 -- "E: could not check for dangling symlinks"
		rv=1
		x=
	}
	[[ -z $(print -r -- "$x" | grep -v '^[(][dfN][)]') ]] || {
		print -ru2 -- "E: bad filetypes found"
		print -r -- "$x" | sed 's/^/N: /' >&2
		rv=1
	}
	x=$(print -r -- "$x" | grep '^[(]N[)]')
	[[ -z $x ]] || {
		if [[ -n $IS_ACTUAL_DEBIAN_BUILD ]]; then
			pf=W:
			(( rv |= 2 ))
		else
			pf=E:
			rv=1
		fi
		print -ru2 -- "$pf dangling symlinks found"
		print -r -- "$x" | sed 's/^/N: /' >&2
	}
fi
if (( rv & 1 )); then
	exit 1
elif (( rv == 0 )); then
	print -ru2 -- "I: build done"
fi
exit 0

#!/bin/mksh
# Starts the http-server and runs mocha-phantomjs-based tests
# Note that you must run `npm run build-jsonly` or
# `npm run build` or `npm run watch` before running this.
# Additional arguments are passed to mocha-phantomjs, e.g.
# run-tests.sh --grep interaction-model
set -e
case $KSH_VERSION {
(*MIRBSD\ KSH*) ;;
(*) echo E: do not call me with bash or something; exit 255 ;;
}

# Run http-server and save its PID
http-server -p 8081 > /dev/null &
SERVER_PID=$!
function finish {
  kill -TERM $SERVER_PID
}
trap finish EXIT

# the following sleep step is not really necessary
# as http-server starts almost instantenously;
# but letting the server settle might help prevent
# possible racing conditions
sleep 1

if [[ $1 = min ]]; then
  shift
  fn=runner-min.html
else
  fn=runner.html
fi

# Start the tests
mocha-phantomjs http://localhost:8081/auto_tests/$fn "$@"

#!/bin/mksh
# This script generates the bundled JS files and pushes all content to a web site using rsync.

if [[ -n $1 ]]; then
  echo "usage: $0 destination"
  exit 1
fi

set -ex
site=$1

# Produce dist/*.js and docroot/ and all in site/
npm run build

# Copy everything to the site.
rsync -avzr site/ $site/

#!/bin/mksh
# Generate code coverage data for posting to Coveralls.
# This requires dist/*.js to be in place.
# Output is coverage/lcov.info

set -o errexit
if [[ -e node_modules ]]; then
	babel_js=babel
else
	babel_js=babeljs
fi
babelrc=$PWD/babel.config.json
set -x

rm -rf disttmp
mkdir disttmp

# Generate per-file ES6 --> ES5 transpilations
$babel_js \
  --config-file "$babelrc" \
  --retain-lines \
  --out-dir disttmp/src \
  src
$babel_js \
  --config-file "$babelrc" \
  --retain-lines \
  --out-dir disttmp/auto_tests/tests \
  auto_tests/tests

# Instrument the source code with Istanbul's __coverage__ variable.
rm -rf coverage  # Clear out everything to ensure a hermetic run.
mkdir coverage
istanbul instrument --output coverage/src disttmp/src
cp -r disttmp/auto_tests coverage/

# Build a combined file for running the tests in-browser
browserify coverage/auto_tests/tests/*.js -o coverage/tests.js

# Run http-server and save its PID for cleanup
http-server -p 8082 > /dev/null &
SERVER_PID=$!
function finish {
  kill -TERM $SERVER_PID
}
trap finish EXIT

# Give the server a chance to start up
sleep 1

# Run the tests using mocha-phantomjs & mocha-phantomjs-istanbul
# This produces coverage/coverage.json
phantomjs \
  ./node_modules/mocha-phantomjs/lib/mocha-phantomjs.coffee \
  http://localhost:8082/auto_tests/coverage.html \
  spec '{"hooks": "mocha-phantomjs-istanbul", "coverageFile": "coverage/coverage.json"}'

if [ $CI ]; then
  # Convert the JSON coverage to LCOV for coveralls.
  istanbul report --include coverage/*.json lcovonly

  # Monkey patch in the untransformed source paths.
  perl -i -pe 's,dist(?:tmp)?/,,' coverage/lcov.info
  echo ''  # reset exit code -- failure to post coverage shouldn't be an error.

else

  # Convert the JSON coverage to HTML for viewing
  istanbul report --include coverage/*.json html
  set +x

  echo 'To browse coverage, run:'
  echo
  echo '  open coverage/index.html'
  echo

fi

rm -rf disttmp

#!/bin/mksh
# From MirOS: www/mk/common,v 1.12 2021/12/11 20:10:49 tg Exp $'
#-
# Copyright © 2007, 2008, 2012, 2013, 2014, 2018, 2021, 2022
#	mirabilos <m$(date +%Y)@mirbsd.de>
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.

export LC_ALL=C
unset LANGUAGE
case $KSH_VERSION {
(*MIRBSD\ KSH*) ;;
(*) echo E: do not call me with bash or something; exit 255 ;;
}

fw=' <b class="extlink" title="WARNING: accesses external resources (Google jsapi)">⚠</b>'
ah='<a href="/">dygraphs JavaScript charting library</a> Pages'

# RFC 2396 and some optional characters _plus_ apostrophe
# -> escapes all shell meta-characters as well
function uri_escape {
	if (( $# )); then
		print -nr -- "$@"
	else
		cat
	fi | sed -e '
	    s.%.%25.g
	    s.;.%3B.g
	    s./.%2F.g
	    s.?.%3F.g
	    s.:.%3A.g
	    s.@.%40.g
	    s.&.%26.g
	    s.=.%3D.g
	    s.+.%2B.g
	    s.\$.%24.g
	    s.,.%2C.g
	    s.	.%09.g
	    s. .%20.g
	    s.<.%3C.g
	    s.>.%3E.g
	    s.#.%23.g
	    s.".%22.g
	    s.{.%7B.g
	    s.}.%7D.g
	    s.|.%7C.g
	    s.\\.%5C.g
	    s.\^.%5E.g
	    s.\[.%5B.g
	    s.\].%5D.g
	    s.`.%60.g
	    s.'\''.%27.g
	'
}

# escape XHTML characters (three mandatory XML ones plus double quotes,
# the latter in an XML safe fashion numerically though)
function xhtml_fesc {
	REPLY=${1//'&'/'&amp;'}
	REPLY=${REPLY//'<'/'&lt;'}
	REPLY=${REPLY//'>'/'&gt;'}
	REPLY=${REPLY//'"'/'&#34;'}
}

set -eo pipefail
print -ru2 "I: mkdiridx.sh beginning"

while [[ $1 = --* ]]; do
	if [[ $1 = --ah && -n $2 ]]; then
		ah=$2
		shift 2
	else
		break
	fi
done

trap 'print -ru2 "E: could not grep for extjs tests"; exit 1' USR1
(grep -FrlZ 'src="http' . || kill -USR1 $$) |&
set -A exttests
nexttests=0
while IFS= read -d '' -pr fn; do
	exttests[nexttests++]=$(realpath "$fn")
done
trap - USR1

owd=$PWD
for dir in "$@"; do
	cd "$owd"
	print -ru2 "I: processing $dir"
	cd "$dir"
	if [[ -s index.html ]]; then
		print -ru2 "I: already present, skipping"
		continue
	fi
	set -A subdirs
	nsubdirs=0
	set -A files
	set -A fwarns
	nfiles=0
	for de in *; do
		if [[ $de = index.html ]]; then
			print -ru2 "W: huh, index.html present but unusable?"
		elif [[ -d $de ]]; then
			subdirs[nsubdirs++]=$de
		elif [[ -e $de ]]; then
			da=$(realpath "$de")
			i=-1
			while (( ++i < nexttests )); do
				if [[ $da = "${exttests[i]}" ]]; then
					fwarns[nfiles]=1
					break
				fi
			done
			files[nfiles++]=$de
		else
			print -ru2 "W: does not exist: $de"
		fi
	done
	exec >index.html
	if [[ $dir = . ]]; then
		t='Index of /'
	else
		t="Index of /${dir#./}"
		t=${|xhtml_fesc "${t%%+(/)}/";}
	fi
	cat <<-EOF
		<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.1//EN"
		 "http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd">
		<html xmlns="http://www.w3.org/1999/xhtml" xml:lang="en"><head>
		 <meta http-equiv="Content-Type" content="text/html; charset=utf-8" />
		 <meta http-equiv="Content-Style-Type" content="text/css" />
		 <meta name="tdm-reservation" content="1" />
		 <meta name="MSSmartTagsPreventParsing" content="TRUE" />
		 <title>$t</title>
		</head><body>
		<h1>$t</h1>
		<ul>
		 <li><a href="../">Parent Directory</a></li>
	EOF
	i=-1
	while (( ++i < nsubdirs )); do
		d=${subdirs[i]}
		dl=$(uri_escape "$d")
		dh=${|xhtml_fesc "$d";}
		print -r -- " <li><a href=\"${dl}/\">${dh}/</a></li>"
	done
	i=-1
	while (( ++i < nfiles )); do
		f=${files[i]}
		fl=$(uri_escape "$f")
		fh=${|xhtml_fesc "$f";}
		print -r -- " <li><a href=\"${fl}\">${fh}</a>${fwarns[i]:+$fw}</li>"
	done
	print -r -- '</ul>'
	print -r -- '<!--@@@IFIMPRINT:<hr /><p>@@@PLACE_IMPRINT_LINK_HERE_IF_NECESSARY@@@</p>:FIIMPRINT@@@-->'
	print -r -- "<address>$ah</address>"
	print -r -- '</body></html>'
	exec >&2
done
print -ru2 "I: mkdiridx.sh finished"

#!/bin/mksh

# uncomment this to skip posting to coveralls.io
# comment the line out to post
#exit 0

if test -n "$CI"; then
  <coverage/lcov.info ./node_modules/.bin/coveralls || \
    echo "N: above errors posting to coveralls.io are ignored"
fi

true  # reset exit code -- failure to post coverage shouldn't be an error.

#!/bin/mksh
# This tracks the effect of pull requests on the size of dygraphs.
# See https://github.com/danvk/travis-weigh-in
set -e
case $KSH_VERSION {
(*MIRBSD\ KSH*) ;;
(*) echo E: do not call me with bash or something; exit 255 ;;
}

if [ -z "$GITHUB_TOKEN" ]; then
    echo "GITHUB_TOKEN not set. Skipping size checks."
    exit 0

else

  curl -O https://raw.githubusercontent.com/danvk/travis-weigh-in/master/weigh_in.py
  python weigh_in.py dist/dygraph.min.js
  mkdir -p disttmp
  gzip -cn9 <dist/dygraph.min.js >disttmp/dygraph.min.js.gz
  python weigh_in.py disttmp/dygraph.min.js.gz
  rm -r disttmp

fi

#!/bin/mksh
# © 2022, 2026 mirabilos <m$(date +%Y)@mirbsd.de> Ⓕ MIT

set -eo pipefail
case $KSH_VERSION {
(*MIRBSD\ KSH*) ;;
(*) echo E: do not call me with bash or something; exit 255 ;;
}
mydir=$(realpath "$0/..")

if [[ $1 = development ]]; then
	rpl=true
elif [[ $1 = production ]]; then
	rpl=false
else
	print -ru2 "$0: syntax error"
	exit 1
fi
shift

if command -v nodejs >/dev/null 2>&1; then
	node_js=nodejs
else
	node_js=node
fi

grep -FrlZ process.env.NODE_ENV "$@" | while IFS= read -d '' -r fn; do
	rm -f env-patcher.tmp*
	print -ru2 "I: patching $fn for !prod=$rpl"
	python3 "$mydir"/smap-out.py "$fn" env-patcher.tmp-in.js env-patcher.tmp-in.map
	print -r -- 'await require(process.env.s)(process.env.r);' | \
	    s="$mydir"/env-patcher.js r="$rpl" \
	    NODE_REPL_HISTORY= NODE_NO_READLINE=1 $node_js -i
	python3 "$mydir"/smap-in.py env-patcher.tmp-out.js env-patcher.tmp-out.map "$fn" --nonl
done
rm -f env-patcher.tmp*
print -ru2 "I: done patching"

#!/bin/sh

echo ::group::Check whether .pages.sh failed anywhere
state=$(cat .post.state || echo state-file-missing)
if test -n "$state"; then
	echo "E: .pages.sh failed in: $state"
	exit 1
fi
echo "I: nope, ok"
echo ::endgroup::

#!/bin/mksh
#
# Generates JSDoc in the /jsdoc dir. Clears any existing jsdoc there.

set -e
case $KSH_VERSION {
(*MIRBSD\ KSH*) ;;
(*) echo E: do not call me with bash or something; exit 255 ;;
}

v=$(sed -n '/^Dygraph.VERSION = "\(.*\)";$/s//\1/p' <src/dygraph.js)
test -n "$v" || {
  echo 'E: could not determine version'
  exit 1
}

if [[ -n $IS_ACTUAL_DEBIAN_BUILD ]]; then
  dv=$(dpkg-parsechangelog -S Version)
  v="$v (Debian $dv)"
elif [[ $GITHUB_REF = refs/heads/debian && $GITHUB_REPOSITORY = mirabilos/dygraphs ]]; then
  v="$v+WIP"
fi

rm -rf jsdoc jsdoc.tmp
mkdir jsdoc.tmp
t=$PWD/jsdoc.tmp
(cd /usr/share/nodejs/jsdoc2/templates/jsdoc && pax -rw . "$t/")
find jsdoc.tmp -type f -print0 | xargs -0r perl -pi -e \
  "s/\r+$//;" -e \
  "s! on [{][+]new Date[(][)][+][}]! for dygraph $v!g;"

echo Generating JSDoc...
srcfiles=src/dygraph.js
#srcfiles+=\ src/iframe-tarp.js
srcfiles+=\ src/dygraph-canvas.js
srcfiles+=\ src/dygraph-default-attrs.js
srcfiles+=\ src/dygraph-gviz.js
srcfiles+=\ src/dygraph-interaction-model.js
srcfiles+=\ src/dygraph-internal.externs.js
srcfiles+=\ src/dygraph-layout.js
srcfiles+=\ src/dygraph-options-reference.js
srcfiles+=\ src/dygraph-options.js
srcfiles+=\ src/dygraph-tickers.js
srcfiles+=\ src/dygraph-types.js
#srcfiles+=\ src/dygraph-utils.js
#srcfiles+=\ src/datahandler/datahandler.js
#srcfiles+=\ src/datahandler/default-fractions.js
#srcfiles+=\ src/datahandler/default.js
#srcfiles+=\ src/datahandler/bars.js
#srcfiles+=\ src/datahandler/bars-custom.js
#srcfiles+=\ src/datahandler/bars-error.js
#srcfiles+=\ src/datahandler/bars-fractions.js
srcfiles+=\ src/plugins/annotations.js
srcfiles+=\ src/plugins/axes.js
srcfiles+=\ src/plugins/chart-labels.js
srcfiles+=\ src/plugins/grid.js
#srcfiles+=\ src/plugins/legend.js
#srcfiles+=\ src/plugins/range-selector.js
#srcfiles+=\ src/extras/crosshair.js
#srcfiles+=\ src/extras/hairlines.js
#srcfiles+=\ src/extras/rebase.js
#srcfiles+=\ src/extras/shapes.js
#srcfiles+=\ src/extras/smooth-plotter.js
#srcfiles+=\ src/extras/super-annotations.js
#srcfiles+=\ src/extras/synchronizer.js
#srcfiles+=\ src/extras/unzoom.js
jsdoc2 \
  -t="$t" \
  -d=jsdoc \
  $srcfiles \
2>&1 | tee jsdoc.tmp/.errs

if test -s jsdoc.tmp/.errs; then errs=true; else errs=false; fi
rm -rf jsdoc.tmp

if $errs; then
  echo Please fix any jsdoc errors/warnings
  exit 1
fi

chmod -R a+rX jsdoc

echo Done

#!/bin/mksh
set -e
case $KSH_VERSION {
(*MIRBSD\ KSH*) ;;
(*) echo E: do not call me with bash or something; exit 255 ;;
}

v=$(sed -n '/^Dygraph.VERSION = "\(.*\)";$/s//\1/p' <src/dygraph.js)
test -n "$v" || {
  echo 'E: could not determine version'
  exit 1
}

if [[ -n $IS_ACTUAL_DEBIAN_BUILD ]]; then
  dv=$(dpkg-parsechangelog -S Version)
  dv="$dv ($v)"
else
  dv=
fi

rm -f docs/download.html docs/options.html docs/versions.html
python3 scripts/generate-download.py ${dv:+"$dv"} >docs/download.html
python3 scripts/generate-documentation.py >docs/options.html
mksh scripts/generate-versions.sh >docs/versions.html
chmod a+r docs/download.html docs/options.html docs/versions.html
for file in docs/download.html docs/options.html docs/versions.html; do
  test -s "$file" || {
    echo >&2 "E: generating $file failed"
    exit 1
  }
done

mksh scripts/generate-jsdoc.sh

rm -rf docroot
mkdir docroot
cd docs
pax -rw . ../docroot/
python3 ssi_expander.py ../docroot/
cd ../docroot
rm -f NOTES TODO common footer.html header.html *.py
mv README README-docs.txt
cd ..
rm -f docs/download.html docs/options.html docs/versions.html
pax -rw -l \
	common \
	gallery \
	jsdoc \
	tests \
	README.md \
	Screenshot.png \
	thumbnail.png \
    docroot/
rm -rf */__pycache__

#!/bin/mksh
# © 2023 mirabilos <t.glaser@tarent.de> Ⓕ MIT

set -eo pipefail

cat <<\EOF
<!--#set var="pagetitle" value="version history" -->
<!--#include virtual="header.html" -->

<h2>Version History</h2>

<p>For links to download each release, see the <a href="/download.html">Downloads</a> page.</p>

<table class="versions">
EOF

# could be a function converting */_ to b/i but we do not use it anywhere
alias out='print -rn --'

sed \
    -e 's!&!\&amp;!g' \
    -e 's!<!\&lt;!g' \
    -e 's!>!\&gt;!g' \
    -e 's!>!\&gt;!g' \
    -e 's!"!\&#34;!g' \
    -e 's!\[\([^]]*\)\](\(https*://[!-~]*\))!<a href="\2">\1</a>!g' \
    -e 's!\([ >]\)`\([^`]*\)`!\1<tt>\2</tt>!g' \
    <CHANGES.md |&
in_tr=0
in_ul=0
in_li=0
in_p=0
while IFS= read -pr line; do
	was_in_ul=$in_ul
	if (( in_li )); then
		if [[ $line = '  '* ]]; then
			print -r '<br />'
			out "         ${line##+( )}"
			continue
		fi
		print -r '</li>'
		in_li=0
	fi
	if (( in_ul )); then
		if [[ $line = '- '* ]]; then
			out "        <li>${line##-+( )}"
			in_li=1
			continue
		fi
		print -r '      </ul>'
		in_ul=0
	fi
	if [[ $line = '#'* ]]; then
		if [[ $line = '# '* ]]; then
			if (( in_tr )); then
				print -r '    </td>'
				print -r '  </tr>'
			fi
			line=${line#'# '}
			line=${line%')'}
			title=${line%' ('*}
			tag=$title
			if [[ $tag = v+([0-9.]) ]]; then
				title="<a href=\"#$tag\" title=\"link\">&#x1F517;</a>"$title
			else
				tag=
			fi
			print -r "  <tr${tag:+ id=\"$tag\"}>"
			print -r "    <td>$title<p class=\"date\">${line##*' ('}</p></td>"
			print -r '    <td class="notes">'
			in_tr=1
			continue
		fi
		if [[ $line = '## '* ]]; then
			print -r "      <h4>${line#'## '}</h4><ul>"
			in_ul=1
			continue
		fi
		print -ru2 "E: unsupported headline level"
		print -ru2 "N: $line"
		exit 1
	fi
	if [[ -z $line ]]; then
		if (( in_p )); then
			print -r '</p>'
			in_p=0
		fi
		# in_li/in_ul already handled above
		continue
	fi
	if [[ $line = '- '* ]]; then
		if (( in_p )); then
			print -r '</p>'
			in_p=0
			print -r '      <ul class="moveup">'
		else
			print -r '      <ul>'
		fi
		in_ul=1
		out "        <li>${line##-+( )}"
		in_li=1
		continue
	fi
	if (( in_p )); then
		print -r '<br />'
		print -nr '       '
	elif (( was_in_ul )); then
		print -nr '      <p class="moveup">'
		in_p=1
	else
		print -nr '      <p>'
		in_p=1
	fi
	out "$line"
done
if (( in_p )); then
	print -r '</p>'
fi
if (( in_li )); then
	print -r '</li>'
fi
if (( in_ul )); then
	print -r '      </ul>'
fi
cat <<\EOF
    </td>
  </tr>
</table>

<!--#include virtual="footer.html" -->
EOF

---
title: "Dygraphs in R Markdown"
output: html_document
---

```{r}
library(dygraphs)

hw <- HoltWinters(ldeaths)
p <- predict(hw, n.ahead = 36, prediction.interval = TRUE, level = 0.95)
ts <- cbind(ldeaths, p)

dygraph(ts, "Deaths from Lung Disease (UK)") %>%
  dySeries("ldeaths", label = "Deaths") %>%
  dySeries(c("p.lwr", "p.fit", "p.upr"), label = "Predicted")
```

Filing a bug report? Please include the following:

1. Link to a page which demonstrates the problem, preferably a jsfiddle (use http://dygraphs.com/fiddle as a template)
2. Browser and Operating System that exhibit the problem
3. Version of dygraphs that you're using

It also helps if you include the non-compacted version of the JS on your
page. For instance, instead of doing this:

```html
<script type="text/javascript" src="dygraph.min.js"></script>
```

do this:

```html
<script type="text/javascript" src="dygraph.js"></script>
```

This makes error messages and debugging simpler. The jsfiddle does this automatically.

Please read the guide to making dygraphs changes:
https://dygraphs.com/changes.html

Pull Requests will only be accepted if:

- You clearly explain what you're adding and why you believe it's an
  improvement. For example: "Fixes issue #123".
- You adhere to the style of the rest of the dygraphs code base.
- You write an `auto_test` for the code that you're adding.
- You certify that no “AI”/LLMs were used while preparing the PR.

Be sure to document any new options you add. Also be aware that PRs which add
options are likely to be rejected. dygraphs already has many options. If you
can fit your feature into one of those or implement it as a plugin, it will be
more likely to get merged.

Filing a bug report? Please include the following:

1. Link to a page which demonstrates the problem, preferably a jsfiddle (use [this jsfiddle](http://dygraphs.com/fiddle) as a template)
2. Browser and Operating System that exhibit the problem

It also helps if you include the non-compacted version of the JS on your
page. For instance, instead of doing this:

```html
<script type="text/javascript" src="dygraph.min.js"></script>
```

do this:

```html
<script type="text/javascript" src="dygraph.js"></script>
```

This makes error messages and debugging simpler. The jsfiddle does this automatically.

Sending a Pull Request? See our [guide to making dygraphs contributions](/DEVELOP.md).

Here's what it looks like:

![screen recording](smooth-plotter-p1-recording.gif)

This uses cubic splines (aka bezier curves) to smooth out the lines
between points. To draw a bezier curve between two points, you must also
specify two control points which specify how it bends:

![bezier_curve](smooth-plotter-p2-bezier.png)

The problem is to determine those control points.

When you connect a bunch of bezier curves, each original point in the
series has both a left control point and a right control point. To avoid
forming a kink in the chart, the left control point, the series point
and the right control point all need to fall on a line.

The algorithm I wound up using follows the HighCharts implementation of
this. Here's how it works...

The setup is that you have a point whose left & right controls you're
trying to determine. You also know the previous and next point:

![setup](smooth-plotter-p3-setup.png)

Start by placing the control points α of the way along each line
segment. This is the only parameter we'll need to control the amount of
smoothing:

![controls on line](smooth-plotter-p4-on-line.png)

But now we've broken the rule that the original point has to be on the
line between the control points! To fix that, we shift the control
points up by ∆y, which can be determined with some simple algebra:

![controls shifted](smooth-plotter-p5-shifted.png)

But this introduces a new problem: the right control point is above the
data points, and so it will create an impression of a false maximum to
the right of the data point. To fix this, we cap the y-value of the
control points:

![controls capped](smooth-plotter-p6-capped.png)

But now we've re-broken that rule about being on a line. So we have to
do the mirror-image adjustment to the left control point:

![controls capped and adjusted](smooth-plotter-p7-adjusted.png)

and now we have our control points!

Hopefully the algorithm makes good sense now. I implemented this as a
separate plotter in `/extras` for now, but it might make more sense to
move this into core dygraphs. That would make it easier to have the
smooth plots play nicely with other options like
`connectSeparatedPoints`, `strokePattern` and `drawPoints`.

This is intended to be a catch-all for legend formatting needs. This
[comes](https://stackoverflow.com/a/25090552/2171120)
[up](https://stackoverflow.com/a/24175648/2171120)
[often](https://stackoverflow.com/a/18211338/2171120) in Stack Overflow
[questions](https://stackoverflow.com/a/23354308/2171120) which can’t be
easily answered with `valueFormatter`. Nowadays I wave my hands and say
"write your own legend, either as a plugin or using `highlightCallback`
and `unhighlightCallback`." This is a simpler option.

The `legendFormatter` is repeatedly called with a `data` object
describing the selection or lack of selection. It contains all the
information you need to generate a standard legend (e.g. formatted
values), but there’s nothing preventing you from doing something crazier
on your own.

For example, here’s what a simple `legendFormatter` might look like:

```js
function legendFormatter(data) {
  if (data.x == null) return '';  // no selection
  return data.xHTML + data.series.map(v => v.labelHTML + ': ' + v.yHTML).join(' ');
}
```

Here’s what the `data` object looks like when nothing is selected:

```json
{
  "dygraph": "(dygraph)",
  "series": [
    {
      "dashHTML": "<div style=\"display: inline-block; position: relative; bottom: .5ex; padding-left: 1em; height: 1px; border-bottom: 2px solid rgb(0,128,0);\"></div>",
      "label": "Y1",
      "labelHTML": "Y1",
      "visible": true,
      "color": "rgb(0,128,0)"
    },
    {
      "dashHTML": "<div style=\"display: inline-block; position: relative; bottom: .5ex; padding-left: 1em; height: 1px; border-bottom: 2px solid rgb(0,0,128);\"></div>",
      "label": "Y2",
      "labelHTML": "Y2",
      "visible": true,
      "color": "rgb(0,0,128)"
    }
  ]
}
```

The `dashHTML` properties help you render lines which match each series’
line on the chart. When `strokePattern` is set, they become dotted or
dashed lines as needed.

Each value has a corresponding `HTML` variant, which is properly
formatted and escaped according to all the relevant options which have
been set for the chart.

Here’s what it looks like when a row is selected:

```json
{
  "dygraph": "(dygraph)",
  "x": 93,
  "xHTML": 93,
  "series": [
    {
      "dashHTML": "<div style=\"display: inline-block; position: relative; bottom: .5ex; padding-left: 1em; height: 1px; border-bottom: 2px solid rgb(0,128,0);\"></div>",
      "label": "Y1",
      "labelHTML": "Y1",
      "visible": true,
      "color": "rgb(0,128,0)",
      "y": 93,
      "yHTML": "93"
    },
    {
      "dashHTML": "<div style=\"display: inline-block; position: relative; bottom: .5ex; padding-left: 1em; height: 1px; border-bottom: 2px solid rgb(0,0,128);\"></div>",
      "label": "Y2",
      "labelHTML": "Y2",
      "visible": true,
      "color": "rgb(0,0,128)",
      "y": 26.04,
      "yHTML": "26.04"
    }
  ]
}
```

Here’s what it looks like when a single series is selected (e.g. with
`highlightSeriesOpts`):

```json
{
  "dygraph": "(dygraph)",
  "x": 94,
  "xHTML": 94,
  "series": [
    {
      "dashHTML": "<div style=\"display: inline-block; position: relative; bottom: .5ex; padding-left: 1em; height: 1px; border-bottom: 2px solid rgb(0,128,0);\"></div>",
      "label": "Y1",
      "labelHTML": "Y1",
      "visible": true,
      "color": "rgb(0,128,0)",
      "y": 94,
      "yHTML": "94",
      "isHighlighted": true
    },
    {
      "dashHTML": "<div style=\"display: inline-block; position: relative; bottom: .5ex; padding-left: 1em; height: 1px; border-bottom: 2px solid rgb(0,0,128);\"></div>",
      "label": "Y2",
      "labelHTML": "Y2",
      "visible": true,
      "color": "rgb(0,0,128)",
      "y": 22.56,
      "yHTML": "22.56"
    }
  ]
}
```

(Note the `isHighlighted` property set on `Y1`.)

## dygraphs developer notes

So you've made a change to dygraphs and would like to contribute it back to the open source project. Wonderful!

This is a step-by-step guide explaining how to do it.

### How-to

To install dependencies, run:

    # for building
    apt-get install jq mksh pax python3
    # npm dev dependencies
    npm install

To build dygraphs and the tests, run:

    npm run clean  # if necessary
    npm run build-jsonly

To run the tests, run:

    npm run test       # on the unminified bundle
    npm run test-min   # on the .min.{css,js}

The prerequisites for a full “npm run build” are:

    # for building
    apt-get install ed jsdoc-toolkit mksh pax python3
    # for docs
    apt-get install libjs-bootstrap libjs-jquery libjs-jquery-ui

To iterate on the code, install envify and run:

    npm run watch

and open `tests/demo.html` (or one of the other demos) in your browser.

To iterate on a unit test, run the `watch` command above and open

    auto_tests/runner.html

in your browser. You can use the Mocha UI to run just a single test or suite.
Or you can change `it` to `it.only` to do run just one test in code.

To run a single test from the command line, you can use:

  npm run test -- --grep highlight-series-background

(Note the extra `--`.)

### dygraphs style

When making a change, please try to follow the style of the existing dygraphs code. This will make the review process go much more smoothly.

A few salient points:

1. We try to adhere to Google's [JS style guide][gstyle] and would appreciate it if you try to as well. This means:
  *   No tabs! Indent using two spaces.
  *   Use camelCase for variable and function names.
  *   Limit lines to 80 characters.
1.  If you've added a new feature, add a test for it (in the tests/ directory) or a gallery entry.
1.  If you've added an option, document it in `dygraph-options-reference.js`. You'll get lots of warnings if you don't.
1.  If you've fixed a bug or added a feature, add a unit test (in `auto_tests`) for it.

Adding a unit test ensures that we won't inadvertently break your feature in the future. To do this, either add to an existing test in `auto_tests/tests` or create a new one.

### Sending a Pull Request

To make a change, you'll need to send a Pull Request. See GitHub's documentation [here][pr].

[gstyle]: https://google.github.io/styleguide/javascriptguide.xml
[pr]: http://help.github.com/send-pull-requests/

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/highlight.R
\name{dyHighlight}
\alias{dyHighlight}
\title{dygraph series mouse-over highlighting}
\usage{
dyHighlight(dygraph, highlightCircleSize = 3,
  highlightSeriesBackgroundAlpha = 0.5, highlightSeriesOpts = list(),
  hideOnMouseOut = TRUE)
}
\arguments{
\item{dygraph}{Dygraph to configure highlighting behavior for.}

\item{highlightCircleSize}{The size in pixels of the dot drawn over 
highlighted points.}

\item{highlightSeriesBackgroundAlpha}{Fade the background while highlighting 
series. 1=fully visible background (disable fading), 0=hidden background 
(show highlighted series only).}

\item{highlightSeriesOpts}{When set, the options from this list are applied 
to the series closest to the mouse pointer for interactive highlighting. 
Example: list(strokeWidth = 3). See the documentation on 
\code{\link{dySeries}} for additional details on options that can be set.}

\item{hideOnMouseOut}{Whether to hide the highlighting effects when the mouse
leaves the chart area. Note that this also affects the hiding of the 
\code{\link{dyLegend}} on mouse out.}
}
\value{
A dygraph with customized highlighting options
}
\description{
Configure options for data series mouse-over highlighting. Note that 
highlighting is always enabled for dygraphs so this function is used to 
customize rather than enable highlighting.
}
\note{
See the
\href{https://rstudio.github.io/dygraphs/gallery-series-highlighting.html}{online
documentation} for additional details and examples.
}
\examples{
library(dygraphs)
lungDeaths <- cbind(ldeaths, mdeaths, fdeaths)
dygraph(lungDeaths, main = "Deaths from Lung Disease (UK)") \%>\%
  dyHighlight(highlightCircleSize = 5, 
              highlightSeriesBackgroundAlpha = 0.2,
              hideOnMouseOut = FALSE)   
 
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/event.R
\name{dyLimit}
\alias{dyLimit}
\title{dygraph limit line}
\usage{
dyLimit(dygraph, limit, label = NULL, labelLoc = c("left", "right"),
  color = "black", strokePattern = "dashed")
}
\arguments{
\item{dygraph}{Dygraph to add limit line to}

\item{limit}{Numeric position of limit.}

\item{label}{Label for limit. Defaults to blank.}

\item{labelLoc}{Location for label (left or right).}

\item{color}{Color of limit line. This can be of the form "#AABBCC" or 
"rgb(255,100,200)" or "yellow". Defaults to black.}

\item{strokePattern}{A predefined stroke pattern type ("dotted", "dashed",
"dotdash", or "solid") or a custom pattern array where the even index is 
a draw and odd is a space in pixels. Defaults to dashed.}
}
\value{
A dygraph with the specified limit line.
}
\description{
Add a horizontal limit line to a dygraph
}
\note{
See the \href{https://rstudio.github.io/dygraphs/gallery-event-lines.html}{online 
  documentation} for additional details and examples.
}
\examples{
library(dygraphs)

dygraph(presidents, main = "Presidential Approval") \%>\%
  dyAxis("y", valueRange = c(0, 100)) \%>\% 
  dyLimit(max(presidents, na.rm = TRUE), "Max",
          strokePattern = "solid", color = "blue")
 
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/exports.R
\name{dygraphs-exports}
\alias{dygraphs-exports}
\alias{\%>\%}
\alias{as.yearmon}
\alias{as.yearqtr}
\title{dygraph exported operators and S3 methods}
\description{
The following functions are imported and then re-exported
from the dygraphs package to avoid listing the magrittr and
zoo packages as Depends of dygraphs.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plugins.R
\name{dyRibbon}
\alias{dyRibbon}
\title{dyRibbon plugin adds a horizontal band of colors that runs through the chart. It
can be useful to visualize categorical variables
(http://en.wikipedia.org/wiki/Categorical_variable) that change over time (along
the x-axis).}
\usage{
dyRibbon(dygraph, data = NULL, palette = NULL, parser = NULL, top = 1,
  bottom = 0)
}
\arguments{
\item{dygraph}{Dygraph to add plugin to}

\item{data}{Vector of numeric values in the range from 0 to 1}

\item{palette}{Vector with colors palette}

\item{parser}{JavaScrip function (function (data, dygraph)) returning the array of
numeric values. Parser is used if no data was provided}

\item{top}{Vertical position of the top edge of ribbon relative to chart height.}

\item{bottom}{Vertical position of the bottom edge of ribbon relative to chart height.}
}
\description{
dyRibbon plugin adds a horizontal band of colors that runs through the chart. It
can be useful to visualize categorical variables
(http://en.wikipedia.org/wiki/Categorical_variable) that change over time (along
the x-axis).
}
\examples{
\dontrun{
library(quantmod)

getSymbols("SPY", from = "2016-12-01", auto.assign=TRUE)

difference <- SPY[, "SPY.Open"] - SPY[, "SPY.Close"]
decreasing <- which(difference < 0)
increasing <- which(difference > 0)

dyData <- SPY[, "SPY.Close"]

ribbonData <- rep(0, nrow(dyData))
ribbonData[decreasing] <- 0.5
ribbonData[increasing] <- 1

dygraph(dyData) \%>\%
    dyRibbon(data = ribbonData, top = 0.1, bottom = 0.02)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/callbacks.R
\name{dyCallbacks}
\alias{dyCallbacks}
\title{Callbacks for dygraph events}
\usage{
dyCallbacks(dygraph, clickCallback = NULL, drawCallback = NULL,
  highlightCallback = NULL, pointClickCallback = NULL,
  underlayCallback = NULL, unhighlightCallback = NULL,
  zoomCallback = NULL, drawHighlightPointCallback = NULL,
  drawPointCallback = NULL, annotationClickHandler = NULL,
  annotationMouseOverHandler = NULL, annotationMouseOutHandler = NULL,
  annotationDblClickHandler = NULL)
}
\arguments{
\item{dygraph}{Dygraph to add callbacks to}

\item{clickCallback}{A function to call when the canvas is clicked.}

\item{drawCallback}{When set, this callback gets called every time the 
dygraph is drawn. This includes the initial draw, after zooming and 
repeatedly while panning.}

\item{highlightCallback}{When set, this callback gets called every time a new
point is highlighted.}

\item{pointClickCallback}{A function to call when a data point is clicked. 
and the point that was clicked.}

\item{underlayCallback}{When set, this callback gets called before the chart 
is drawn.}

\item{unhighlightCallback}{When set, this callback gets called every time the
user stops highlighting any point by mousing out of the graph.}

\item{zoomCallback}{A function to call when the zoom window is changed 
(either by zooming in or out).}

\item{drawHighlightPointCallback}{Draw a custom item when a point is 
highlighted. Default is a small dot matching the series color. This method 
should constrain drawing to within pointSize pixels from (cx, cy)}

\item{drawPointCallback}{Draw a custom item when drawPoints is enabled. 
Default is a small dot matching the series color. This method should 
constrain drawing to within pointSize pixels from (cx, cy).}

\item{annotationClickHandler}{JavaScript function to call when an annotation
is clicked. This can also be specified on a per-annotation basis.}

\item{annotationMouseOverHandler}{JavaScript function to call when the mouse
hovers over an annotation. This can also be specified on a per-annotation
basis.}

\item{annotationMouseOutHandler}{JavaScript function to call when the mouse
exits an annotation. This can also be specified on a per-annotation basis.}

\item{annotationDblClickHandler}{JavaScript function to call when an
annotation is double clicked. This can also be specified on a
per-annotation basis.}
}
\value{
Dygraph with callbacks
}
\description{
Set JavaScript callbacks for various dygraph events. See the 
\href{http://dygraphs.com/options.html}{dygraph options} reference for 
additional details on the signature of each callback.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shading.R
\name{dyShading}
\alias{dyShading}
\title{dygraph region shading}
\usage{
dyShading(dygraph, from, to, color = "#EFEFEF", axis = "x")
}
\arguments{
\item{dygraph}{Dygraph to add shading to}

\item{from}{Date/time or numeric to shade from (for date/time this must be a
\code{as.POSIXct} object or another object convertible via
\code{as.POSIXct}).}

\item{to}{Date/time or numeric to shade to (for date/time this must be a
\code{as.POSIXct} object or another object convertible via
\code{as.POSIXct}).}

\item{color}{Color of shading. This can be of the form "#AABBCC" or 
"rgb(255,100,200)" or "yellow". Defaults to a very light gray.}

\item{axis}{Axis to apply shading.  Choices are "x" or "y".}
}
\value{
A dygraph with the specified shading
}
\description{
Specify that a region of a dygraph be drawn with a background shading
}
\note{
See the
  \href{https://rstudio.github.io/dygraphs/gallery-annotations.html}{online
  documentation} for additional details and examples.
}
\examples{
library(dygraphs)

dygraph(nhtemp, main = "New Haven Temperatures") \%>\% 
  dyShading(from = "1920-1-1", to = "1930-1-1") \%>\%
  dyShading(from = "1940-1-1", to = "1950-1-1")
  
dygraph(nhtemp, main = "New Haven Temperatures") \%>\% 
  dyShading(from = "48", to = "52", axis = "y") \%>\%
  dyShading(from = "50", to = "50.1", axis = "y", color = "black")
    
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/axis.R
\name{dyAxis}
\alias{dyAxis}
\title{dygraph axis}
\usage{
dyAxis(dygraph, name, label = NULL, valueRange = NULL, logscale = NULL,
  ticker = NULL, rangePad = NULL, labelWidth = NULL, labelHeight = NULL,
  axisHeight = NULL, axisLineColor = NULL, axisLineWidth = NULL,
  pixelsPerLabel = NULL, axisLabelColor = NULL, axisLabelFontSize = NULL,
  axisLabelWidth = NULL, axisLabelFormatter = NULL, valueFormatter = NULL,
  drawGrid = NULL, gridLineColor = NULL, gridLineWidth = NULL,
  independentTicks = NULL)
}
\arguments{
\item{dygraph}{Dygraph to add an axis definition to}

\item{name}{Axis name ('x', 'y', or 'y2')}

\item{label}{Label to display for axis (defaults to none).}

\item{valueRange}{Explicitly set the vertical range of the graph to 
\code{c(low, high)}. This may be set on a per-axis basis to define each 
y-axis separately. If either limit is unspecified, it will be calculated 
automatically (e.g. \code{c(NULL, 30)} to automatically calculate just the 
lower bound).}

\item{logscale}{When set for the y-axis or x-axis, the graph shows that axis
in log scale. Any values less than or equal to zero are not displayed.
Showing log scale with ranges that go below zero will result in an
unviewable graph. Not compatible with showZero. connectSeparatedPoints is
ignored. This is ignored for date-based x-axes.}

\item{ticker}{This lets you specify an arbitrary JavaScript function to 
generate tick marks on an axis. The tick marks are an array of (value, 
label) pairs. The built-in functions go to great lengths to choose good 
tick marks so, if you set this option, you'll most likely want to call one 
of them and modify the result. See dygraph-tickers.js and the 
\href{http://dygraphs.com/options.html}{dygraphs documentation} for 
additional details).}

\item{rangePad}{Add the specified amount of extra space (in pixels) around 
the value range to ensure points at the edges remain visible.}

\item{labelWidth}{Width of the div which contains the y-axis label. Since the
y-axis label appears rotated 90 degrees, this actually affects the height 
of its div.}

\item{labelHeight}{Height of the x-axis label, in pixels. This also controls 
the default font size of the x-axis label. If you style the label on your 
own, this controls how much space is set aside below the chart for the 
x-axis label's div.}

\item{axisHeight}{Height, in pixels, of the x-axis. If not set explicitly, 
this is computed based on \code{axisLabelFontSize} and \code{axisTickSize}.}

\item{axisLineColor}{Color of the x- and y-axis lines. Accepts any value 
which the HTML canvas strokeStyle attribute understands, e.g. 'black' or 
'rgb(0, 100, 255)'.}

\item{axisLineWidth}{Thickness (in pixels) of the x- and y-axis lines.}

\item{pixelsPerLabel}{Number of pixels to require between each x- and 
y-label. Larger values will yield a sparser axis with fewer ticks. Defaults
to 50 (x-axis) or 30 (y-axes).}

\item{axisLabelColor}{Color for x- and y-axis labels. This is a CSS color 
string. This may also be set globally using \code{dyOptions}.}

\item{axisLabelFontSize}{Size of the font (in pixels) to use in the axis 
labels, both x- and y-axis. This may also be set globally using 
\code{dyOptions}.}

\item{axisLabelWidth}{Width, in pixels, of the axis labels}

\item{axisLabelFormatter}{JavaScript function to call to format the tick 
values that appear along an axis (see the 
\href{http://dygraphs.com/options.html}{dygraphs documentation} for 
additional details).}

\item{valueFormatter}{JavaScript function to call to provide a custom display
format for the values displayed on mouseover (see the 
\href{http://dygraphs.com/options.html}{dygraphs documentation} for 
additional details).}

\item{drawGrid}{Whether to display grid lines in the chart.}

\item{gridLineColor}{The color of the grid lines.}

\item{gridLineWidth}{Thickness (in pixels) of the grid lines drawn under the 
chart.}

\item{independentTicks}{Only valid for y and y2, has no effect on x: This 
option defines whether the y axes should align their ticks or if they 
should be independent. Possible combinations: 1.) y=true, y2=false 
(default): y is the primary axis and the y2 ticks are aligned to the the 
ones of y. (only 1 grid) 2.) y=false, y2=true: y2 is the primary axis and 
the y ticks are aligned to the the ones of y2. (only 1 grid) 3.) y=true, 
y2=true: Both axis are independent and have their own ticks. (2 grids) 4.) 
y=false, y2=false: Invalid configuration causes an error.}
}
\value{
Axis options
}
\description{
Define options for an axis on a dygraph plot. Note that options will use the 
default global setting (as determined by \code{\link{dyOptions}}) when not 
specified explicitly.
}
\note{
See the 
  \href{https://rstudio.github.io/dygraphs/gallery-axis-options.html}{online 
  documentation} for additional details and examples.
}
\examples{
library(dygraphs)

dygraph(nhtemp, main = "New Haven Temperatures") \%>\%
  dyAxis("y", label = "Temp (F)", valueRange = c(40, 60)) \%>\%
  dyOptions(axisLineWidth = 1.5, fillGraph = TRUE, drawGrid = FALSE)
    
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/roller.R
\name{dyRoller}
\alias{dyRoller}
\title{dygraph rolling average period text box}
\usage{
dyRoller(dygraph, showRoller = TRUE, rollPeriod = 1)
}
\arguments{
\item{dygraph}{Dygraph to add roller to}

\item{showRoller}{Whether to show the roller}

\item{rollPeriod}{Number of time scale units (e.g. days, months, years) to 
average values over.}
}
\value{
A dygraph that displays a range selector
}
\description{
Add a rolling average period text box to the bottom left of the plot. Y 
values are averaged over the specified number of time scale units.
}
\note{
See the
\href{https://rstudio.github.io/dygraphs/gallery-roll-periods.html}{online
documentation} for additional details and examples.
}
\examples{
library(dygraphs)

dygraph(discoveries, main = "Important Discoveries") \%>\% 
  dyRoller(rollPeriod = 5)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dygraph.R
\name{dygraph}
\alias{dygraph}
\title{dygraph interactive plot for time series data}
\usage{
dygraph(data, main = NULL, xlab = NULL, ylab = NULL, periodicity = NULL,
  group = NULL, elementId = NULL, width = NULL, height = NULL)
}
\arguments{
\item{data}{Either time series data or numeric data. For time series, this 
must be an \link[xts]{xts} object or an object which is convertible to 
\code{xts}. For numeric data, this must be a named list or data frame,
where the first element/column provides x-axis values and all subsequent
elements/columns provide one or more series of y-values.}

\item{main}{Main plot title (optional)}

\item{xlab}{X axis label}

\item{ylab}{Y axis label}

\item{periodicity}{Periodicity of time series data (automatically detected
via \link[xts:periodicity]{xts::periodicity} if not specified).}

\item{group}{Group to associate this plot with. The x-axis zoom level of 
plots within a group is automatically synchronized.}

\item{elementId}{Use an explicit element ID for the widget (rather than an
automatically generated one). Useful if you have other JavaScript that
needs to explicitly discover and interact with a specific widget instance.}

\item{width}{Width in pixels (optional, defaults to automatic sizing)}

\item{height}{Height in pixels (optional, defaults to automatic sizing)}
}
\value{
Interactive dygraph plot
}
\description{
R interface to interactive time series plotting using the 
\href{http://dygraphs.com}{dygraphs} JavaScript library.
}
\note{
See the \href{https://rstudio.github.io/dygraphs/}{online documentation} for
additional details and examples.
}
\examples{
library(dygraphs)
lungDeaths <- cbind(mdeaths, fdeaths)
dygraph(lungDeaths)

indoConc <- Indometh[Indometh$Subject == 1, c("time", "conc")]
dygraph(indoConc)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plugins.R
\name{dyCrosshair}
\alias{dyCrosshair}
\title{The dyCrosshair plugin draws a crosshair line over the point closest to the
mouse when the user hovers over the graph. It has a "direction" option which
is provided in the R wrapper function and then forwarded to the plugin using
the "options" argument to dyPlugin.}
\usage{
dyCrosshair(dygraph, direction = c("both", "horizontal", "vertical"))
}
\arguments{
\item{dygraph}{Dygraph to add plugin to}

\item{direction}{Crosshair direction. Valid options are: "both", "horizontal",
"vertical"}
}
\value{
Dygraph with Crosshair plugin enabled
}
\description{
The dyCrosshair plugin draws a crosshair line over the point closest to the
mouse when the user hovers over the graph. It has a "direction" option which
is provided in the R wrapper function and then forwarded to the plugin using
the "options" argument to dyPlugin.
}
\examples{
library(dygraphs)
dygraph(ldeaths) \%>\%
  dyRangeSelector() \%>\%
  dyCrosshair(direction = "vertical")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dependency.R
\name{dyPlugin}
\alias{dyPlugin}
\title{Include a dygraph plugin}
\usage{
dyPlugin(dygraph, name, path, options = list(), version = "1.0")
}
\arguments{
\item{dygraph}{Dygraph to add plugin to}

\item{name}{Name of plugin}

\item{path}{Path to plugin JavaScript file}

\item{options}{Named list of options to pass to plugin constructor}

\item{version}{Plugin version (e.g. version of package which provides the plugin)}
}
\value{
A dygraph with the specified plugin enabled.
}
\description{
Include a dygraph plugin
}
\details{
You can use dygraphs plugins to customize the appearance of dygraphs as well
as add new interactive behaviors. For additional information on creating
dygraphs plugins see the
\href{https://github.com/danvk/dygraphs/tree/master/src/plugins}{dygraphs plugins}
documentation.
Once you've created a dygraphs plugins you can use the dyPlugin function to
create an R wrapper for it. See
\href{https://rstudio.github.io/dygraphs/gallery-plugins.html}{https://rstudio.github.io/dygraphs/gallery-plugins.html}
for details on how to do this.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/range-selector.R
\name{dyRangeSelector}
\alias{dyRangeSelector}
\title{dygraph interactive range selection and zooming}
\usage{
dyRangeSelector(dygraph, dateWindow = NULL, height = 40,
  fillColor = " #A7B1C4", strokeColor = "#808FAB", keepMouseZoom = TRUE,
  retainDateWindow = FALSE)
}
\arguments{
\item{dygraph}{Dygraph to add range selector to}

\item{dateWindow}{Initially zoom in on a section of the graph. Is a two 
element vector [earliest, latest], where earliest/latest objects 
convertible via \code{as.POSIXct}.}

\item{height}{Height, in pixels, of the range selector widget.}

\item{fillColor}{The range selector mini plot fill color. This can be of the 
form "#AABBCC" or "rgb(255,100,200)" or "yellow". You can also specify ""
to turn off fill.}

\item{strokeColor}{The range selector mini plot stroke color. This can be of 
the form "#AABBCC" or "rgb(255,100,200)" or "yellow". You can also specify 
"" to turn off stroke.}

\item{keepMouseZoom}{Keep mouse zoom when adding a range selector}

\item{retainDateWindow}{Whether to retain the user's current date window 
(zoom level) when updating an existing dygraph with new data and/or 
options.}
}
\value{
A dygraph that displays a range selector
}
\description{
Add a range selector to the bottom of the chart that allows users to pan and 
zoom to various date ranges.
}
\note{
See the
\href{https://rstudio.github.io/dygraphs/gallery-range-selector.html}{online
documentation} for additional details and examples.

Shiny applications can respond to changes in the \code{dateWindow}
via a special date window shiny input value. For example, if the 
output id of a dygraph is `series` then the current date window
can be read from \code{input$series_date_window} as an array of 
two date values (from and to).
}
\examples{
library(dygraphs)

dygraph(nhtemp, main = "New Haven Temperatures") \%>\% 
  dyRangeSelector()   
 
dygraph(nhtemp, main = "New Haven Temperatures") \%>\% 
  dyRangeSelector(dateWindow = c("1920-01-01", "1960-01-01"))   
    
dygraph(nhtemp, main = "New Haven Temperatures") \%>\% 
  dyRangeSelector(height = 20, strokeColor = "")     

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/series.R
\name{dySeriesData}
\alias{dySeriesData}
\title{Add series data to dygraph}
\usage{
dySeriesData(dygraph, name, values)
}
\arguments{
\item{dygraph}{Dygraph to add data to}

\item{name}{Name of series}

\item{values}{Data values}
}
\value{
Dygraph with additional series data
}
\description{
Add an additional column of series data to a dygraph. This
is typically used in the construction of custom series types
(e.g. log scaled, smoothed, etc.)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/candlestick.R, R/plotters.R
\name{dyCandlestick}
\alias{dyCandlestick}
\alias{Plotters}
\alias{dyBarChart}
\alias{dyStackedBarChart}
\alias{dyMultiColumn}
\alias{dyBarSeries}
\alias{dyStemSeries}
\alias{dyShadow}
\alias{dyFilledLine}
\alias{dyErrorFill}
\alias{dyMultiColumnGroup}
\alias{dyCandlestickGroup}
\alias{dyStackedBarGroup}
\alias{dyStackedLineGroup}
\alias{dyStackedRibbonGroup}
\title{Employ a dygraph plotter on a series, a group of series, or the whole dygraph}
\usage{
dyCandlestick(dygraph, compress = FALSE)

dyBarChart(dygraph)

dyStackedBarChart(dygraph)

dyMultiColumn(dygraph)

dyBarSeries(dygraph, name, ...)

dyStemSeries(dygraph, name, ...)

dyShadow(dygraph, name, ...)

dyFilledLine(dygraph, name, ...)

dyErrorFill(dygraph, name, ...)

dyMultiColumnGroup(dygraph, name, ...)

dyCandlestickGroup(dygraph, name, ...)

dyStackedBarGroup(dygraph, name, ...)

dyStackedLineGroup(dygraph, name, ...)

dyStackedRibbonGroup(dygraph, name, ...)
}
\arguments{
\item{dygraph}{Dygraph to add plotter to}

\item{compress}{(For dyCandlestick) If true, compress data yearly, quarterly, monthly, weekly or daily
according to overall amount of bars and/or current zoom level.}

\item{name}{name - or chrarcter vector of names - of (the) series within the data set}

\item{...}{additional options to pass to dySeries}
}
\value{
A dygraph with the specified plotter(s) employed.
}
\description{
Plotters provide variuos ways to customize how your data appears
on the dygraph.  Series-based plotters allow users to mix-and-match different plotters on a 
per-series or (with dyGroup) a per-group basis.  See \code{\link{dyPlotter}} for additional detail.
}
\section{Available plotters}{



Currently the dygraphs package provides the following plotters:

\describe{
   \item{dyBarChart()}{Draws a bar plot rather than a line plot. If the provided dygraph
features more than one series, dyBarChart will call dyMultiColumn instead.}

   \item{dyStackedBarChart()}{Draws a bar chart stacking all the underlying series.}

   \item{dyMultiColumn()}{Draws multiple column bar chart.} 
   \item{dyBarSeries()}{Draws a single set of bars for just the provided series.}

   \item{dyStemSeries()}{Draws a single set of stems for just the provided series.}

   \item{dyShadow()}{An extraction of the _fillplotter from dygraph-combined-dev.js, 
   drawing the filled area without the line.}

   \item{dyFilledLIne()}{An extraction of the _fillplotter and _lineplotter combo 
   from dygraph-combined-dev.js. dyFilledLine allows users to fill only a single series.}

   \item{dyMultiColumnGroup()}{The multicolumn plotter, but on a subset of the series, leaving 
   the others for other plotters.}

   \item{dyCandlestick()}{Draw a candlestick chart.}

   \item{dyCandleStickGroup()}{Employed on the provided series, 
   but still plotting the others.}
   
   \item{dyStackerBarGroup()}{Return the data group as stacked bars}

   \item{dyStackerRibbonGroup()}{Return the data group as stacked ribbons}

}
}

\examples{
## The following two examples will results in the same dygraph:

dygraph(mdeaths) \%>\% 
  dyBarChart()
 
lungDeaths <- cbind(mdeaths, fdeaths)
dygraph(lungDeaths) \%>\%
  dyMultiColumn()
  

## Per-series plotters:

lungDeaths <- cbind(mdeaths, fdeaths)
dygraph(lungDeaths) \%>\%
  dyBarSeries('fdeaths')

lungDeaths <- cbind(mdeaths, fdeaths)
dygraph(lungDeaths) \%>\%
  dyStemSeries('fdeaths')
  
lungDeaths <- cbind(mdeaths, fdeaths)
dygraph(lungDeaths) \%>\%
  dyShadow('fdeaths')

lungDeaths <- cbind(mdeaths, fdeaths)
dygraph(lungDeaths) \%>\%
  dyFilledLine('fdeaths')

## A bunch of different plotters together:

lungDeaths <- cbind(fdeaths, mdeaths, ldeaths, foo = fdeaths/2, bar = fdeaths/3)
dygraph(lungDeaths) \%>\%
  dyRangeSelector() \%>\%
  dyBarSeries('bar') \%>\% 
  dyStemSeries('mdeaths') \%>\% 
  dyShadow('foo') \%>\% 
  dyFilledLine('fdeaths')

## Group-based plotters:
  

## Candlestick plotters:
  
library(xts)
data(sample_matrix)
library(dygraphs)
dygraph(sample_matrix) \%>\%
  dyCandlestick()
  
sample<-data.frame(sample_matrix)
sample_2<-sample*2
names(sample_2)<-c('O', 'H', 'L', 'C')
sample<-cbind(sample, sample_2)
dygraph(sample) \%>\% 
  dyOptions(stackedGraph = TRUE) \%>\% 
  dyCandlestickGroup(c('Open', 'High', 'Low', 'Close')) \%>\% 
  dyCandlestickGroup(c('O', 'H', 'L', 'C'))

## Stacked Bar and Ribbon Graphs:  

dygraph(lungDeaths) \%>\% 
  dySeries('mdeaths', axis = 'y2') \%>\%
  dyAxis('y', valueRange = c(-100, 1000)) \%>\% 
  dyStackedBarGroup(c('ldeaths', 'fdeaths'))
  
lungDeaths <- cbind(ldeaths, fdeaths, mdeaths, 
                    additive = rep.int(200, length(ldeaths)),
                    line = rep.int(3000, length(ldeaths)))
dygraph(lungDeaths) \%>\% 
  dySeries('line', strokePattern = 'dashed') \%>\% 
  dySeries('ldeaths', stepPlot = TRUE) \%>\% 
  dyStackedBarGroup(c('additive', 'mdeaths')) \%>\% 
  dyStackedRibbonGroup(c('fdeaths', 'ldeaths'))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/group.R
\name{dyGroup}
\alias{dyGroup}
\title{dygraph series group}
\usage{
dyGroup(dygraph, name = NULL, label = NULL, color = NULL, axis = "y",
  stepPlot = NULL, stemPlot = NULL, fillGraph = NULL, drawPoints = NULL,
  pointSize = NULL, pointShape = NULL, strokeWidth = NULL,
  strokePattern = NULL, strokeBorderWidth = NULL,
  strokeBorderColor = NULL, plotter = NULL)
}
\arguments{
\item{dygraph}{Dygraph to add a series definition to}

\item{name}{character vector of the series within data set. If no name is specified then 
series are bound to implicitly based on their order within the underlying 
time series object. This parameter can also be a character vector of length
3 that specifies a set of input column names to use as the lower, value,
and upper for a series with a shaded bar drawn around it.}

\item{label}{Labels to display for series (uses name if no label defined)}

\item{color}{Color for series. These can be of the form "#AABBCC" or 
"rgb(255,100,200)" or "yellow", etc. If not specified then the global 
colors option (typically based on equally-spaced points around a color wheel).}

\item{axis}{Y-axis to associate the series with ("y" or "y2")}

\item{stepPlot}{When set, display the graph as a step plot instead of a line 
plot.}

\item{stemPlot}{When set, display the graph as a stem plot instead of a line
plot.}

\item{fillGraph}{Should the area underneath the graph be filled? This option 
is not compatible with error bars.}

\item{drawPoints}{Draw a small dot at each point, in addition to a line going
through the point. This makes the individual data points easier to see, but
can increase visual clutter in the chart.}

\item{pointSize}{The size of the dot to draw on each point in pixels. A dot 
is always drawn when a point is "isolated", i.e. there is a missing point 
on either side of it. This also controls the size of those dots.}

\item{pointShape}{The shape of the dot to draw. Can be one of the following:
"dot" (default), "triangle", "square", "diamond", "pentagon", "hexagon",
"circle", "star", "plus" or "ex".}

\item{strokeWidth}{The width of the lines connecting data points. This can be
used to increase the contrast or some graphs.}

\item{strokePattern}{A predefined stroke pattern type ("dotted", "dashed", or
"dotdash") or a custom pattern array where the even index is a draw and odd
is a space in pixels. If \code{NULL} then it draws a solid line. The array 
should have an even length as any odd length array could be expressed as 
a smaller even length array.}

\item{strokeBorderWidth}{Draw a border around graph lines to make crossing 
lines more easily distinguishable. Useful for graphs with many lines.}

\item{strokeBorderColor}{Color for the line border used if 
\code{strokeBorderWidth} is set.}

\item{plotter}{A function which plots the series group. See the 
\href{http://dygraphs.com/tests/plotters.html}{dygraphs documentation} for 
additional details on plotting functions.}
}
\value{
Dygraph with additional series
}
\description{
Add a data series group to a dygraph plot. Note that options will use the default 
global setting (as determined by \code{\link{dyOptions}}) when not specified 
explicitly. Importantly, any dySeries options passed can be passed as a vector of values 
and will be replicated across all series named as part of the group. If arguments differ in
length than the number of series named, then the argument vector will be 
cycled across the named series.
}
\details{
NOTE: dyGroup will turn off \code{stackedGraph}, as the option will calculated cumulatives using
all series in the underlying dygraph, not just a subset.

The dyGroup function can also replicated similar arguments across multiple series, or 
be used to apply a grouped custom plotter - i.e., multi-column plotter - to a subset of the
dygraph data series.
}
\note{
See the 
  \href{https://rstudio.github.io/dygraphs/gallery-series-options.html}{online
  documentation} for additional details and examples.
}
\examples{
\dontrun{
library(dygraphs)

lungDeaths <- cbind(ldeaths, mdeaths, fdeaths)

dygraph(lungDeaths, main = "Deaths from Lung Disease (UK)") \%>\%
  dySeries("fdeaths", stepPlot = TRUE, color = "red") \%>\% 
  dyGroup(c("mdeaths", "ldeaths"), drawPoints = TRUE, color = c("blue", "green"))
}   

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dependency.R
\name{dyDataHandler}
\alias{dyDataHandler}
\title{Include a dygraph data handler}
\usage{
dyDataHandler(dygraph, name, path, version = "1.0")
}
\arguments{
\item{dygraph}{Dygraph to add data handler to}

\item{name}{Name of data handler}

\item{path}{Path to data handler JavaScript file}

\item{version}{Data handler version (e.g. version of package which provides the
data handler)}
}
\value{
A dygraph with the specified data handler enabled.
}
\description{
Include a dygraph data handler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/options.R
\name{dyOptions}
\alias{dyOptions}
\title{dygraph options}
\usage{
dyOptions(dygraph, stackedGraph = FALSE, fillGraph = FALSE,
  fillAlpha = 0.15, stepPlot = FALSE, stemPlot = FALSE,
  drawPoints = FALSE, pointSize = 1, pointShape = c("dot", "triangle",
  "square", "diamond", "pentagon", "hexagon", "circle", "star", "plus", "ex"),
  drawGapEdgePoints = FALSE, connectSeparatedPoints = FALSE,
  strokeWidth = 1, strokePattern = NULL, strokeBorderWidth = NULL,
  strokeBorderColor = "white", plotter = NULL, colors = NULL,
  colorValue = 0.5, colorSaturation = 1, drawXAxis = TRUE,
  drawYAxis = TRUE, includeZero = FALSE, drawAxesAtZero = FALSE,
  logscale = FALSE, axisTickSize = 3, axisLineColor = "black",
  axisLineWidth = 0.3, axisLabelColor = "black", axisLabelFontSize = 14,
  axisLabelWidth = 60, drawGrid = TRUE, gridLineColor = NULL,
  gridLineWidth = 0.3, titleHeight = NULL, rightGap = 5,
  digitsAfterDecimal = 2, labelsKMB = FALSE, labelsKMG2 = FALSE,
  labelsUTC = FALSE, maxNumberWidth = 6, sigFigs = NULL,
  panEdgeFraction = NULL, animatedZooms = FALSE,
  mobileDisableYTouch = TRUE, timingName = NULL, useDataTimezone = FALSE,
  retainDateWindow = FALSE, disableZoom = FALSE)
}
\arguments{
\item{dygraph}{Dygraph to add options to}

\item{stackedGraph}{If set, stack series on top of one another rather than 
drawing them independently. The first series specified in the input data 
will wind up on top of the chart and the last will be on bottom.}

\item{fillGraph}{Should the area underneath the graph be filled? This option 
is not compatible with error bars. This option can also be set on a 
per-series basis.}

\item{fillAlpha}{Transparency for filled regions of the plot. A value of 0.0 
means that the fill will not be drawn, whereas a value of 1.0 means that 
the fill will be as dark as the line of the series itself.}

\item{stepPlot}{When set, display the graph as a step plot instead of a line 
plot. This option can also be set on a per-series basis.}

\item{stemPlot}{When set, display the graph as a stem plot instead of a line
plot. This option can also be set on a per-series basis.}

\item{drawPoints}{Draw a small dot at each point, in addition to a line going
through the point. This makes the individual data points easier to see, but
can increase visual clutter in the chart. This option can also be set on a 
per-series basis.}

\item{pointSize}{The size of the dot to draw on each point in pixels. A dot 
is always drawn when a point is "isolated", i.e. there is a missing point 
on either side of it. This also controls the size of those dots. This 
option can also be set on a per-series basis.}

\item{pointShape}{The shape of the dot to draw. Can be one of the following:
"dot" (default), "triangle", "square", "diamond", "pentagon", "hexagon",
"circle", "star", "plus" or "ex". This option can also be set on a per-series
basis.}

\item{drawGapEdgePoints}{Draw points at the edges of gaps in the data. This 
improves visibility of small data segments or other data irregularities.}

\item{connectSeparatedPoints}{Usually, when dygraphs encounters a missing 
value in a data series, it interprets this as a gap and draws it as such. 
If, instead, the missing values represents an x-value for which only a 
different series has data, then you'll want to connect the dots by setting 
this to true.}

\item{strokeWidth}{The width of the lines connecting data points. This can be
used to increase the contrast or some graphs. This option can also be set 
on a per-series basis.}

\item{strokePattern}{A custom pattern array where the even index is a draw 
and odd is a space in pixels. If null then it draws a solid line. The array
should have a even length as any odd length array could be expressed as a 
smaller even length array. This is used to create dashed lines. This option
can also be set on a per-series basis.}

\item{strokeBorderWidth}{Draw a border around graph lines to make crossing 
lines more easily distinguishable. Useful for graphs with many lines. This 
option can also be set on a per-series basis.}

\item{strokeBorderColor}{Color for the line border used if 
\code{strokeBorderWidth} is set. This option can also be set on a 
per-series basis.}

\item{plotter}{A function (or array of functions) which plot each data series
on the chart. May also be set on a per-series basis. See the 
\href{http://dygraphs.com/tests/plotters.html}{dygraphs documentation} for 
additional details on plotting functions.}

\item{colors}{Character vector of colors for the data series. These can be of
the form "#AABBCC" or "rgb(255,100,200)" or "yellow", etc. If not 
specified, equally-spaced points around a color wheel are used. This option
can also be set on a per-series basis. Note that in both global and 
per-series specification of custom colors you must provide a color for all 
series being displayed. Note also that global and per-series color 
specification cannot be mixed.}

\item{colorValue}{If custom colors are not specified, value of the data 
series colors, as in hue/saturation/value (0.0-1.0, default 0.5).}

\item{colorSaturation}{If custom colors are not specified, saturation of the 
automatically-generated data series colors (0.0-1.0, default 0.5).}

\item{drawXAxis}{Whether to draw the x-axis. Setting this to false also 
prevents x-axis ticks from being drawn and reclaims the space for the chart
grid/lines.}

\item{drawYAxis}{Whether to draw the y-axis. Setting this to false also 
prevents y-axis ticks from being drawn and reclaims the space for the chart
grid/lines.}

\item{includeZero}{Usually, dygraphs will use the range of the data plus some
padding to set the range of the y-axis. If this option is set, the y-axis 
will always include zero, typically as the lowest value. This can be used 
to avoid exaggerating the variance in the data.}

\item{drawAxesAtZero}{When set, draw the X axis at the Y=0 position and the Y
axis at the X=0 position if those positions are inside the graph's visible 
area. Otherwise, draw the axes at the bottom or left graph edge as usual.}

\item{logscale}{When set the graph shows the y-axis in log scale. Any values 
less than or equal to zero are not displayed.}

\item{axisTickSize}{The spacing between axis labels and tick marks.}

\item{axisLineColor}{Color of the x- and y-axis lines. Accepts any value 
which the HTML canvas strokeStyle attribute understands, e.g. 'black' or 
'rgb(0, 100, 255)'. This can also be set on a per-axis basis.}

\item{axisLineWidth}{Thickness (in pixels) of the x- and y-axis lines. This 
can also be set on a per-axis basis.}

\item{axisLabelColor}{Color for x- and y-axis labels. This is a CSS color 
string. This may also be set on a per-axis basis.}

\item{axisLabelFontSize}{Size of the font (in pixels) to use in the axis 
labels, both x- and y-axis. This may also be set on a per-axis basis.}

\item{axisLabelWidth}{Width (in pixels) of the containing divs for x- and 
y-axis labels.}

\item{drawGrid}{Whether to display grid lines in the chart. This may be set 
on a per-axis basis to define the visibility of each axis' grid separately.
Defaults to \code{TRUE} for x and y, and \code{FALSE} for y2.}

\item{gridLineColor}{The color of the grid lines. This option can also be set
on a per-series basis.}

\item{gridLineWidth}{Thickness (in pixels) of the grid lines drawn under the 
chart. This option can also be set on a per-series basis.}

\item{titleHeight}{Height of the chart title, in pixels. This also controls 
the default font size of the title. If you style the title on your own, 
this controls how much space is set aside above the chart for the title's 
div.}

\item{rightGap}{Number of pixels to leave blank at the right edge of the 
Dygraph. This makes it easier to highlight the right-most data point.}

\item{digitsAfterDecimal}{Unless it's run in scientific mode (see the 
\code{sigFigs} option), dygraphs displays numbers with 
\code{digitsAfterDecimal} digits after the decimal point. Trailing zeros 
are not displayed, so with a value of 2 you'll get '0', '0.1', '0.12', 
'123.45' but not '123.456' (it will be rounded to '123.46'). Numbers with 
absolute value less than 0.1^digitsAfterDecimal (i.e. those which would 
show up as '0.00') will be displayed in scientific notation.}

\item{labelsKMB}{Show K/M/B for thousands/millions/billions on y-axis.}

\item{labelsKMG2}{Show k/M/G for kilo/Mega/Giga on y-axis. This is different 
than \code{labelsKMB} in that it uses base 2, not 10.}

\item{labelsUTC}{Show date/time labels according to UTC (instead of local 
time). Note that this option cannot is incompatible with
\code{useDataTimezone} (you must use one or the other).}

\item{maxNumberWidth}{When displaying numbers in normal (not scientific) 
mode, large numbers will be displayed with many trailing zeros (e.g. 
100000000 instead of 1e9). This can lead to unwieldy y-axis labels. If 
there are more than maxNumberWidth digits to the left of the decimal in a 
number, dygraphs will switch to scientific notation, even when not 
operating in scientific mode. If you'd like to see all those digits, set 
this to something large, like 20 or 30.}

\item{sigFigs}{By default, dygraphs displays numbers with a fixed number of 
digits after the decimal point. If you'd prefer to have a fixed number of 
significant figures, set this option to that number of significant figures.
A value of 2, for instance, would cause 1 to be display as 1.0 and 1234 to 
be displayed as 1.23e+3.}

\item{panEdgeFraction}{A value representing the farthest a graph may be 
panned, in percent of the display. For example, a value of 0.1 means that 
the graph can only be panned 10% pased the edges of the displayed values. 
null means no bounds.}

\item{animatedZooms}{Set this option to animate the transition between zoom 
windows. Applies to programmatic and interactive zooms. Note that if you 
also set a drawCallback, it will be called several times on each zoom. If 
you set a zoomCallback, it will only be called after the animation is 
complete.}

\item{mobileDisableYTouch}{Set this option to automatically disable touch
events on the Y axis for mobile devices (since this interferes with 
swiping/scrolling on mobile devices).}

\item{timingName}{Set this option to log timing information. The value of the
option will be logged along with the timing, so that you can distinguish 
multiple dygraphs on the same page.}

\item{useDataTimezone}{Whether to use the time zone of the underlying xts 
object for display. Defaults to \code{FALSE} which uses the time zone of 
the client workstation. Note that this option is incompatible with 
\code{labelsUTC} (you must use one or other other).}

\item{retainDateWindow}{Whether to retain the user's current date window 
(zoom level) when updating an existing dygraph with new data and/or 
options.}

\item{disableZoom}{Set this option to disable click and drag zooming.}
}
\value{
dygraph with additional options
}
\description{
Add options to a dygraph plot.
}
\note{
See the \href{https://rstudio.github.io/dygraphs/}{online documentation}
  for additional details and examples.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/annotation.R
\name{dyAnnotation}
\alias{dyAnnotation}
\title{Annotation for dygraph chart}
\usage{
dyAnnotation(dygraph, x, text, tooltip = NULL, width = NULL,
  height = NULL, cssClass = NULL, tickHeight = NULL,
  attachAtBottom = FALSE, clickHandler = NULL, mouseOverHandler = NULL,
  mouseOutHandler = NULL, dblClickHandler = NULL, series = NULL)
}
\arguments{
\item{dygraph}{Dygraph to add an annotation to}

\item{x}{Either numeric or date value indicating where to place the
annotation. For date value, this should be of class \code{POSIXct} or
convertible to \code{POSIXct}.}

\item{text}{Text to overlay on the chart at the location of x}

\item{tooltip}{Additional tooltip text to display on mouse hover}

\item{width}{Width (in pixels) of the annotation flag.}

\item{height}{Height (in pixels) of the annotation flag.}

\item{cssClass}{CSS class to use for styling the annotation.}

\item{tickHeight}{Height of the tick mark (in pixels) connecting the point to
its flag or icon.}

\item{attachAtBottom}{If true, attach annotations to the x-axis, rather than 
to actual points.}

\item{clickHandler}{JavaScript function to call when an annotation is 
clicked.}

\item{mouseOverHandler}{JavaScript function to call when the mouse hovers 
over an annotation.}

\item{mouseOutHandler}{JavaScript function to call when the mouse exits an 
annotation.}

\item{dblClickHandler}{JavaScript function to call when an annotation is 
double clicked.}

\item{series}{Series to attach the annotation to. By default, the last series
defined using \code{\link{dySeries}}.}
}
\value{
Dygraph with specified annotation
}
\description{
Define a text annotation for a data-point on a dygraph chart.
}
\note{
Annotations are bound to specific series in the input data. If you have
  only one series or if you define annotations immediately after a call to 
  \code{\link{dySeries}} then you need not specify the series explicitly. 
  Otherwise, you should use the \code{series} parameter to indicate which
  series the annotation should be bound to.
  
  Annotation event handlers can also specified globally (see 
  \code{\link{dyCallbacks}}).

See the
  \href{https://rstudio.github.io/dygraphs/gallery-annotations.html}{online
  documentation} for additional details and examples.
}
\examples{
library(dygraphs)
  
dygraph(presidents, main = "Presidential Approval") \%>\%
  dyAxis("y", valueRange = c(0, 100)) \%>\%
  dyAnnotation("1950-7-1", text = "A", tooltip = "Korea") \%>\%
  dyAnnotation("1965-1-1", text = "B", tooltip = "Vietnam")   
  
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dependency.R
\name{dyPlotter}
\alias{dyPlotter}
\title{Include a dygraph plotter}
\usage{
dyPlotter(dygraph, name, path, version = "1.0")
}
\arguments{
\item{dygraph}{Dygraph to add plotter to}

\item{name}{Name of plotter}

\item{path}{Path to plotter JavaScript file}

\item{version}{Plotter version (e.g. version of package which provides the plotter)}
}
\value{
A dygraph with the specified plotter enabled.
}
\description{
Include a dygraph plotter
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/css.R
\name{dyCSS}
\alias{dyCSS}
\title{CSS for dygraph labels and legend}
\usage{
dyCSS(dygraph, css)
}
\arguments{
\item{dygraph}{Dygraph to add CSS styles to}

\item{css}{Path to css file to be used for styling textual elements of the 
graph.}
}
\value{
dygraph with additional CSS styles
}
\description{
Apply custom CSS to the text drawn within a dygraph. See the 
\href{http://dygraphs.com/css.html}{CSS documentation} on the dygraphs 
website for additional details on which styles are available.
}
\note{
See the 
  \href{https://rstudio.github.io/dygraphs/gallery-css-styling.html}{CSS 
  Styling} article on the dygraphs for R website for additional details.
  
  Note that CSS styles are global so will affect all dygraphs on a given web
  page. This also implies that for a page with multiple plots you only need
  to specify styles for the first one (alternatively you can just add them
  directly to the page by other means).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/legend.R
\name{dyLegend}
\alias{dyLegend}
\title{dygraph legend}
\usage{
dyLegend(dygraph, show = c("auto", "always", "onmouseover", "follow",
  "never"), width = 250, showZeroValues = TRUE, labelsDiv = NULL,
  labelsSeparateLines = FALSE, hideOnMouseOut = TRUE)
}
\arguments{
\item{dygraph}{Dygraph to configure legend options for.}

\item{show}{When to display the legend. Specify "always" to always show the 
legend. Specify "onmouseover" to only display it when a user mouses over 
the chart. Specify "follow" to have the legend show as overlay to the chart
which follows the mouse. The default behavior is "auto", which results in
"always" when more than one series is plotted and "onmouseover" when only a
single series is plotted.}

\item{width}{Width (in pixels) of the div which shows the legend.}

\item{showZeroValues}{Show zero value labels in the legend.}

\item{labelsDiv}{Show data labels in an external div, rather than on the 
graph. This value should be a div element id.}

\item{labelsSeparateLines}{Put a <br/> between lines in the label string. 
Often used in conjunction with \code{labelsDiv}.}

\item{hideOnMouseOut}{Whether to hide the legend when the mouse leaves the 
chart area. This option applies when \code{show} is set to "onmouseover". 
Note that this also affects the hiding of the \code{\link{dyHighlight}} on 
mouse out.}
}
\value{
A dygraph with customized legend options
}
\description{
Configure options for the dygraph series legend.
}
\note{
See the 
  \href{https://rstudio.github.io/dygraphs/gallery-plot-labels.html}{online
  documentation} for additional details and examples.
}
\examples{
library(dygraphs)

dygraph(nhtemp, main = "New Haven Temperatures") \%>\% 
  dySeries("V1", label = "Temperature (F)") \%>\%
  dyLegend(show = "always", hideOnMouseOut = FALSE)   
    
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/event.R
\name{dyEvent}
\alias{dyEvent}
\title{dygraph event line}
\usage{
dyEvent(dygraph, x, label = NULL, labelLoc = c("top", "bottom"),
  color = "black", strokePattern = "dashed", date)
}
\arguments{
\item{dygraph}{Dygraph to add event line to}

\item{x}{Either numeric or date/time for the event, depending on the format
of the x-axis of the dygraph. (For date/time must be a \code{POSIXct}
object or another object convertible to \code{POSIXct} via
\code{as.POSIXct})}

\item{label}{Label for event. Defaults to blank.}

\item{labelLoc}{Location for label (top or bottom)}

\item{color}{Color of event line. This can be of the form "#AABBCC" or 
"rgb(255,100,200)" or "yellow". Defaults to black.}

\item{strokePattern}{A predefined stroke pattern type ("dotted", "dashed",
"dotdash", or "solid") or a custom pattern array where the even index is 
a draw and odd is a space in pixels. Defaults to dashed.}

\item{date}{(deprecated) See argument \code{x}.}
}
\value{
A dygraph with the specified event line.
}
\description{
Add a vertical event line to a dygraph
}
\note{
See the \href{https://rstudio.github.io/dygraphs/gallery-event-lines.html}{online 
  documentation} for additional details and examples.
}
\examples{
library(dygraphs)

dygraph(presidents, main = "Presidential Approval") \%>\%
  dyAxis("y", valueRange = c(0, 100)) \%>\%
  dyEvent("1950-6-30", "Korea", labelLoc = "bottom") \%>\%
  dyEvent("1965-2-09", "Vietnam", labelLoc = "bottom")
dygraph(presidents, main = "Presidential Approval") \%>\%
  dyAxis("y", valueRange = c(0, 100)) \%>\%
  dyEvent(c("1950-6-30", "1965-2-09"), c("Korea", "Vietnam"), labelLoc = "bottom") 
 
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dependency.R
\name{dyDependency}
\alias{dyDependency}
\title{Add external assets as a dygraph dependency}
\usage{
dyDependency(dygraph, dependency)
}
\arguments{
\item{dygraph}{Dygraph to add dependency to}

\item{dependency}{An HTML dependency}
}
\value{
A dygraph with the specified dependency added.
}
\description{
Add external assets as a dygraph dependency
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rebase.R
\name{dyRebase}
\alias{dyRebase}
\title{Rebase data handler for straw broom charts with Dygraph}
\usage{
dyRebase(dygraph, value = 100, percent = FALSE)
}
\arguments{
\item{dygraph}{Dygraph to draw chart on}

\item{value}{Value to rebase to}

\item{percent}{If true, ingnore value argument and rebase to percentage difference}
}
\value{
Dygraph with specified straw broom chart
}
\description{
Draw a straw broom chart.
}
\examples{
\dontrun{
library(quantmod)
tickers <- c("AAPL", "MSFT")
getSymbols(tickers)
closePrices <- do.call(merge, lapply(tickers, function(x) Cl(get(x))))
dateWindow <- c("2008-01-01", "2009-01-01")
dygraph(closePrices, main = "Value", group = "stock") \%>\%
  dyRebase(value = 100) \%>\%
  dyRangeSelector(dateWindow = dateWindow)
dygraph(closePrices, main = "Percent", group = "stock") \%>\%
  dyRebase(percent = TRUE) \%>\%
  dyRangeSelector(dateWindow = dateWindow)
dygraph(closePrices, main = "None", group = "stock") \%>\%
  dyRangeSelector(dateWindow = dateWindow)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dygraph.R
\name{dygraph-shiny}
\alias{dygraph-shiny}
\alias{dygraphOutput}
\alias{renderDygraph}
\title{Shiny bindings for dygraph}
\usage{
dygraphOutput(outputId, width = "100\%", height = "400px")

renderDygraph(expr, env = parent.frame(), quoted = FALSE)
}
\arguments{
\item{outputId}{output variable to read from}

\item{width, height}{Must be a valid CSS unit (like \code{"100\%"},
\code{"400px"}, \code{"auto"}) or a number, which will be coerced to a
string and have \code{"px"} appended.}

\item{expr}{An expression that generates a dygraph}

\item{env}{The environment in which to evaluate \code{expr}.}

\item{quoted}{Is \code{expr} a quoted expression (with \code{quote()})? This 
is useful if you want to save an expression in a variable.}
}
\description{
Output and render functions for using dygraph within Shiny 
applications and interactive Rmd documents.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plugins.R
\name{dyUnzoom}
\alias{dyUnzoom}
\title{The dyUnzoom plugin adds an "Unzoom" button to the graph when it's displaying
in a zoomed state (this is a bit more discoverable than the default double-
click gesture for unzooming).}
\usage{
dyUnzoom(dygraph)
}
\arguments{
\item{dygraph}{Dygraph to add plugin to}
}
\value{
Dygraph with Unzoom plugin enabled
}
\description{
The dyUnzoom plugin adds an "Unzoom" button to the graph when it's displaying
in a zoomed state (this is a bit more discoverable than the default double-
click gesture for unzooming).
}
\examples{
library(dygraphs)
dygraph(ldeaths) \%>\%
  dyRangeSelector() \%>\%
  dyUnzoom()

}
